#!/usr/bin/env python

#^#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import sys
import socket
import signal
import os
import logging
import logging.handlers
from os.path import dirname, join, exists, isdir, basename
from os import O_CREAT, O_RDWR, O_NONBLOCK
from fcntl import fcntl, flock, LOCK_EX, LOCK_NB, LOCK_UN, F_GETFL, F_SETFL
from OVSConstants import DLM_DOMAIN, AGENT_LIVENESS_LOCK, LOG_DIR, XEND_CONFIG
from OVSDaemon import daemonize
from OVSCommons import run_cmd
from OVSXHA import ha_config_xend_external_lock

#----------------------------------------------------------------------
#global variables
#asm/fcntl.h
#define O_NOATIME       01000000
O_NOATIME = 01000000

dlm_lock_fd = -1

#----------------------------------------------------------------------
#handler for NOWAYOUT lock request
def lock_handler(signum, frame):
    global dlm_lock_fd
    logging.info("receive signal=%d" % signum)
    if dlm_lock_fd != -1:
        if signum == signal.SIGUSR1:
            logging.info("disable NOWAYOUT lock.")
            flags = fcntl(dlm_lock_fd, F_GETFL)
            fcntl(dlm_lock_fd, F_SETFL, flags & ~O_NOATIME)
            ha_config_xend_external_lock(operation="remove")
            
        if signum == signal.SIGUSR2:
            logging.info("enable NOWAYOUT lock.")
            flags = fcntl(dlm_lock_fd, F_GETFL)
            fcntl(dlm_lock_fd, F_SETFL, flags | O_NOATIME)
            ha_config_xend_external_lock(operation="add")

#setup signal trap 
signal.signal(signal.SIGUSR1, lock_handler) #SIGUSR1=10, disable NOWAYOUT
signal.signal(signal.SIGUSR2, lock_handler) #SIGUSR2=12, enable NOWAYOUT
 
#----------------------------------------------------------------------
def main():
    #setup the logging system, no dependency to the log server.
    formatter = logging.Formatter(fmt="%(asctime)s %(levelname)s=> %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler = logging.handlers.RotatingFileHandler(join(LOG_DIR, "ovs_guard.log"), "a", 1024*1024, 3)
    handler.setFormatter(formatter)
    logging.getLogger("").setLevel(logging.DEBUG)
    logging.getLogger("").addHandler(handler)

    #only startup the guard process when the agent is alive
    agent_alive = False
    fd = os.open(AGENT_LIVENESS_LOCK, O_CREAT)
    try:
        try:
            flock(fd, LOCK_EX | LOCK_NB)
            flock(fd, LOCK_UN)
        except:
            agent_alive = True
    finally:
        os.close(fd)
    if not agent_alive:
        logging.error("Agent Server is not yet running! nothing to guard.")
        sys.exit(1)

    #run as daemon
    guard_name =  basename(__file__).split(".")[0] 
    daemonize(guard_name)
    
    #join the domain
    from OVSXHA import ha_join_dlm_domain
    rs = ha_join_dlm_domain(DLM_DOMAIN)
    if not rs.startswith("success"):
        raise Exception(rs)
    ha_config_xend_external_lock(operation="add")
     
    #main process
    logging.info("Agent Server is running, guard it!")
    hostname = socket.gethostname().split(".")[0] 
    dlm_lock = join(DLM_DOMAIN, hostname)
        
    try:
        #acquire the NOWAYOUT lock.
        global dlm_lock_fd 
        dlm_lock_fd = os.open(dlm_lock, O_CREAT | O_RDWR | O_NONBLOCK | O_NOATIME)
        #block wait on agent's EX lock, which should be hold by the Agent Server
        if not exists(AGENT_LIVENESS_LOCK):
            raise Exception("Agent Server not exists.")
        while True:
            agent_alive_lock_fd = os.open(AGENT_LIVENESS_LOCK, O_CREAT)
            try:
                try:
                    logging.error("OVSAgentGuard: try to get the EX lock on '%s'" % AGENT_LIVENESS_LOCK)
                    flock(agent_alive_lock_fd, LOCK_EX)
                    logging.error("OVSAgentGuard: release the EX lock on '%s'" % AGENT_LIVENESS_LOCK)
                    flock(agent_alive_lock_fd, LOCK_UN)
                    logging.info("Agent Server died.")
                    break
                except Exception, e:
                    logging.error("OVSAgentGuard: main() => %s" % str(e))
            finally:
                os.close(agent_alive_lock_fd)
    except Exception, e:
        logging.error("guard: error. => %s %s" % (e.__class__, e))
        sys.exit(1)

if __name__ == "__main__":
    main()

