#!/usr/bin/python
#^#
# OVSAsyncTaskAux.py
# Asynchronous task Auxiliary
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
# hongxi.ma@oracle.com 2008.1.20
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import logging
from os import getpid, kill
from OVSCDS import CDS
from OVSCommons import exposed, fmt_err_msg

#----------------------------------------------------------------
# Asynchronous task Auxiliary interfaces 
#----------------------------------------------------------------
class AsyncenvAgentcallProxy:
    def __init__(self, server_proxy, func_name):
        self.sp = server_proxy
        self.func = func_name

    def __getattr__(self, name):
        if in_asyncenv():
            return self.exec_func
        else:
            return getattr(self.sp, name)

    def exec_func(self, *params):
        if in_asyncenv():
            taskid = get_asynctaskid()
            return self.sp.directcall_in_asyncenv(self.func, *(params + (taskid,)))
        else:
            return getattr(self.sp, self.func)(*params)

def agentcall_in_asyncenv(server_proxy, func, *params):
    if in_asyncenv():
        taskid = get_asynctaskid()
        return server_proxy.directcall_in_asyncenv(func, *(params + (taskid,)))
    else:
        return getattr(server_proxy, func)(*params)

_ataskid = None

#Last param should be taskid, this function is called by OVSAsyncTaskCore.py
@exposed
def directcall_in_asyncenv(func, *params):
    global _ataskid
    logger = logging.getLogger("ovs.operation")
    try:
        if len(params) == 0:
            raise Exception('params not enough:at least async_task_id should be passed in as last one')
        taskid = params[-1]
        _ataskid = taskid
        associate_asynctaskid(taskid)
        save_asynctask_funcparams(taskid, func, params[:-1])
        rs = direct_agent_api_call(func, params[:-1])
        logger.info('directcall_in_asyncenv: success. %s %s => %s' % (func, str(params), rs))
        return rs
    except Exception, e: 
        errmsg = fmt_err_msg(e)
        logger.error('directcall_in_asyncenv: failed! %s %s => %s' % (func, str(params), errmsg))
        return 'failed:%s' % (errmsg,)

@exposed
def remote_async_task_kill(taskid):
    logger = logging.getLogger("ovs.operation")
    try:
        pids = get_async_auxdata(taskid, 'pids')
        if pids is None or len(pids) == 0:
            raise Exception('process id for task: %s not found' % taskid)
        pid = pids[0] #or kill all the pids right now?
        try:
            kill(pid, 9)
        except:
            pass
        if len(pids) > 1: #remove pid is also ok?
            pids.pop(0)
            pids.append(pid)
            save_async_auxdata3(taskid, {'pids': pids})
        logger.info('remote_async_task_kill: success. %s => %s' % (taskid, pid))
        return 'success:%s' % pid
    except Exception, e: 
        errmsg = fmt_err_msg(e)
        logger.error('remote_async_task_kill: failed! %s => %s' % (taskid, errmsg))
        return 'failed:%s' % (errmsg,)
            
    
def associate_asynctaskid(taskid):
    pid = getpid()
    cds = CDS('ataskaux')
    try:
        if not taskid in cds.db_ataskaux:
            cds.db_ataskaux[taskid] = {'pids': [pid]}
        else:
            ad = cds.db_ataskaux[taskid]
            pids = ad.get('pids', [])
            pids.append(pid)
            cds.db_ataskaux[taskid] = ad
    finally:
        cds.close()

def save_asynctask_funcparams(taskid, func, params):
    cds = CDS('ataskaux')
    try:
        ad = cds.db_ataskaux.get(taskid, None)
        if ad != None:
            fp = ad.get('funcparams', {})
            fp[func] = params
            cds.db_ataskaux[taskid] = ad
    finally:
        cds.close()


def in_asyncenv():
    taskid = get_asynctaskid()
    return taskid != '' and taskid != None

def get_asynctaskid():
    global _ataskid
    if _ataskid == None:
        pid = getpid()
        _ataskid = ''
        cds = CDS('ataskaux')
        try:
            for taskid in cds.db_ataskaux:
                if pid in cds.db_ataskaux[taskid].get('pids', []):
                    _ataskid = taskid
                    break
        finally:
            cds.close()
    return _ataskid


def save_async_auxdata(key, value):
    return save_async_auxdata2({key: value})

# data as {'utility_agent':'https://oracle:oracle@oops2.cn.oracle.com:/8899'}    
def save_async_auxdata2(data):
    if in_asyncenv():
        return save_async_auxdata3(get_asynctaskid(), data)    

def save_async_auxdata3(taskid, data):
    cds = CDS('ataskaux')
    try:
        ad = cds.db_ataskaux.get(taskid, None)
        if ad != None:
            ad.update(data)
            cds.db_ataskaux[taskid] = ad
            return True
    finally:
        cds.close()

def get_async_auxdata(taskid, key):
    ad = get_async_auxdata2(taskid)
    if ad != None and key in ad:
        return ad[key]
        
def get_async_auxdata2(taskid):
    cds = CDS('ataskaux')
    try:
        return cds.db_ataskaux.get(taskid, None)        
    finally:
        cds.close()

def get_async_func_params(taskid, func_name = ''):
    fp = get_async_auxdata(taskid, 'funcparams')    
    if fp is None: 
        return
    if len(func_name) > 0 and func_name in fp:
        return fp[func_name]
    if len(func_name) == 0 and len(fp) == 1:
        return fp.values[0]
    
def direct_agent_api_call(func, params):
    import OVSServices
    func_addr = None
    for mod_name in dir(OVSServices):
        if mod_name.startswith('_'):
            continue
        mod = getattr(OVSServices, mod_name)
        if func not in dir(mod):
            continue 
        func_addr = getattr(mod, func)
        break
    if func_addr == None:
        raise Exception('function: %s not found in direct_agent_api_call' % func)
    return func_addr(*params)
