#!/usr/bin/python
#^#
# OVSAsyncTaskCore.py
# Asynchronous task scheduler
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
# hongxi.ma@oracle.com 2008.1.18
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#
#---------------------------------------------------------------
#TODO: when pid reaches 32768, it will start from 1 again, as so
# there is bugs in current async code, since in db we associate 
# taskid with pid, one of the solution is clear the pid information
# in some proper time: service start, remove_async_task, ......
#---------------------------------------------------------------
import logging
from OVSCDS import CDS
from OVSServerProxy import ServerProxy
from OVSCommons import exposed, fmt_err_msg

#----------------------------------------------------------------
# Asynchronous proxy 
#----------------------------------------------------------------
class AsyncProxy:
    async_method_prefix = 'async_'
    def _dispatch(self, method, args):
        if not method.startswith(self.async_method_prefix):
            raise Exception('method %s is not supported' % method)
        real_method = method[len(self.async_method_prefix):].strip()
        if real_method == '':
            raise Exception('method %s is invalid' % method)
        return new_async_task(real_method, * args)        
#----------------------------------------------------------------
# constants 
#----------------------------------------------------------------
TASK_STATUS_RUNNING = 1
TASK_STATUS_END = 0
TASK_STATUS_ERROR = -1
TASK_STATUS_CANCEL = -2
TASK_STATUS_TEXT = {TASK_STATUS_ERROR:'error', TASK_STATUS_RUNNING:'running', TASK_STATUS_END:'end', TASK_STATUS_CANCEL:'cancel'}

#----------------------------------------------------------------
# Asynchronous task scheduler interfaces 
#----------------------------------------------------------------
@exposed
def new_async_task(func, *params):
    return new_async_task_ext('', func, *params)

@exposed
def new_async_task_ext(agent_url, func, *params):
    logger = logging.getLogger("ovs.operation")
    try:
        taskid = new_task(agent_url, func, params)
        logger.info('new_async_task_ext: success. %s->%s %s => %s' % (agent_url, func, str(params), taskid))
        return 'success:%s' % taskid    
    except Exception, e: 
        errmsg = fmt_err_msg(e)
        logger.error('new_async_task_ext: failed! %s->%s %s => %s' % (agent_url, func, str(params), errmsg))
        return 'failed:%s' % (errmsg,)

@exposed
def get_async_task_status(taskid):
    logger = logging.getLogger("ovs.query")
    try:
        ti = get_task_info(taskid)
        if ti == None:
            raise Exception('task with id: %s not exist' % taskid)
        #TODO
        #check if pid still alive, if not change status to 'killed'
        if get_task_running_status(taskid) == True:
            ti = get_task_info(taskid)
        ti_status = ti['status']
        ti_message = ti.get('message', '')
        rs = '%s:%s' % (TASK_STATUS_TEXT[ti_status], ti_message)
        logger.info('get_async_task_status: success. taskid: %s => %s' % (taskid, rs))
        return 'success:%s' % rs    
    except Exception, e: 
        errmsg = fmt_err_msg(e)
        logger.error('get_async_task_status: failed! taskid: %s => %s' % (taskid, errmsg))
        return 'failed:%s' % (errmsg,)
    
@exposed
def cancel_async_task(taskid):
    logger = logging.getLogger("ovs.operation")
    try:
        ti = get_task_info(taskid)
        if ti == None:
            raise Exception('task with id: %s not exist') % taskid
        if ti['status'] != TASK_STATUS_RUNNING:
            raise Exception('Task is not running( but %s), can not be cancelled' % TASK_STATUS_TEXT[ti['status']])
        kill_task(taskid)
        cancel_task(taskid)
        logger.info('cancel_async_task: success. taskid: %s' % (taskid,))
        return 'success'
    except Exception, e: 
        errmsg = fmt_err_msg(e)
        logger.error('cancel_async_task: failed! taskid: %s => %s' % (taskid, errmsg))
        return 'failed:%s' % (errmsg,)

@exposed
def remove_async_task(taskid, force = 0):
    logger = logging.getLogger("ovs.operation")
    try:
        ti = get_task_info(taskid)
        if ti == None:
            raise Exception('task with id: %s not exist') % taskid
        ti_status = ti['status']
        if ti_status == TASK_STATUS_RUNNING:
            if force == 1:
                cancel_async_task(taskid, 1)
            else:
                raise Exception('task:%s is running, not removed' % taskid)
        remove_task(taskid)    
        logger.info('remove_async_task: success. taskid: %s' % (taskid,))
        return 'success'
    except Exception, e: 
        errmsg = fmt_err_msg(e)
        logger.error('remove_async_task: failed! taskid: %s => %s' % (taskid, errmsg))
        return 'failed:%s' % (errmsg,)


#----------------------------------------------------------------
# internal functions
#----------------------------------------------------------------
#TODO
#should also check are tere any task marked 'running', since a cold-reset may happened, 
#if so, change the status to 'killed', and remove the pids field,since they are meaningless.
def init_asyncenv():
    from os import remove
    def force_remove(filename):
        try:
            remove(filename)
        except:
            pass
    from OVSConstants import CONFIG_DIR
    from os.path import join
    db_base = join(CONFIG_DIR, 'db')
    force_remove(join(db_base, 'atask.db'))
    force_remove(join(db_base, 'atask.lock'))
    force_remove(join(db_base, 'ataskaux.db'))
    force_remove(join(db_base, 'ataskaux.lock'))
    
    
_local_agent_url = ''
def local_agent_url():
    global _local_agent_url
    if _local_agent_url == '':
        from OVSConfig import get_config
        config = get_config()
        _admin_username = "admin"
        _admin_passwd = config.get("security", "admin.password")
        if config.get("security", "ssl").lower().startswith("enable"):
            protocol = "https"
        else: 
            protocol = "http" 
        _local_agent_url = "%s://%s:%s@%s:%s" % \
            (protocol, _admin_username, _admin_passwd, "127.0.0.1", config.get("agent", "port"))
    return _local_agent_url

def agent_api_call(agent_url, func, params):
    if agent_url == '':
        agent_url = local_agent_url()
    sp = ServerProxy(agent_url)
    rs = getattr(sp, func)(*params)
    return rs

def agent_api_call_success(rs):
    return rs.startswith('success')

def agent_api_call_result(rs):
    idx = rs.find(':')
    if idx > -1:
        return rs[idx + 1:]
    else:
        return ''
    

def exec_async_task(taskid, agent_url, func, params):
    from os import fork, _exit
    pid = fork()
    if pid != 0:
        return pid
    try:
        try:
            params = (func,) + params + (taskid,)
            rs = agent_api_call(agent_url, 'directcall_in_asyncenv', params)
        except Exception, e:
            errmsg = fmt_err_msg(e)
            set_task_info(taskid, {'status': TASK_STATUS_ERROR, 'message': errmsg})
        else:
            #put rs into async.db and update status as done
            set_task_info(taskid, {'status': TASK_STATUS_END, 'message': rs})
    except Exception, e: 
        errmsg = fmt_err_msg(e)
        logger = logging.getLogger("ovs.operation")
        logger.error('exec_async_task: failed! taskid: %s %s->%s %s => %s' % (taskid, agent_url, func, str(params), errmsg))
    _exit(0)    

def kill_task(taskid):
    from os import kill
    ti = get_task_info(taskid)
    if ti != None:
        pid = ti.get('pid', -1)
        if pid > 0:
            try:
                kill(pid, 9)
            except:
                pass
        agent_url = ti['agent_url']
        params = (taskid,)
        rs = agent_api_call(agent_url, 'remote_async_task_kill', params)
        return agent_api_call_success(rs)

def cancel_task(taskid):
    ti = get_task_info(taskid)
    if ti != None:
        agent_url = ti['agent_url']
        func = ti['func'] + '_cancel'
        params = (taskid,)
        rs = None
        try:
            rs = agent_api_call(agent_url, func, params)
        except Exception, e:
            s = fmt_err_msg(e)
            if s.find('method %s is not supported' % func) < 0:
                message = 'failed:%s' % s
            else:
                message = 'success:no additional cancel interface defined'
        else:
            message = str(rs)
                
        set_task_info(taskid, {'status': TASK_STATUS_CANCEL, 'message': message})
        if rs != None:
            return agent_api_call_success(rs)

def get_task_running_status(taskid):
    ti = get_task_info(taskid)
    if ti != None and ti['status'] == TASK_STATUS_RUNNING:
        try:
            agent_url = ti['agent_url']
            func = ti['func'] + '_status'
            params = (taskid,)
            rs = agent_api_call(agent_url, func, params)
            result = agent_api_call_success(rs)
            if result == True:
                try:
                    set_task_info(taskid, {'message': agent_api_call_result(rs)})
                except:
                    pass
            return result
        except:
            pass
        
def new_uuid():
    from sys import version_info
    if version_info[0] == 2 and version_info[1] >= 5:
        from uuid import uuid1
        uuid_string = str(uuid1())
    else:
        #python 2.3 and 2.4 doesn't have uuid module, so a bit stupid
        from commands import getoutput
        uuid_string = getoutput('uuidgen')
    return uuid_string.replace('-','').upper()

def new_taskid(agent_url, func, params):
    return new_uuid()

def new_task(agent_url, func, params):
    taskid = ''
    cds = CDS('atask')
    try:
        while(True):
            taskid = new_taskid(agent_url, func, params)
            if not taskid in cds.db_atask:
                break
        pid = exec_async_task(taskid, agent_url, func, params)
        if pid > 0:
            ti = {'agent_url': agent_url, 'func': func, 'params': params, 'status': TASK_STATUS_RUNNING, 'pid': pid}
            cds.db_atask[taskid] = ti
        else:
            raise Exception('failed to fork new process')
    finally:
        cds.close()
    return taskid

def remove_task(taskid):
    cds = CDS('atask')
    try:
        if taskid in cds.db_atask:
            return cds.db_atask.pop(taskid, None)
    finally:
        cds.close()    

def set_task_info(taskid, info):
    cds = CDS('atask')
    try:
        if not taskid in cds.db_atask:
            raise Exception('task: %s is not saved.' %  taskid)
        ti = cds.db_atask[taskid]
        if ti['status'] != TASK_STATUS_RUNNING:
            raise Exception('only running task can be updated.')
        ti.update(info)
        cds.db_atask[taskid] = ti
    finally:
        cds.close()

def get_task_info(taskid):
    cds = CDS('atask')
    try:
        return cds.db_atask.get(taskid)
    finally:
        cds.close()
