#!/usr/bin/env python

#^#
# OVSCommons.py
#
# common function used in agent.
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import zlib
import traceback
import sys
import string
import re
import random
import popen2
import os
from subprocess import Popen, PIPE
from os.path import abspath, exists, isfile, isdir, join, basename, dirname
from StringIO import StringIO
from OVSConfig import get_config
from OVSConstants import OVS_BASE

def parse_result(result):
    """
    all return result following same convension, (our own marshal mechanism):
    success                         #if success completed without return value
    success:key=value;key=value...  #if success completed with return result
    failed:errmsg                   #failed with error message
    """
    if result.startswith("failed"): return

    #e.g. result="success:key=value;key2=value2..."
    d = {}
    result = result[8:].split(";")
    if len(result) == 1 and result[0].strip() == "":
        return d
    for kv in result:
        k,v = kv.split("=")
        d[k] = v
    return d

def host_alive(ip):
    """Determine whether the given host is alive."""
    status = os.popen("ping -q -c 2 %s" % ip).close()
    return not status

def fmt_err_msg(e):
    errmsg = "<%s: %s>" % (e.__class__.__name__, e)
    try:
        config = get_config()
        if config.get("debug", "traceback").startswith("enable"):
            tb = sys.exc_info()[2]
            if tb:
                sio = StringIO()
                try:
                    traceback.print_tb(tb, 10, sio)
                    errmsg = "%s \nStackTrace: \n%s" % (errmsg, sio.getvalue())
                finally:
                    sio.close()
    except:
        pass
    return errmsg

def run_cmd(cmd):
    """
    run a cmd and catch its output, if run into any problem raise exception
    """
    #N.B.: !!security checkpoint!!
    #possible cmd injection, like sql injection
    p = popen2.Popen3(cmd, True)
    if p.wait() != 0:
        raise Exception('%s=>%s' % (cmd, p.childerr.read()))
    return p.fromchild.read()

def run_cmd_bg(cmd):
    #run cmd in background without wait.
    #NOTE: even implemented as a daemon, redirect 0,1,2 is still necessary or
    #you will not get a clean ssh disconnect
    devnull = open(os.devnull, "rw")
    p = Popen(cmd.split(), stdin=devnull, stdout=devnull, stderr=devnull, close_fds=True)

def chk_vm_path(vm_path):
    """
    check path agaist vm operation
    """
    vm_path = abspath(vm_path)
    config = get_config()
    paths = config.get("security", "vm.store").split(":")
    for path in paths:
        if vm_path.startswith(path.strip()): return
    raise Exception("'%s' is not inside vm store!" % vm_path)

def exposed(func):
    #---------------------------------------
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    #---------------------------------------
    wrapper.exposed = True
    wrapper.real_func = func
    return wrapper

def randomMAC():
    """Generate a random MAC address.

    Uses OUI (Organizationally Unique Identifier) 00-16-3E, allocated to
    Xensource, Inc. The OUI list is available at
    http://standards.ieee.org/regauth/oui/oui.txt.

    The remaining 3 fields are random, with the first bit of the first
    random field set 0.

    @return: MAC address string
    """
    mac = [ 0x00, 0x16, 0x3e,
            random.randint(0x00, 0x7f),
            random.randint(0x00, 0xff),
            random.randint(0x00, 0xff) ]
    return ':'.join(map(lambda x: "%02x" % x, mac))

def marshall_string(s):
    """
    compress with zlib, then encode base64
    """
    return zlib.compress(s).encode("base64").replace("\n", "")

def unmarshall_string(s):
    """
    decode base64, then decompress with zlib
    """
    return zlib.decompress(s.decode("base64"))


def parse_shared_fs():
    shared_fs, shared_fs_type, shared_fs_line = "", "", ""
    #f = open("/proc/mounts")
    #FIXME:
    #entry in /proc/mounts may not be consistent with mount
    #mount -t nfs -o rw,sync,vers=3,rsize=32768,wsize=32768,hard,proto=tcp,timeo=600,retrans=2,sec=sys,addr=ostest105.cn.oracle.com ostest105.cn.oracle.com:/OVS /OVS
    #the addr=ostest105.cn.oracle.com option is not recognized by mount
    f = open("/etc/mtab")
    try:
        for line in f:
            parts = line.split()
            if parts[1] == OVS_BASE:
                #bugz#6707
                if shared_fs:
                    raise Exception("more than one filesystems are mounted on /OVS, please verify.")
                shared_fs, shared_fs_type = parts[0], parts[2]
                shared_fs_line = line
    finally:
        f.close()

    return shared_fs, shared_fs_type, shared_fs_line

def parse_cluster_conf():
    cfg_file = "/etc/ocfs2/cluster.conf"
    f = open(cfg_file)
    lines = f.readlines()
    f.close()

    conf = { "nodes": [], "cluster": None }
    for line in lines:
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        elif line.startswith("node:"):
            d = {}
            conf["nodes"].append(d)
        elif line.startswith("cluster:"):
            d = {"name": "ocfs2"} #with default value
            conf["cluster"] = d
        else:
            parts = map(string.strip, line.split("="))
            if len(parts) == 2:
                d[parts[0]] = parts[1]

    return conf

def check_shutdown_mode():
    from OVSConstants import AGENT_SHUTDOWN_MODE_FLAG
    if exists(AGENT_SHUTDOWN_MODE_FLAG):
        raise Exception("operation cannot be invoked when agent in shutdown mode!")

def sync_file_attr(path):
    #[bugz 6759], sync file attribute by lock/unlock
    if not isfile(path):
        return
    try:
        from fcntl import lockf, LOCK_EX, LOCK_UN
        fd = os.open(path, os.O_RDWR)
        lockf(fd, LOCK_EX)
        lockf(fd, LOCK_UN)
        os.close(fd)
    except:
        pass

def get_host_by_name(hostname):
    ptn_ip = re.compile(r"\d+\.\d+\.\d+\.\d+")
    if ptn_ip.match(hostname):
        return hostname

    ip = ""
    try:
        import socket
        #NOTE: if return value is not constant, we treat it as configuration error.
        ip = socket.gethostbyname(hostname)
        if ip.startswith("127") and not hostname.startswith("localhost"):
                stdout = run_cmd("host %s" % hostname)
                l = [ line.split()[-1] for line in stdout.splitlines()]
                l = [_ip for _ip in l if ptn_ip.match(_ip)]
                l.sort()
                if l:
                    ip = l[0]
    except:
        pass
    return ip

