#!/usr/bin/env python
#^#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import string
import re
import os
import logging
from time import sleep
from os.path import exists
from os import O_CREAT, O_RDWR, O_NONBLOCK
from OVSXHA import ha_check_cluster_heartbeat
from OVSSiteDB import get_live_vm_info
from OVSLog import init_log
from OVSDaemon import daemonize
from OVSDB import db_update_srv, db_get_srv
from OVSConstants import DLM, DLM_DOMAIN
from OVSConfig import init_config, get_config
from OVSCommons import run_cmd, fmt_err_msg, parse_cluster_conf
from OVSCDS import CDS

def get_my_ips():
    ptn = re.compile(r"inet addr:([.0-9]+)")
    lines = run_cmd("/sbin/ifconfig")
    ll = [] #list local
    for line in lines.splitlines():
        m = ptn.search(line)
        if m:
            ip = m.group(1)
            ll.append(ip)
    return ll

#get my cluster ip -- the ip alive and in /etc/ocfs2/cluster.conf
def get_my_member_ip():
    ll = get_my_ips()
    d = parse_cluster_conf()
    lc = [node["ip_address"] for node in d["nodes"]]
    l = list(set(ll).intersection(set(lc)))
    if len(l) > 0:
        return l[0]

def is_master():
    rs = False
    logger = logging.getLogger()
    try:
        my_ips = get_my_ips()
        config = get_config()
        if config.get("ha", "remaster").startswith("enable"):
            cds = CDS("master")
            try:
                mdb = cds.db_master
                if "master_vip" in mdb:
                    vip = mdb["master_vip"]
                    if vip in get_my_ips():
                        rs = True
            finally:
                cds.close()
        else:
            for ip in my_ips:
                try:
                    srv = db_get_srv(ip)
                    if "site" in srv["srv_role"]:
                        rs = True
                        break
                except:
                    pass
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("is_master: error. => %s" % errmsg)

    return rs

def check_server_status():
    logger = logging.getLogger()

    trigger_recovery = False
    #check ocfs2 cluster status
    rs = ha_check_cluster_heartbeat()
    if not rs.startswith("success"):
        return trigger_recovery
    #check debugfs status
    livenodes_file = "/sys/kernel/debug/o2hb/livenodes"
    try:
        if not exists(livenodes_file):
            cmd = "mount -t debugfs debugfs /sys/kernel/debug"
            run_cmd(cmd)
            if not exists(livenodes_file):
                raise Exception("debugfs cannot be mounted!")
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("check_server_status: error. => %s" % errmsg)
        return trigger_recovery

    #main process
    try:
        #cluster members
        my_ip = get_my_member_ip()
        d = parse_cluster_conf()
        members = dict([(node["ip_address"], node["number"]) for node in d["nodes"]])
        if my_ip in members:
            del members[my_ip]
        #live node map
        f = open(livenodes_file)
        livenodes = f.read().split()
        f.close()

        for ip in members:
            try:
                try:
                    srv = db_get_srv(ip)
                except:
                    #member in ocfs2 cluster but not in the same pool
                    continue
                num = members[ip]
                if num not in livenodes:
                    if "real_status" not in srv or srv["real_status"] != "DEAD":
                        db_update_srv(ip, {"real_status": "DEAD"})
                        logger.error("check_server_status: srv('%s') goes DEAD!" % ip)
                        trigger_recovery = True
                else:
                    if "real_status" not in srv or srv["real_status"] != "ALIVE":
                        db_update_srv(ip, {"real_status":"ALIVE"})
                        logger.error("check_server_status: srv('%s') comes to ALIVE!" % ip)
                        trigger_recovery = True
            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.error("check_server_status: ip('%s') error. => %s" % (ip, errmsg))
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("check_server_status: error. =>%s" % errmsg)

    return trigger_recovery

def recover_process():
    #recover process according to policy
    import OVSPolicyServer
    OVSPolicyServer.execute_policy()

def run():
    logger = logging.getLogger()
    cnt = 0
    while True:
        cnt = cnt % 10000
        try:
            init_config() #let the config modification take effect immediate
            config = get_config()

            srv_monitor_interval, vm_status_sync_interval, collect_vm_status = 10, 60, False
            try:
                srv_monitor_interval = int(config.get("ha", "server.monitor.interval"))
                vm_status_sync_interval = int(config.get("ha", "policy.execute.interval"))
            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.error("OVSMonitorServer.run(): error.=>%s" % errmsg)

            if is_master():
                if vm_status_sync_interval < srv_monitor_interval \
                        or cnt % (vm_status_sync_interval / srv_monitor_interval) == 0:
                    #cannot collect vm status more frequently than collect server status
                    collect_vm_status = True

                trigger_recovery = check_server_status()
                if trigger_recovery or collect_vm_status:
                    #logging.debug("OVSMonitorServer.run(): trigger collect vm status.")
                    get_live_vm_info()
                if trigger_recovery and config.get("ha", "policy.execute").startswith("enable"):
                    logging.debug("OVSMonitorServer.run(): trigger recovery process.")
                    recover_process()
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.error("OVSMonitorServer.run(): error. => %s" % errmsg)

        #sleep and next cycle ...
        sleep(srv_monitor_interval)

def main():
    from os.path import basename
    pid_filename = basename(__file__).split(".")[0]
    daemonize(pid_filename)

    init_log("client")
    run()

if __name__ == "__main__":
    main()
