#!/usr/bin/env python
#^#
# OVSRemasterServer.py
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

# dlmfs document
# http://www.mjmwired.net/kernel/Documentation/filesystems/dlmfs.txt

import socket
import re
import os
import logging
from time import sleep
from os.path import exists, join
from OVSXHA import ha_check_cluster_heartbeat
from OVSLog import init_log
from OVSDaemon import daemonize
from OVSConstants import DLM, DLM_DOMAIN, DLM_LOCK_MASTER
from OVSConfig import init_config, get_config
from OVSCommons import fmt_err_msg, run_cmd, get_host_by_name
from OVSCDS import CDS

ETXTBUSY = 26

def _get_vip_dev(vip, not_used=False):
    cmd = "/sbin/ifconfig"
    output = run_cmd(cmd)
    lines = output.splitlines()
    d = {}
    ptn = re.compile(r"inet addr:([0-9.]+)")
    for idx, line in enumerate(lines):
        if not line.startswith(" "):
            if line.strip():
                dev = line.split()[0]
                m = ptn.search(lines[idx + 1])
                if m:
                    ip = m.group(1)
                    d[ip] = dev
    if not_used:
        if vip not in d:
            vs = d.values()
            vs.sort()
            tgt_dev, tgt_idx = '', -1
            for dev in vs:
                #FIXME: any other identifier for nic?
                if dev.startswith("eth"):
                    if tgt_dev and not dev.startswith(tgt_dev):
                        break #break when we got one
                    parts = dev.split(":")
                    if len(parts) == 2:
                        tgt_dev = parts[0]
                        tgt_idx = int(parts[1])
                    else:
                        tgt_dev = parts[0]
                        tgt_idx = -1

            if tgt_dev:
                return "%s:%d" % (tgt_dev, tgt_idx + 1)
    else:
        if vip in d:
            return d[vip]

#release_dlm_lock has to be controled by external -- agtctl
def release_dlm_lock():
    try:
        if exists(DLM_LOCK_MASTER):
            os.remove(DLM_LOCK_MASTER)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        print "release_dlm_lock(): error. =>%s" % errmsg

def release_master_vip():
    try:
        dev = ''
        cds = CDS("master")
        try:
            mdb = cds.db_master
            if "master_vip" in mdb:
                vip = mdb["master_vip"]
                dev = _get_vip_dev(vip)
        finally:
            cds.close()

        if dev:
            cmd = "/sbin/ifconfig %s down" % dev
            run_cmd(cmd)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        #print "release_master_vip(): error. =>%s" % errmsg

def cleanup_master_env():
    release_master_vip()
    release_dlm_lock()

def setup_master_env():
    #TODO: setup vip or report that i am the master.
    logger = logging.getLogger()
    try:
        cds = CDS("master")
        try:
            mdb = cds.db_master
            hostname = socket.gethostname()
            master_real_ip = get_host_by_name(hostname)
            mdb["master_ip"] = master_real_ip
            if "master_vip" in mdb:
                vip = mdb["master_vip"]
                dev = _get_vip_dev(vip, not_used=True)
                if dev:
                    cmd = "/sbin/ifconfig %s %s" % (dev, vip)
                    run_cmd(cmd)
                else:
                    raise Exception("cannot find any device to bind!")
                cmd = "/sbin/route add -host %s %s" % (vip, dev)
                run_cmd(cmd)
        finally:
            cds.close()

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("setup_master_env(): failed. => %s" % errmsg)


master_lock = None

def acquire_master_lock():
    global master_lock
    logger = logging.getLogger()
    try:
        rs = ha_check_cluster_heartbeat()
        if not rs.startswith("success"):
            raise Exception("acquire_master_lock(): ocfs2 dlmfs not ready!")
        
        from OVSXHA import ha_join_dlm_domain
        rs = ha_join_dlm_domain(DLM_DOMAIN)
        if not rs.startswith("success"):
            raise Exception(rs)

        #"""
        # close() does not downconvert the lock level nor does it drop the lock. The
        # holder still owns the lock at that level. All close() does is allows any
        # downconvert request to succeed.
        # However, a downconvert request is only generated for queued requests. And
        # O_NONBLOCK is specifically a noqueue dlm request.
        #"""

        #here is the situation,
        # 1) O_CREAT | O_NONBLOCK will create a lock file if it does not exist, whether
        #    we are the lock holder or not.
        # 2) if we hold O_RDWR lock, and we close but not delete it, we still hold it.
        #    afterward, O_RDWR will succeed, but O_RDWR | O_NONBLOCK will not.
        # 3) but if we donnot hold the lock, O_RDWR will hang there waiting,
        #    which is not the desirable -- any uninterruptable hang is undesirable.
        # 4) if noboday else hold the lock either, but the lock file exists as side effect
        #    of 1), with O_NONBLOCK, it may result in ETXTBUSY
        #so,
        # a) we need O_NONBLOCK to avoid 3)
        # b) we need to delete it ourself to avoid 2) (for e.g. agtctl restart), but when?
        #   *) if not succeed in 1), remove the lock file (side product), avoid 4)
        #   *) if everything is good, we drop it and we clean out.
        #   *) if stopped with agtctl (killed), agtctl should remove the lock.
        #   *) if crashed, agtctl stop --force should clean out it.

        try:
            if not exists(DLM_LOCK_MASTER):
                #logger.info("acquire_master_lock(): touch lock '%s'." % DLM_LOCK_MASTER)
                fd = os.open(DLM_LOCK_MASTER, os.O_CREAT | os.O_NONBLOCK)
                os.close(fd)

            #logger.info("acquire_master_lock(): create exclusive lock '%s'." % DLM_LOCK_MASTER)
            master_lock = os.open(DLM_LOCK_MASTER, os.O_RDWR | os.O_NONBLOCK)
            #so if we've reached here without error, we've got the lock. close it and hold it.
            os.close(master_lock)

            logger.info("acquire_master_lock(): got the lock and i'm the MASTER!")
            return True
        except OSError, e:
            if e.errno == ETXTBUSY:
                try:
                    # if we donnot remove the lock file (the side effect of touch),
                    # we may never succeed to hold the exclusive lock, i.e. 4)
                    if exists(DLM_LOCK_MASTER):
                        os.remove(DLM_LOCK_MASTER)
                except Exception, e1:
                    errmsg = fmt_err_msg(e1)
                    logger.warning("acquire_master_lock(): remove lock file failed. %s" % errmsg)
            else:
                errmsg = fmt_err_msg(e)
                logger.debug("acquire_master_lock(): try to lock the file. => %s" % errmsg)
                raise e

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.info("acquire_master_lock(): error. =>%s" % errmsg)

def run():
    global master_lock
    logger = logging.getLogger()
    try:
        while True:
            init_config() #let the config modification take effect immediate
            config = get_config()
            server_remaster_interval = 5
            try:
                server_remaster_interval = int(config.get("ha", "server.remaster.interval"))
            except:
                pass
            # grap the lock only when we are not the holder
            # once OVSRemasterServer hold the lock, it will never drop it.
            # similar to Oracle DB, startup -> mount -> open -> shutdown,
            # it is a oneway nonreversible operation
            if not master_lock \
                    or master_lock == -1 \
                    or not exists(DLM_LOCK_MASTER):
                if acquire_master_lock():
                    setup_master_env()
                else: #somebody else is the master
                    sleep(server_remaster_interval)
            else: #i am the master
                sleep(server_remaster_interval)
    finally:
        release_dlm_lock()

def main():
    from os.path import basename
    pid_filename = basename(__file__).split(".")[0]
    daemonize(pid_filename)

    init_log("client")
    run()

if __name__ == "__main__":
    main()

