#!/usr/bin/env python
#^#
# OVSServerSelection.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#
import logging
from OVSServerProxy import ServerProxy
from OVSCommons import exposed, fmt_err_msg, parse_result
from OVSCDS import CDS
from OVSDB import db_get_srv, db_update_srv, db_get_vm, db_get_real_vm_path, db_update_vm
from OVSXXenVMConfig import xen_get_vm_type
from OVSException import *

#----------------------------------------<> helper functions <>----
def select_tgt(l, field, goal="max"):
    """
    l is a list of (k, d) and d is a dict with (field in d)
    """
    if not l: 
        return
    (vk, vd) = l[0]
    if goal == "max":
        for (k,d) in l[1:]:
            if vd[field] < d[field]:
                (vk, vd) = (k, d)
    else:
        for (k,d) in l[1:]:
            if vd[field] > d[field]:
                (vk, vd) = (k, d)
    return vk

def _get_agent_version(agt_url):
    sp = ServerProxy(agt_url)
    try:
        rs = getattr(sp, "get_agent_version")()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)
    except:
        rs = {"agent_version": "2.2-15"} # this is the base release

    t = []
    #2.2-15 => (2, 2, 0, 15)
    parts = rs["agent_version"].split("-")
    #version 2.2 => (2, 2, 0)
    _parts = parts[0].split(".")
    for i in range(0, 3):
        try: t.append(int(_parts[i]))
        except: t.append(0)
    #release
    try: t.append(int(parts[1]))
    except: t.append(0)
    return tuple(t)

#----------------------------------------<> <>----
def collect_srv_stat(candidates=None):
    """
    collect server statistics data throu agent
    save collected info to srv.db

    @param candidates    list of hostnames
    """
    logger = logging.getLogger("ovs.performance")

    cds = CDS("srv")
    try:
        if not candidates:
            candidates = cds.db_srv.keys()
        else:
            candidates = [candidate for candidate in candidates if candidate in cds.db_srv]
    finally:
        cds.close()

    for hostname in candidates:
        try:
            srv = db_get_srv(hostname)
            sp = ServerProxy(srv['agt_url'])
            #TODO: combine these 2 rpc calls into 1
            rs = sp.sys_vmstat_info()
            rs = parse_result(rs)
            srv["cpu_idle"] = float(rs["cpu_idle"])
            srv["mem_free"] = int(float(rs["mem_free"]))
            #FIXME: any better algorithm to describe the cpu power?
            if "cpu_speed" not in srv:
                rs = sp.sys_cpu_info()
                if not rs.startswith("success"): 
                    raise Exception(rs)
                rs = parse_result(rs)
                srv["cpu_num"] = int(rs["cpu_num"])
                srv["cpu_speed"] = int(float(rs["bogomips"]))
            srv["cpu_avail"] = int(srv["cpu_num"]) * float(srv["cpu_speed"]) * float(srv["cpu_idle"])

            #info from hypervisor's view
            try:
                vtech = "xen"
                if "srv_role" in srv and vtech in srv["srv_role"]:
                    try:
                        rs = getattr(sp, "%s_hypervisor_info" % vtech)()
                        rs = parse_result(rs)
                        srv["mem_free"] = int(float(rs["mem_free"]))
                        srv["vcaps"] = rs["vcaps"]
                    except:
                        logger.warning("collect_srv_stat: srv('%s') may not yet support api: %s_hypervisor_info()" % \
                            (hostname, vtech))
                        #fallback
                        rs = getattr(sp, "%s_mem_free" % vtech)()
                        rs = parse_result(rs)
                        srv["mem_free"] = int(float(rs["mem_free"]))
                        srv["vcaps"] = "UNKNOWN"
            except Exception, e:
                logger.warning("collect_srv_stat: exception %s" % fmt_err_msg(e))

            #agent version info
            srv['agent_version'] = _get_agent_version(srv["agt_url"])

            srv["status"] = "RUNNING"
        except KeyError, e:
            #when unregister one of the candidates occurs before this loop
            errmsg = fmt_err_msg(e)
            logger.warning("collect_srv_stat: srv('%s') removed?=>%s" % \
                    (hostname, errmsg))
            continue
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.warning("collect_srv_stat: srv('%s') error=>%s" % \
                    (hostname, errmsg))
            srv["status"] = "UNKNOWN"

        db_update_srv(hostname, srv)

    logger.info("collect_srv_stat: success. %s" % repr(candidates))

@exposed
def select_tgt_host(vm_path, any=False, excludes=None, with_attr=None):
    """
    select server for running the vm
    ~-- balancing strategy --~
    1) enough memory to accommodate the new vm
    2) the fast one in candidates 1)
    this is only one of the strategies, other possiblities for e.g.:
    *) chose the one with memory best-fit size of vm memory
    *) ...

    @param excludes
    @param with_attr
    these 2 param are for migrate_vm,
    1) excludes itself or some specific server
    2) server arch must be the same

    """
    logger = logging.getLogger("ovs.operation")
    #---------------------------------------------------------------------
    def is_subset(d1, d2):
        if d1 is None: return True
        try:
            for k in d1:
                if d1[k] != d2[k]: return False
        except:
            return False
        return True

    def capable_to_run(srv_caps, vm_type):
        #FIXME: currently we only check para or hvm
        #vm_type in [xen, hvm]
        #bit detection will be added later
        #logger.debug("capable_to_run: srv_caps=%s, vtype=%s" % (srv_caps, vm_type))
        if srv_caps == "UNKNOWN":
            return True
        if srv_caps.lower().find(vm_type.lower().strip()) != -1:
            return True
    #---------------------------------------------------------------------

    #check if vm got "vtype" attr
    vm = db_get_vm(vm_path)
    if "vtype" in vm:
        vtech, vtype = vm["vtech"], vm["vtype"]
    else:
        real_vm_path = db_get_real_vm_path(vm_path)
        rs = xen_get_vm_type(real_vm_path)
        if not rs.startswith("success"):
            raise Exception("check vm type exception! =>%s" % rs)
        rs = parse_result(rs)
        if rs["type"] == "para": #para -> xen, hvm -> hvm
            rs["type"] = "xen"
        db_update_vm(vm_path, {"real_path": real_vm_path, "vtype": rs["type"]})
        vtech = vm["vtech"]
        vtype = rs["type"]

    #-------------------------<>main process<>------------------------------
    tgt_host = None
    cds = CDS("srv")
    try:
        candidates = [ hostname for (hostname, srv) in cds.db_srv.items()
                        if vtech in srv["srv_role"] and
                            srv["status"] == "RUNNING" and
                            not srv.get("shutdown_mode") and
                            is_subset(with_attr, srv) ]
        if "preferred_servers" in vm:
            preferred_servers = vm["preferred_servers"]
            if preferred_servers:
                candidates = list(set(candidates).intersection(set(preferred_servers)))
        if isinstance(excludes, list) or isinstance(excludes, tuple):
            candidates = list(set(candidates) - set(excludes))
    finally:
        cds.close()

    if not candidates:
        raise OVSException("no server of type '%s' ready to run the vm." % vm['vtech'], ERR_SERVER_SELECTION)

    #sync server status
    collect_srv_stat(candidates)

    if any:
        #any server not depends on vm memory size
        #for e.g. vm installation
        srvs = {}
        cds = CDS("srv")
        try:
            for hostname in candidates:
                if hostname in cds.db_srv:
                    srv = cds.db_srv[hostname]
                    if srv["status"] == "RUNNING":
                        srvs[hostname] = srv
        finally:
            cds.close()
        #here we choose server with max memory
        tgt_host = select_tgt(srvs.items(), "mem_free")
        srv = srvs[tgt_host]
        sp = ServerProxy(srv['agt_url'])
        if "agent_version" in srv:
            sp.agent_version = srv["agent_version"]
        else:
            sp.agent_version = _get_agent_version(srv['agt_url'])

        logger.info("select_tgt_host: vm('%s') will run on srv('%s')" % \
                (vm_path, tgt_host))
        return (tgt_host, sp)


    #FIXME:
    #infact, if the vm claim to use 1G ram, the xen server
    #won't allocate so much ram, it will dynamically allocate memory
    #so a host server with 2 G mem may accommodate 3 vms each with 2 G ram
    real_vm_path = db_get_real_vm_path(vm_path)
    (hostname, sp) = select_one_server(candidates=candidates)
    rs = getattr(sp, "%s_get_memory" % vtech)(real_vm_path)
    if rs.startswith("success"):
        rs = parse_result(rs)
        if "memory" in rs:
            memory = int(rs["memory"])
        else: #compatible with ovs-agent-2.2-15
            memory = int(rs["mem_size"])
        logger.debug("select_tgt_host: vm('%s')=>memory=%d" % (vm_path, memory))
    else:
        raise Exception(rs)

    srvs = {}
    cds = CDS("srv")
    try:
        for hostname in candidates:
            if hostname in cds.db_srv:
                srv = cds.db_srv[hostname]
                if (srv["status"] == "RUNNING") and (srv["mem_free"] >= memory):
                    if "vcaps" in srv: #filter server without capability
                        if not capable_to_run(srv["vcaps"], vtype):
                            continue
                    srvs[hostname] = srv
    finally:
        cds.close()

    if not srvs:
        raise OVSException("no server selected to run vm('%s') memory=%d" % \
            (vm_path, memory), ERR_SERVER_SELECTION)
    #FIXME:we change our police to run vm on srv with max free mem.
    tgt_host = select_tgt(srvs.items(), "mem_free")
    srv = srvs[tgt_host]
    sp = ServerProxy(srv['agt_url'])
    if "agent_version" in srv:
        sp.agent_version = srv["agent_version"]
    else:
        sp.agent_version = _get_agent_version(srv['agt_url'])

    logger.info("select_tgt_host: vm('%s') will run on srv('%s')" % \
            (vm_path, tgt_host))
    return (tgt_host, sp)

def get_utility_server(preference=None):
    """
    there may be multiple utility servers.
    balancing policy is 'the one with the maximum cpu resources'.
    """
    if preference:
        tgt_host = preference
        srv = db_get_srv(tgt_host)
    else:
        #FIXME: should we add another db to store utility,xen,site as index?
        cds = CDS("LOCK_SH", "srv")
        try:
            candidates = [k for k in cds.db_srv if "utility" in cds.db_srv[k]["srv_role"]]
        finally:
            cds.close()

        if not candidates:
            raise OVSException("no available utility server!", ERR_SERVER_SELECTION)

        if len(candidates) == 1:
            tgt_host = candidates[0]
            srv = db_get_srv(tgt_host)
        else:
            collect_srv_stat(candidates)
            cds = CDS("srv")
            try:
                l = [ (hostname, srv) for (hostname, srv) in cds.db_srv.items()
                        if (hostname in candidates) and (srv["status"] == "RUNNING") ]
            finally:
                cds.close()
            if not l: 
                raise OVSException("no available utility server!", ERR_SERVER_SELECTION)
            tgt_host = select_tgt(l, "cpu_avail")
            srv = dict(l)[tgt_host]

    sp = ServerProxy(srv['agt_url'])
    if "agent_version" in srv:
        sp.agent_version = srv["agent_version"]
    else:
        sp.agent_version = _get_agent_version(srv['agt_url'])

    return (tgt_host, sp)

def select_one_server(vtech="xen", candidates=None):
    """
    get one server of type vtech, to get cfg etc lite operation
    """
    vtech = vtech.lower().strip()
    tgt_host = None
    if not candidates:
        cds = CDS("srv")
        try:
            candidates = [ hostname for (hostname, srv) in cds.db_srv.items()
                        if (vtech in srv["srv_role"]) and (srv["status"] == "RUNNING") ]
        finally:
            cds.close()
    if not candidates:
        raise OVSException("no server of type '%s'." % vtech, ERR_SERVER_SELECTION)

    tgt_host = candidates[0]
    srv = db_get_srv(tgt_host)
    sp = ServerProxy(srv['agt_url'])
    if "agent_version" in srv:
        sp.agent_version = srv["agent_version"]
    else:
        sp.agent_version = _get_agent_version(srv['agt_url'])

    return (tgt_host, sp)

def get_vm_host(vm_path):
    """
    get vm host server from site db
    """
    vm = db_get_vm(vm_path)
    host_srv = vm['host_srv']
    if not host_srv.strip():
        raise Exception("cannot find host server for vm('%s')" % vm_path)
    try:
        srv = db_get_srv(host_srv)
    except Exception, e:
        raise Exception("get host server('%s') for vm('%s') failed! => %s" % \
                (host_srv, vm_path, fmt_err_msg(e)))
    sp = ServerProxy(srv['agt_url'])
    if "agent_version" in srv:
        sp.agent_version = srv["agent_version"]
    else:
        sp.agent_version = _get_agent_version(srv['agt_url'])

    return (host_srv, sp)

def get_master_server():
    """
    @return (hostname, serverproxy) of master server
    """
    hostname, sp = None, None
    cds = CDS("master")
    try:
        #only available after registered to the master server
        mdb = cds.db_master
        hostname = mdb.get("master_hostname")
        agt_url = mdb.get("master_agt_url")
        if agt_url:
            sp = ServerProxy(agt_url)
    finally:
        cds.close()

    return hostname, sp

