#!/usr/bin/env python
#^#
# OVSSiteDB.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import re
import logging
from os.path import join, basename, exists
from OVSXXenVMConfig import xen_update_uuid
from OVSXUtility import find_real_path
from OVSServerSelection import collect_srv_stat, get_utility_server
from OVSServerProxy import ServerProxy
from OVSDB import db_get_vm, db_update_vm, db_del_vm, db_get_real_vm_path
from OVSDB import db_get_srv, db_update_srv, db_del_srv, db_get_master_srv
from OVSConstants import DEFAULT_AGENT_USERNAME, DEFAULT_AGENT_PASSWD, DEFAULT_AGENT_PORT, OVS_BASE
from OVSCommons import exposed, fmt_err_msg, parse_result, chk_vm_path, host_alive, get_host_by_name
from OVSCDS import CDS

#------------------<> register/unregister server/vm <>----
@exposed
def register_server(hostname, srv_role, is_shared=True,
        agt_username=DEFAULT_AGENT_USERNAME,
        agt_passwd=DEFAULT_AGENT_PASSWD,
        agt_port=DEFAULT_AGENT_PORT,
        ssl_enabled=True):
    """
    NOTE:
      when register a server, agent on the server must be up and running
    for collecting server infomation

    @param srv_role "site", "xen" or "utility"
    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.debug("register_server: hostname=%s, srv_role=%s, is_shared=%s, agt_username=%s, agt_port=%s, ssl_enable=%s" % \
                (hostname, srv_role, repr(is_shared), agt_username, agt_port, repr(ssl_enabled)))

        #check if master server already registered.
        srv_role = srv_role.lower()
        if "site" in srv_role:
            cds = CDS("srv")
            try:
                candidates = [_hostname for (_hostname, srv) in cds.db_srv.items()
                        if (srv.has_key("srv_role") and "site" in srv["srv_role"])]
            finally:
                cds.close()

            if len(candidates) > 1:
                raise Exception("More than one server pool master exists: %s" % repr(candidates))
            if len(candidates) == 1 and candidates[0] != hostname:
                raise Exception("Master pool server already exists: %s" % candidates[0])

        #check if server with same ip is already registered.
        ip = get_host_by_name(hostname)
        cds = CDS("srv", "srvp")
        try:
            for _hostname in cds.db_srv:
                if _hostname != hostname and get_host_by_name(_hostname) == ip:
                    raise Exception("srv('%s') with ip('%s') is already registered." % (_hostname, ip))
            cds.db_srvp[ip] = hostname
        finally:
            cds.close()

        srv = {
            #is in shared server pool, vm can be balanced to this server.
            "is_shared": (is_shared and True or False), #FIXME: not used
            "srv_role": srv_role,
            "ip": ip,
            "agt_addr": hostname,
            "agt_username": agt_username,
            "agt_passwd": agt_passwd,
            "agt_port": agt_port,
            "agt_url": "http%s://%s:%s@%s:%s" % \
                (ssl_enabled and "s" or "", agt_username, agt_passwd, hostname, agt_port),
            }

        #logger.debug("register_server: agt_url=%s" % srv['agt_url'])
        sp = ServerProxy(srv["agt_url"])
        rs = sp.sys_os_info()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)
        srv["arch"] = rs["arch"]
        srv["cpu_num"] = int(rs["cpu_num"])
        srv["cpu_speed"] = float(rs["bogomips"])
        srv["mem_total"] = int(rs["mem_total"])
        srv["status"] = "RUNNING"
        if "hostname" in rs:
            if rs["hostname"].startswith("localhost"):
                raise Exception("hostname cannot be 'localhost*'!")
            srv["hostname"] = rs["hostname"]
        if "xen" in srv_role:
            try:
                rs = sp.xen_hypervisor_info()
                if rs.startswith("success"):
                    rs = parse_result(rs)
                    srv["vcaps"] = rs["vcaps"]
                    srv["mem_free"] = rs["mem_free"]
                else:
                    raise Exception(rs)
            except Exception, e: #xen_hypervisor_info() may not be supported in 2.2-15(oracle vm 2.1)
                errmsg = fmt_err_msg(e)
                logger.warning("register_server: warning. hostname=%s => %s" % \
                    (hostname, errmsg))
            
        #register master to the member
        if "site" in srv_role:
            rs = sp.register_master(srv["agt_url"], hostname)
            if not rs.startswith("success"):
                raise Exception(rs)
        else:
            master_srv = db_get_master_srv()
            if not master_srv:
                raise Exception("master server is not registered yet! please register master server first.")
            try:
                rs = sp.register_master(master_srv["agt_url"], hostname)
                if not rs.startswith("success"):
                    raise Exception(rs)
            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.warning("register_server: reverse register master to member failed. => %s" % errmsg)

        db_update_srv(hostname, srv, True)

        #add hostname to ocfs2 cluster        
        try:
            from OVSSiteHA import ha_sync_ocfs2_cluster_conf 
            rs = ha_sync_ocfs2_cluster_conf()
            if not rs.startswith("success"):
                raise Exception(rs)
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.warning("register_server:('%s') failed to sync cluster.conf => %s" % (hostname, errmsg))


        logger.info("register_server: success. %s->%s" % (hostname, srv_role))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("register_server: failed! %s->%s => %s" \
                % (hostname, srv_role, errmsg))
        return "failed:%s" % (errmsg,)

@exposed
def unregister_server(hostname):
    """
    unregister a server from site

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        srv, sp = {}, None
        try:
            srv = db_get_srv(hostname)
            sp = ServerProxy(srv['agt_url'])
            sp.unregister_master()
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.warning("unregister_server: failed on get server proxy for '%s' => %s" % (hostname, errmsg))

        _srv = {}
        _srv.update(srv)
        db_del_srv(hostname)

        try:
            if ("srv_role" in _srv) and ("site" in _srv["srv_role"].lower()) and sp:
                sp.truncate_db()
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.warning("unregister_server: failed on truncate_db %s => %s" % (hostname, errmsg))

        #remove hostname from ocfs2 cluster        
        try:
            from OVSSiteHA import ha_sync_ocfs2_cluster_conf 
            rs = ha_sync_ocfs2_cluster_conf()
            if not rs.startswith("success"):
                raise Exception(rs)
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.warning("unregister_server:('%s') failed to sync cluster.conf => %s" % (hostname, errmsg))

        logger.info("unregister_server: success. srv('%s') removed." % hostname)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unregister_server: failed. srv('%s') => %s" % \
                (hostname, errmsg))
        return "failed:%s" % errmsg

@exposed
def registered_server(hostname):
    """
    whether a server is registered or not.

    @return         "success" if registered else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        srv = db_get_srv(hostname)
        logger.info("registered_server: success. srv('%s') exists." % hostname)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("registered_server: failed. srv('%s') => %s" % \
                (hostname, errmsg))
        return "failed:%s" % errmsg

@exposed
def register_vm(vm_path, vtech="xen",
        agt_username=DEFAULT_AGENT_USERNAME,
        agt_passwd=DEFAULT_AGENT_PASSWD,
        agt_port=DEFAULT_AGENT_PORT,
        packaged=False):
    """
    register vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        #try to update the uuid for the vm
        real_path = ""
        try:
            real_path = find_real_path(vm_path)
            vm_cfg_file = join(real_path, "vm.cfg")
            if exists(vm_cfg_file):
                xen_update_uuid(vm_cfg_file)
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.warning("register_vm: warning. vm('%s')=> %s" % (vm_path, errmsg))

        d = {
            "vtech": vtech.lower().strip(),
            "agt_addr": "",
            "agt_username": agt_username,
            "agt_passwd": agt_passwd,
            "agt_port": agt_port,
            "real_path": real_path,
            "host_srv": "",
            "packaged": packaged,
            "status": "DOWN",
            "readlock": 0,
        }
        db_update_vm(vm_path, d)

        logger.info("register_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("register_vm: failed. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def unregister_vm(vm_path):
    """
    unregister vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        db_del_vm(vm_path)
        logger.info("unregister_vm: success. vm('%s')" % (vm_path,))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unregister_vm: failed. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_hostserver(vm_path):
    """
    get vm host server, if down, then hostname should be empty

    @return         "success:host_srv=<hostname>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_get_vm(vm_path)

        logger.info("get_vm_hostserver: success. vm('%s') => host_srv='%s'" % \
                (vm_path, vm['host_srv']))
        return "success:host_srv=%s" % vm["host_srv"]
    except Exception, e:
        errmsg = errmsg
        logger.error("get_vm_hostserver: error. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_path(host, vm_name, vtech="xen"):
    """
    get vm path by vm name from xenstore-*
    @return         success:path=<val>

    N.B.
        !!deprecated!!
    """
    logger = logging.getLogger("ovs.query")
    try:
        vtech = vtech.lower().strip()
        srv = db_get_srv(host)
        sp = ServerProxy(srv['agt_url'])
        rs = getattr(sp, "%s_get_vm_path" % vtech)(vm_name)

        if rs.startswith('success'):
            path = parse_result(rs)["path"]
        else:
            raise Exception(rs)

        logger.info("get_vm_path: vm('%s') => %s" % (vm_name, path))
        return "success:path=%s" % path
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_path: failed! vm('%s') => %s" % (vm_name, errmsg))
        return "failed:%s" % errmsg

def get_live_vms(vtech="xen"):
    """
    query each host server for live vms, and update vm.db accordingly
    @return [vm1, vm2,] => list of running vms,
    """
    logger = logging.getLogger("ovs.query")
    logger.info("get_live_vms:start...")

    vtech = vtech.lower().strip()
    collect_srv_stat()
    cds = CDS("srv")
    try:
        candidates = [hostname for (hostname, srv) in cds.db_srv.items() \
                if (vtech in srv["srv_role"] and srv["status"] == "RUNNING")]
    finally: cds.close()

    d_host_vms = {}
    for hostname in candidates:
        srv = db_get_srv(hostname)
        try:
            sp = ServerProxy(srv['agt_url'])
            rs = getattr(sp, "%s_get_vms_name_vncport_macs" % vtech)()
        except: continue
        if rs.startswith('success'):
            name_vncport_macs = parse_result(rs)["vm_info"]
            if not name_vncport_macs:
                continue
            d_host_vms[hostname] = name_vncport_macs.split(",")

    #update db
    site_running_vms = []
    d_vm_in_doubt = {}
    cds = CDS("vm", "vmp", "macip")
    try:
        for hostname in d_host_vms:
            vm_infos = d_host_vms[hostname]
            for vm_info in vm_infos:
                vm_name, vncport, macs = vm_info.split("@")
                site_running_vms.append(vm_name)

                #updated vm info => ip, vncport, host_srv
                if vncport: vncport = int(vncport)
                else: vncport = -1
                d = { "host_srv": hostname, "vnc_port": vncport }
                ip = ""
                for mac in macs.split("^"):
                    if mac in cds.db_macip:
                        ip = cds.db_macip[mac]
                        if ip: break
                d["ip"] = ip

                if vm_name in cds.db_vmp:
                    vm_path = cds.db_vmp[vm_name]
                    if vm_path in cds.db_vm:
                        vm = cds.db_vm[vm_path]
                        vm.update(d)
                        cds.db_vm[vm_path] = vm
                    else: #no entry in vm.db
                        d_vm_in_doubt[vm_name] = d
                else: #no entry in vmp.db
                    d_vm_in_doubt[vm_name] = d
    finally: cds.close()
    #import vm (manage vm not created by ovs manager)
    #FIXME: possible conflict:
    # if user startup a vm outside OVS_BASE but with same name to a vm ovs managed.
    for vm_name in d_vm_in_doubt:
        vm_info = d_vm_in_doubt[vm_name]
        vm_path = join(OVS_BASE, "running_pool/%s" % vm_name)
        try:
            real_path = find_real_path(vm_path)
            register_vm(vm_path, vtech)
            db_update_vm(vm_path, vm_info)
        except: #cannot get real_path
            if vm_name in site_running_vms:
                site_running_vms.remove(vm_name)

    logger.info("get_live_vms: rs=%s" % repr(site_running_vms))
    return site_running_vms

def sync_vm_status(site_running_vms, agent_restart=True):
    """
    sync vm status based on info gethered from host servers

    @return site_running_vms
    """
    logger = logging.getLogger("ovs.query")
    logger.info("sync_vm_status: start...")

    #rebuild index vmp
    if agent_restart:
        cds = CDS("vm", "vmp")
        try:
            for vm_path in cds.db_vm:
                vm_name = basename(vm_path)
                cds.db_vmp[vm_name] = vm_path
        finally:
            cds.close()

    if not agent_restart:
        #routine sync, agent works normal
        #if not in running list, if status is RUNNING then set to DOWN
        l_in_doubt = []
        cds = CDS("vm", "vmp")
        try:
            vm_name_non_exist = []
            for vm_name in cds.db_vmp:
                vm_path = cds.db_vmp[vm_name]
                try:
                    vm = cds.db_vm[vm_path]
                    if not isinstance(vm, dict):
                        del cds.db_vm[vm_path]
                        vm_name_non_exist.append(vm_name)
                        continue
                    if vm["status"] == "RUNNING" and vm_name not in site_running_vms:
                        l_in_doubt.append(vm_path)
                except KeyError:
                    vm_name_non_exist.append(vm_name)
            for vm_name in vm_name_non_exist:
                del cds.db_vmp[vm_name]
        finally: cds.close()

        #double check if the vm in doubt is really running or down
        for vm_path in l_in_doubt:
            try:
                vm = db_get_vm(vm_path)
                srv = db_get_srv(vm['host_srv'])
                if "real_status" not in srv \
                        or ("real_status" in srv and srv["real_status"] == "ALIVE"):
                    sp = ServerProxy(srv['agt_url'])
                    rs = getattr(sp, "%s_get_vm_names" % vm['vtech'])()
                    vm_names = parse_result(rs)["vm_names"].split(",")
                    vm_name = basename(vm_path)
                    if vm_name not in vm_names: #not in running list indeed
                        logger.debug("sync_vm_status: vm('%s') RUNNING -> DOWN!" % vm_path)
                        db_update_vm(vm_path, {"status" : "DOWN"})
                    else:
                        site_running_vms.append(vm_name)
                else: #server is DEAD, vm has gone with it
                    logger.debug("sync_vm_status: vm('%s') RUNNING -> DOWN!" % vm_path)
                    db_update_vm(vm_path, {"status" : "DOWN"})

            except Exception, e:
                logger.warning("sync_vm_status: vm('%s') => %s" % (vm_path, fmt_err_msg(e)))
        #sync DOWN -> RUNNING if in site_running_vms
        cds = CDS("vm", "vmp")
        try:
            for vm_name in site_running_vms:
                try:
                    vm_path = cds.db_vmp[vm_name]
                    vm = cds.db_vm[vm_path]
                    if vm['status'] == "DOWN":
                        vm['status'] = "RUNNING"
                        cds.db_vm[vm_path] = vm
                except:
                    pass
        finally:
            cds.close()

    else:
        #called by agtctl startup
        #if in running list, set to RUNNING else DOWN
        cds = CDS("vm")
        try:
            for vm_path in cds.db_vm:
                try:
                    vm = cds.db_vm[vm_path]
                    vm_name = basename(vm_path)
                    if vm_name in site_running_vms:
                        vm["status"] = "RUNNING"
                    else:
                        vm["status"] = "DOWN"
                        vm["readlock"] = 0
                    cds.db_vm[vm_path] = vm
                except: pass
        finally:
            cds.close()

    logger.info("sync_vm_status: rs=%s" % repr(site_running_vms))
    return site_running_vms

@exposed
def get_live_vm_info(agent_restart=False):
    """
    get info of all live vms, update the database

    @return         "success:<string of host-info>" if success else "failed:<error message>"
    example of string of host-info: "host1=vm1:ip1:vnc_port1;host2=vm2:ip2:vnc_port2,vm3:ip3:vnc_port3"
    string of host-info can be an empty string; ip/vnc_port can be an empty string.
    """
    logger = logging.getLogger("ovs.query")
    #make it ready to be called by utlis/do_rpc.py
    agent_restart = bool(agent_restart)
    try:
        site_running_vms = get_live_vms()
        site_running_vms = sync_vm_status(site_running_vms, agent_restart)

        host_vm_info = {}
        cds = CDS("vm", "vmp")
        try:
            for vm_name in site_running_vms:
                try:
                    vm_path = cds.db_vmp[vm_name]
                    vm = cds.db_vm[vm_path]
                    hostname = vm['host_srv']
                    if hostname:
                        if vm["vnc_port"] < 0: vnc_port = ""
                        else: vnc_port = "%d" % vm["vnc_port"]
                        if hostname in host_vm_info:
                            host_vm_info[hostname].append("%s:%s:%s" % (vm_name, vm["ip"], vnc_port))
                        else:
                            host_vm_info[hostname] = [ "%s:%s:%s" % (vm_name, vm["ip"], vnc_port), ]
                except Exception, e:
                    logging.debug("get_live_vm_info: warning. vm('%s') => %s" % \
                            (vm_name, fmt_err_msg(e)))
        finally: cds.close()
        rs = ""
        if host_vm_info:
            rs = ";".join([ "%s=%s" % (hostname, ",".join(vm_info))
                    for hostname, vm_info in host_vm_info.items() ])
        logger.info("get_live_vm_info: success. vm info => %s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_live_vm_info: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

#----------------------------------------<> network config <>----
@exposed
def ifconfig_vm(vm_path, dev_name, inet_type,
        ip=None, netmask=None, gateway=None):
    """
    config vm network interface

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        d = { dev_name: { "inet_type": inet_type,
                          "ip": ip,
                          "netmask": netmask,
                          "gateway": gateway, } }

        vm = db_get_vm(vm_path)
        if "nics" not in vm: vm["nics"] = d
        else: vm["nics"].update(d)
        db_update_vm(vm_path, vm)

        logger.info("ifconfig_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ifconfig_vm: failed. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def hostname_vm(vm_path, hostname):
    """
    config vm hostname

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        db_update_vm(vm_path, {"hostname": hostname})
        logging.info("hostname_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("hostname_vm: success. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_ip(vm_path):
    """
    get vm primary ip from site db

    @return         "success:ip=<ip>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_get_vm(vm_path)
        logger.info("get_vm_ip: vm('%s')->%s" % (vm_path, vm['ip']))
        return "success:ip=%s" % vm["ip"]
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("get_vm_ip: error! vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_status(vm_path, status):
    """
    @param status    valid status in ('DOWN', 'RUNNING', 'ERROR', ...)
    @return "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        db_update_vm(vm_path, {"status":status})
        logger.info("set_vm_status:success. vm('%s')=>%s" % (vm_path, status))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_status:vm('%s') failed. => %s" % \
                (vm_path, errmsg(e)))
        return "failed:%s" % errmsg(e)

@exposed
def get_srv_agent_status(hostname):
    """
    @return         "success:echo=<msg>" if success else "failed:<error message>"
    """
    rs = "failed"
    logger = logging.getLogger("ovs.query")
    try:
        srv = db_get_srv(hostname)
        sp = ServerProxy(srv['agt_url'])
        rs = sp.echo('hello')
        logger.info("get_srv_agent_status: agent('%s')=> %s" % (hostname, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        try:
            db_update_srv(hostname, {"status" : "UNKNOWN"})
        except:
            pass
        logger.error("get_srv_agent_status: error. agent('%s')=>%s" % \
                (hostname, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_agent_status(vm_path):
    """
    @return         "success:echo=<msg>" if success else "failed:<error message>"
    N.B.
    !!deprecated!! we donnot install agent inside vm in this release
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_get_vm(vm_path)
        sp = ServerProxy(vm)
        rs = sp.echo('hello')
        logger.info("get_vm_agent_status: agent('%s')=> %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_agent_status: error. agent('%s')=>%s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_preferred_servers(vm_path, preferred_servers):
    """
    @param      preferred_servers       preferred servers that the VM will be running on one of them, format:"oops1.cn.oracle.com,oops2.cn.oracle.com"
    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        preferred_server_list = [ server.strip() for server in preferred_servers.strip().split(",") if server.strip() ]
        db_update_vm(vm_path, {"preferred_servers":preferred_server_list})
        logger.info("set_vm_preferred_servers:success. vm('%s')=>%s" % (vm_path, repr(preferred_server_list)))
        #TODO
        #how can remove the field: preferred_servers while preferred_server_list = []? since db_update_vm can only update ,but can not remove or pop
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_preferred_servers:vm('%s') failed. => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_preferred_servers(vm_path):
    """
    @return     "success:<string of hostnames>" if success else "failed:<error message>"
    example return value: "success:oops1.cn.oracle.com,oops2.cn.oracle.com,oops3.cn.oracle.com"
    note: possible return value "success:" indicates no preferred servers information stored for specified VM.
    """
    logger = logging.getLogger("ovs.query")
    try:
        preferred_servers = ",".join(db_get_vm(vm_path).get("preferred_servers") or []) #just incase prefered_servers not exists or someone set it to None
        logger.info("get_vm_preferred_servers:success. vm('%s')=>%s" % (vm_path, repr(preferred_servers)))
        return "success:" + preferred_servers
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_preferred_servers:vm('%s') failed. => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def truncate_db():
    """
    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("vm", "vmp", "srv")
        try:
            cds.db_vm.truncate()
            cds.db_vmp.truncate()
            cds.db_srv.truncate()
        finally:
            cds.close()
        logger.info("truncate_db: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("truncate_db: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def register_master_vip(vip):
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("master")
        try:
            mdb = cds.db_master
            mdb["master_vip"] = vip
        finally:
            cds.close()

        logger.info("register_master_vip: success. vip=%s" % vip)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("register_master_vip: failed. vip=%s => %s" % (vip, errmsg))
        return "failed: %s" % errmsg

@exposed
def get_master_vip():
    logger = logging.getLogger("ovs.query")
    try:
        cds = CDS("master")
        try:
            mdb = cds.db_master
            if not "master_vip" in mdb:
                raise Exception("master_vip has not been registered yet.")
            vip = mdb["master_vip"]
        finally:
            cds.close()

        logger.info("get_master_vip: success. vip=%s" % vip)
        return "success:vip=%s" % vip
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_master_vip: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def get_master_ip():
    logger = logging.getLogger("ovs.query")
    try:
        cds = CDS("master")
        try:
            mdb = cds.db_master
            if "master_ip" in mdb:
                ip = mdb["master_ip"]
            elif "master_hostname" in mdb and mdb["master_hostname"]:
                ip = get_host_by_name(mdb["master_hostname"])
            else:
                raise Exception("master_ip has not been registered yet.")
        finally:
            cds.close()

        logger.info("get_master_ip: success. ip=%s" % ip)
        return "success:ip=%s" % ip
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_master_ip: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def register_master(master_agt_url, member_hostname):
    """
    register master url for call back
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("master")
        try:
            ptn = r"https?://.+:.+@(.+?):[0-9]+"
            m = re.match(ptn, master_agt_url)
            if not m:
                raise Exception("agt_url must be of pattern '%s'" % ptn)
            master_hostname = m.group(1)

            mdb = cds.db_master
            #restrict that one server can only be registered in one pool
            _master_hostname = mdb.get("master_hostname")
            if _master_hostname:
                if get_host_by_name(master_hostname) != get_host_by_name(_master_hostname): #not the same server
                    if "master_agt_url" in mdb:
                        rs = ""
                        try:
                            sp = ServerProxy(mdb["master_agt_url"])
                            rs = sp.registered_server(member_hostname)
                            logger.debug("register_server: registered_server() rs=%s" % rs)
                        except Exception, e:
                            logger.debug("register_server: registered_server() exception => %s" % e)
                        if rs.startswith("success"): #still in original server
                            raise Exception("server already registered to master('%s')." % _master_hostname)

            mdb["master_hostname"] = master_hostname
            mdb["master_agt_url"] = master_agt_url
        finally:
            cds.close()
        
        #setup cluster
        try:
            from OVSXHA import ha_pull_cluster_setup
            rs = ha_pull_cluster_setup()
            if not rs.startswith("success"):
                raise Exception(rs)
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.warning("register_master: ha_pull_cluster_setup() error. => %s" % errmsg)
        
        logger.info("register_master: hostname='%s'" % master_hostname)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("register_master: error. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def unregister_master():
    """
    unregister master server information from member node.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("master")
        try:
            mdb = cds.db_master
            l = mdb.keys()
            for k in l:
                del mdb[k]
        finally:
            cds.close()

        logger.info("unregister_master:success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unregister_master: error. => %s" % errmsg)
        return "failed"

@exposed
def set_vm_policy(vm_path, policies, append=True):
    """
    @policies   priority,policy_name[;priority,policy_nam] priority in (-100, 100)
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("vm")
        try:
            if vm_path not in cds.db_vm:
                raise Exception("vm('%s') does not exist in master db." % vm_path)
            vm = cds.db_vm[vm_path]
            if "policy" not in vm: vm["policy"] = []

            l = [e.strip().split(",") for e in policies.split(";") if e.strip()]
            l = [(int(priority), policy_name) for priority, policy_name in l]

            from OVSPolicy import policy
            for priority, policy_name in l:
                if policy_name not in policy:
                    raise Exception("policy('%s') not in %s" % (policy_name, repr(policy.keys())))

            if not append:
                vm['policy'] = l
            else:
                vm["policy"] += l
            vm["policy"] = list(set(vm["policy"])) # remove duplicate
            cds.db_vm[vm_path] = vm
        finally:
            cds.close()

        logger.info("set_vm_policy: vm('%s') policy=%s append=%s=> success." % (vm_path, repr(vm['policy']),repr(append)))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_policy: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def set_pool_ha_enable(status=False):
    """
    set ha status at pool level
    vm auto start only available when pool ha is enabled
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("master")
        try:
            pool_ha_enable = status and True or False
            cds.db_master["pool_ha_enable"] = pool_ha_enable
        finally:
            cds.close()

        logger.info("set_pool_ha_enable: pool_ha_enable=%s" % repr(pool_ha_enable))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("enable_pool_ha:failed. =>%s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def set_shutdown_mode(ips, status=True):
    """
    register the server to enter shutdown mode or not
    """
    logger = logging.getLogger("ovs.operation")
    try:
        status = bool(status)
        hostname = ""
        for ip in ips.split(","):
            try:
                srv = db_get_srv(ip)
                hostname = srv.get("hostname")
                db_update_srv(hostname, {"shutdown_mode": status})
                break
            except:
                pass
        if hostname:
            logger.info("set_shutdown_mode: srv('%s')=%s" % (hostname, repr(status)))
        else:
            logger.warning("set_shutdown_mode: srv('%s')=%s" % (ips, repr(status)))
            
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("enable_pool_ha:failed. =>%s" % errmsg)
        return "failed:%s" % errmsg

