#!/usr/bin/env python
#^#
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import string
import socket
import re
import logging
from os.path import exists, join, dirname, basename
from os import remove, makedirs
from glob import glob
from OVSXUtility import parse_storage_repos
from OVSServerProxy import ServerProxy
from OVSDB import db_get_srv
from OVSConstants import HA_TOUCH_PREFIX, OVS_BASE
from OVSConfig import get_config
from OVSCommons import exposed, fmt_err_msg, parse_result, marshall_string, unmarshall_string, run_cmd, parse_shared_fs, parse_cluster_conf, get_host_by_name
from OVSCDS import CDS

#---------------------------> helper function <---------------------------------

def _get_host_sp(hosts=""):
    cds = CDS("srv")
    try:
        l = [ (hostname, ServerProxy(srv["agt_url"])) for (hostname, srv) in cds.db_srv.items() if srv.get("status") == "RUNNING"]
    finally:
        cds.close()

    d = l and dict(l) or {}
    for hostname_agent_passwd in hosts.split(","):
        try:
            hostname, agent_passwd = hostname_agent_passwd.split(":")
            config = get_config()
            protocol = config.get("security", "ssl").lower().startswith("enable") and "https" or "http"
            agt_url="%s://oracle:%s@%s:%s" % \
                (protocol, agent_passwd, hostname, config.get("agent", "port"))
            sp = ServerProxy(agt_url)
            d[hostname] = sp
        except:
            pass

    return d.items()

def _update_ocfs2_cluster_conf(hosts=""):
    _need_update = True
    cluster_conf = "/etc/ocfs2/cluster.conf"
    if not exists(cluster_conf):
        if not exists("/etc/ocfs2"):
            makedirs("/etc/ocfs2")
    else:
        l = []
        f = open(cluster_conf)
        try:
            lines = f.readlines()
            for line in lines:
                parts = map(string.strip, line.split("="))
                if parts[0] == "ip_address":
                    l.append(parts[1])
        finally:
            f.close()

        l2 = [get_host_by_name(hostname) for hostname, sp in _get_host_sp(hosts)]
        if set(l2).issubset(set(l)): #donnot need to update
            return

    rs = ha_generate_ocfs2_cluster_conf(hosts)
    if not rs.startswith("success"):
        raise Exception("ha_generate_ocfs2_cluster_conf: %s" % rs)
    f = open(cluster_conf, "w")
    try:
        f.write(rs[8:])
    finally:
        f.close()

@exposed
def ha_get_cluster_conf():
    """
    @return "success:%s" % base64 encoded /etc/ocfs2/cluster.conf
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cluster_config = ""
        f = open("/etc/ocfs2/cluster.conf")
        try:
            cluster_config = f.read()
        finally:
            f.close()

        logger.info("ha_get_cluster_conf: success. '%s'" % cluster_config)
        return "success:%s" % marshall_string(cluster_config)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_get_cluster_conf: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_get_shared_fs_conf():
    """
    @return "success:%s" % base64 encoded the /OVS line in /etc/mtab
    """
    logger = logging.getLogger("ovs.operation")
    try:
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()

        shared_fs_extra = ""
        if shared_fs_type == "ocfs2":
            output = run_cmd("debugfs.ocfs2 %s <<<stats 2>/dev/null | awk '/UUID/ {print $2;}'" % shared_fs)
            shared_fs_extra = "uuid=%s" % output.splitlines()[-1]

        #extended shared storage
        shared_repositories = ""
        f = open("/etc/ovs/repositories")
        try:
            shared_repositories = f.read()
        finally:
            f.close()

        logger.info("ha_get_shared_fs_conf: success.")
        return "success:%s@@@%s@@@%s" % \
                (marshall_string(shared_fs_line), marshall_string(shared_repositories), shared_fs_extra)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_get_shared_fs_conf: error. => %s" % errmsg)
        return "failed:%s" % errmsg

#--------------------------------> HA validation <----------------------------------
@exposed
def ha_check_oracle_vm_version(hosts=""):
    """
    @param hosts    "hostname:agent_passwd[,hostname:agent_passwd]"
    @return     success:oracle_vm_version=<val> if all servers in the pool are at the same version,
                else failed:<error message>
    """
    logger = logging.getLogger("ovs.operation")
    try:
        l = _get_host_sp(hosts)
        logger.info("ha_check_oracle_vm_version: among servers(%s)..." % repr([hostname for hostname, sp in l]))
        d = {}
        for hostname, sp in l:
            try:
                rs = sp.get_oracle_vm_version()
                rs = parse_result(rs)
                d[hostname] = rs["oracle_vm_version"]
            except Exception, e:
                errmsg = fmt_err_msg(e)
                raise Exception("get srv('%s') oracle vm version. failed! => %s" % (hostname, errmsg))

        versions = list(set(d.values()))
        if len(set(versions)) != 1:
            raise Exception("servers are not in the oracle vm version! %s" % repr(d))

        version = versions[0]
        v = [ int(x) for x in re.split(r"[.-]", version)]
        if v <= [2, 1, 1]:
            raise Exception("Oracle VM must be 2.1.2 or upward.")

        rs = "success:oracle_vm_version=%s" % versions
        logger.info("ha_check_oracle_vm_version: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_oracle_vm_version: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_check_agent_version(hosts=""):
    """
    @param hosts "hostname:agent_passwd[,hostname:agent_passwd]"
    @return     "success:agent_version=<val>" if all servers in the pool are at the same version
                else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        l = _get_host_sp(hosts)
        logger.info("ha_check_agent_version: among servers(%s)..." % repr([hostname for hostname, sp in l]))
        d = {}
        for hostname, sp in l:
            try:
                rs = sp.get_agent_version()
                rs = parse_result(rs)
                d[hostname] = rs["agent_version"]
            except Exception, e:
                errmsg = fmt_err_msg(e)
                raise Exception("get srv('%s') agent_version version. failed! => %s" % (hostname, errmsg))

        versions = list(set(d.values()))
        if len(set(versions)) != 1:
            raise Exception("servers are not in the agent version! %s" % repr(d))

        version = versions[0]
        v = [ int(x) for x in re.split(r"[.-]", version)]
        if len(v) == 3:
            v.insert(2, 0)
        if v <= [2, 2, 0, 27]: #2.2-27 -> oracle vm 2.1.1
            raise Exception("Oracle VM must be 2.1.2 or upward.")

        rs = "success:agent_version=%s" % version
        logger.info("ha_check_agent_version: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_agent_version: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_precheck_storage_mount():
    """
    @return "failed" iff /OVS is already mounted  and not in shared file systems [nfs, ocfs2].
    """
    logger = logging.getLogger("ovs.operation")
    try:
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
        if not shared_fs_type:
            raise Exception("/OVS must be mounted.")
        if shared_fs_type and shared_fs_type not in ("nfs", "ocfs2"):
            raise Exception("/OVS is already mounted, but not in ['nfs', 'ocfs2'].")
        #bugz#6757, local formated ocfs2 filesystem should not be supported.
        if shared_fs_type == "ocfs2":
            cmd = "debugfs.ocfs2 %s <<< stats 2>/dev/null | grep -q 'Feature Incompat:.*8 Local'" % shared_fs
            local_mount = False 
            try:
                run_cmd(cmd)
                local_mount = True
            except:
                pass
            if local_mount:
                raise Exception("/OVS is already mounted to local formated ocfs2, which is not supported.")

        logger.info("ha_precheck_storage_mount: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.info("ha_precheck_storage_mount: failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_check_shared_storage(hosts=""):
    """
    @param hosts "hostname:agent_passwd[,hostname:agent_passwd]"
    @return "success" if all servers in the pool are connected to the same storage
            i.e. have same mounted file system at /OVS & /OVS/<uuid>
            else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        #create temp file
        from random import seed, randint
        seed()
        filename = "%05d" % randint(0, 99999)
        d_uuid = parse_storage_repos()
        d_uuid[""] = ""
        for uuid in d_uuid:
            v_touch_file = join(join(OVS_BASE, uuid), filename)
            touch_file = join(join(OVS_BASE, uuid), "%s%s" % (HA_TOUCH_PREFIX, filename))
            open(touch_file, "w").close()
            try:
                l = _get_host_sp(hosts)
                logger.info("ha_check_shared_storage: among servers(%s)..." % repr([hostname for hostname, sp in l]))
                for hostname, sp in l:
                    try:
                        rs = sp.ha_check_exist(v_touch_file)
                        if not rs.startswith("success"):
                            raise Exception(rs)
                    except Exception, e:
                        errmsg = fmt_err_msg(e)
                        raise Exception("srv('%s'). failed! => %s" % (hostname, errmsg))
            finally:
                #clean out
                remove(touch_file)

        logger.info("ha_check_shared_storage: success. /OVS/%s" % repr(d_uuid.keys()))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_shared_storage: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_check_cluster_status(hosts=""):
    """
    check cluster status by o2cb utils
    """
    logger = logging.getLogger("ovs.operation")
    try:

        l = _get_host_sp(hosts)
        logger.info("ha_check_cluster_status: among servers(%s)..." % repr([hostname for hostname, sp in l]))
        d = {}
        for hostname, sp in l:
            try:
                rs = sp.ha_check_cluster_heartbeat()
                if not rs.startswith("success"):
                    raise Exception(rs)
            except Exception, e:
                errmsg = fmt_err_msg(e)
                raise Exception("srv('%s'). failed! => %s" % (hostname, errmsg))

        logger.info("ha_check_cluster_status: success")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_cluster_status: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def ha_check_cluster_prerequisite(hosts=""):
    """
    @param hosts hostname:agent_passwd[,hostname:agent_passwd]
    will check:
        *) oracle vm version
        *) agent version
        *) shared storage status
    """
    logger = logging.getLogger("ovs.operation")
    try:
        rs = []
        rs.append(("ha_check_oracle_vm_version", ha_check_oracle_vm_version(hosts)))
        rs.append(("ha_check_agent_version", ha_check_agent_version(hosts)))
        failed = [(a, b) for (a, b) in rs if not b.startswith("success")]
        if failed:
            msg = "\n".join([" ".join(t) for t in failed])
            raise Exception(msg)

        rs.append(("ha_precheck_storage_mount", ha_precheck_storage_mount()))
        failed = [(a, b) for (a, b) in rs if not b.startswith("success")]
        if failed:
            msg = "\n".join([" ".join(t) for t in failed])
            raise Exception(msg)

        logger.info("ha_check_cluster_prerequisite: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_cluster_prerequisite: error. =>%s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_check_cluster_infrastructure(hosts=""):
    """
    will validate:
        *) oracle vm version
        *) agent version
        *) shared storage status
        *) ocfs2 cluster status
    """
    logger = logging.getLogger("ovs.operation")
    try:
        rs = ha_check_cluster_prerequisite(hosts)
        if not rs.startswith("success"):
            raise Exception("ha_check_cluster_prerequisite: => %s" % rs)

        rs = []
        rs.append(("ha_check_shared_storage", ha_check_shared_storage(hosts)))
        rs.append(("ha_check_cluster_status", ha_check_cluster_status(hosts)))
        failed = [(a, b) for (a, b) in rs if not b.startswith("success")]
        if failed:
            msg = "\n".join([" ".join(t) for t in failed])
            raise Exception(msg)

        logger.info("ha_check_cluster_infrastructure: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_cluster_infrastructure: error. =>%s" % errmsg)
        return "failed:%s" % errmsg

#--------------------------------> HA setup <----------------------------------
@exposed
def ha_sync_shared_storage_mount(hosts=""):
    """
    /OVS    -> mount
    /etc/ovs/repositories -> /etc/init.d/ovsrepositories restart
    """
    logger =  logging.getLogger("ovs.operation")
    try:
        #base shared storage
        #awk '$2=="/OVS" {print $0; };' /etc/mtab
        #/dev/sdd1 /OVS ocfs2 rw 0 0
        #ostest105.cn.oracle.com:/OVS /OVS nfs rw,sync,addr=10.182.125.105 0 0
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
        shared_fs_extra = ""
        if shared_fs_type == "ocfs2":
            output = run_cmd("debugfs.ocfs2 %s <<<stats 2>/dev/null | awk '/UUID/ {print $2;}'" % shared_fs)
            shared_fs_extra = "uuid=%s" % output.splitlines()[-1]

        #extended shared storage
        shared_repositories = ""
        f = open("/etc/ovs/repositories")
        try:
            shared_repositories = f.read()
        finally:
            f.close()

        l = _get_host_sp(hosts)
        logger.info("ha_sync_shared_storage_mount: among servers(%s)..." % repr([hostname for hostname, sp in l]))
        for hostname, sp in l:
            try:
                rs = sp.ha_mount_shared_storage(marshall_string(shared_fs_line), marshall_string(shared_repositories), shared_fs_extra)
                if not rs.startswith("success"):
                    raise Exception(rs)
            except Exception, e:
                errmsg = fmt_err_msg(e)
                raise Exception("srv('%s'). failed! => %s" % (hostname, errmsg))

        logger.info("ha_sync_shared_storage_mount: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_sync_shared_storage_mount: error. =>%s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_start_cluster_service(hosts=""):
    """
    for server in hosts:
        if OVS is ocfs2: load ocfs2 configuration, start o2cb service
        if OVS is nfs: load ocfs2 configuration, start o2cb service, setup dmnfs
    """
    logger = logging.getLogger("ovs.operation")
    try:

        cluster_config = ""
        f = open("/etc/ocfs2/cluster.conf")
        try:
            cluster_config = f.read()
        finally:
            f.close()

        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
        l = _get_host_sp(hosts)
        for hostname, sp in l:
            try:
                rs = sp.ha_setup_ocfs2_cluster(marshall_string(cluster_config), shared_fs_type)
                if not rs.startswith("success"):
                    raise Exception(rs)
            except Exception, e:
                errmsg = fmt_err_msg(e)
                raise Exception("srv('%s'). failed! => %s" % (hostname, errmsg))

        logger.info("ha_start_cluster_service: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_start_cluster_service: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_start_agent_guard_all(hosts=""):
    """
    @xref ha_start_agent_guard()
    """
    logger = logging.getLogger("ovs.operation")
    try:

        l = _get_host_sp(hosts)
        for hostname, sp in l:
            try:
                rs = sp.ha_start_agent_guard()
                if not rs.startswith("success"):
                    raise Exception(rs)
            except Exception, e:
                errmsg = fmt_err_msg(e)
                raise Exception("srv('%s'). failed! => %s" % (hostname, errmsg))

        logger.info("ha_start_agent_guard_all: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_start_agent_guard_all: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_setup_cluster_infrastructure(hosts=""):
    """
    setup the cluster infrastructure (enable DLM):
    if OVS is ocfs2:
        load ocfs2 cluster config
        start o2cb service
        sync mount point on all nodes(start heartbeating)

    if OVS is nfs:
        sync mount point on all nodes
        load ocfs2 cluster config
        start o2cb service
        setup dmnfs device (start heartbeating)
    """
    logger = logging.getLogger("ovs.operation")
    try:
        rs = ha_check_cluster_prerequisite(hosts)
        if not rs.startswith("success"):
            raise Exception("ha_check_cluster_prerequisite: => %s" % rs)

        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
        _update_ocfs2_cluster_conf(hosts)

        logger.info("ha_setup_cluster_infrastructure: ...")
        if shared_fs_type == "ocfs2":
            rs = ha_start_cluster_service(hosts)
            if not rs.startswith("success"):
                raise Exception("ha_start_cluster_service: %s" % rs)

            rs = ha_sync_shared_storage_mount(hosts)
            if not rs.startswith("success"):
                raise Exception("ha_sync_shared_storage_mount: %s" % rs)

            rs = ha_start_agent_guard_all(hosts)
            if not rs.startswith("success"):
                raise Exception("ha_start_agent_guard_all: %s" % rs)

        elif shared_fs_type == "nfs":
            rs = ha_sync_shared_storage_mount(hosts)
            if not rs.startswith("success"):
                raise Exception("ha_sync_shared_storage_mount: %s" % rs)

            rs = ha_start_cluster_service(hosts)
            if not rs.startswith("success"):
                raise Exception("ha_start_cluster_service: %s" % rs)

            rs = ha_start_agent_guard_all(hosts)
            if not rs.startswith("success"):
                raise Exception("ha_start_agent_guard_all: %s" % rs)
        else:
            raise Exception("only [ocf2, nfs] shared filesystem is supported.")

        logger.info("ha_setup_cluster_infrastructure: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_setup_cluster_infrastructure: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_generate_ocfs2_cluster_conf(hosts=""):
    """
    @return success:<content of generated /etc/ocfs2/cluster.conf>
    """
    def _read(filename):
        f = open(filename, "r")
        r = f.read()
        f.close()
        return r

    logger = logging.getLogger("ovs.operation")
    try:
        l = _get_host_sp(hosts)
        hostnames = [hostname for hostname, sp in l]
        
        #always include my own hostname
        _hostname = socket.gethostname()
        if _hostname.split(".")[0] not in [h.split(".")[0] for h in hostnames]:
            hostnames.append(_hostname)
        
        cluster_config_file = "/etc/ocfs2/cluster.conf"
        if exists(cluster_config_file):
            d = parse_cluster_conf()
            cluster_name = d["cluster"]["name"]
            nodes = dict([ (node["ip_address"], node) for node in d["nodes"] ])
        else:
            d = {"nodes":[], "cluster":{}}
            cluster_name = "ocfs2" #default
            nodes = {}
        
        try:
            #consider what already in the configfs
            ptn = "/sys/kernel/config/cluster/%s/node/*" % cluster_name
            l = glob(ptn)
            for node_dir in l:
                node = {}
                node["ip_address"] = _read(join(node_dir, "ipv4_address"))
                node["ip_port"] = _read(join(node_dir, "ipv4_port"))
                node["number"] = _read(join(node_dir, "num"))
                node["name"] = basename(node_dir)
                node["cluster"] = cluster_name
                nodes[node["ip_address"]] = node
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.debug("ha_generate_ocfs2_cluster_conf: get current used node number failed. => %s" % errmsg)

        numbers = [ int(node["number"]) for node in d["nodes"] ]
        for hostname in hostnames:
            ip = get_host_by_name(hostname)
            if ip not in nodes:
                d["ip_port"] = "7777"
                d["ip_address"] = ip
                #calculate the free slot
                _number = 0
                while True:
                    if _number in numbers:
                        _number += 1
                    else:
                        numbers.append(_number)
                        break
                d["number"] = str(_number)
                #name must be the real hostname
                try:
                    srv = db_get_srv(hostname)
                    if "hostname" in srv:
                        d["name"] = srv["hostname"]
                    else:
                        d["name"] = hostname
                except:
                    d["name"] = hostname
                d["cluster"] = cluster_name
                nodes[ip] = d
            nodes[ip]["in"] = True 
        
        #sort by node number
        l = [ (int(node["number"]), node) for node in nodes.values()
                    if node.get("in") #filter server not in the pool.
            ]
        l.sort()
        lines = []
        for (_number, node) in l:
            lines.append("node:\n")
            lines.append("\tip_port = %s\n" % node["ip_port"])
            lines.append("\tip_address = %s\n" % node["ip_address"])
            lines.append("\tnumber = %s\n" % node["number"])
            lines.append("\tname = %s\n" % node["name"])
            lines.append("\tcluster = %s\n" % node["cluster"])
            lines.append("\n")
        lines.append("cluster:\n")
        lines.append("\tnode_count = %d\n" % len(l))
        lines.append("\tname = %s\n" % cluster_name)
        lines.append("\n")

        clusterconf = "".join(lines)
        logger.info("ha_generate_ocfs2_cluster_conf:\n%s" % clusterconf)
        return "success:%s" % clusterconf
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_generate_ocfs2_cluster_conf: error. => %s" % errmsg)
        return "failed:%s" % errmsg

def ha_sync_ocfs2_cluster_conf():
    """
    sync the change to /etc/ocfs2/cluster.conf when add/remove server
    """
    logger = logging.getLogger("ovs.operation")
    try:
        rs = ha_generate_ocfs2_cluster_conf()
        if not rs.startswith("success"):
            raise Exception(rs)
        
        if not exists("/etc/ocfs2"):
            makedirs("/etc/ocfs2")
        f = open("/etc/ocfs2/cluster.conf", "w")
        try:
            f.write(rs[8:]) #strip 'success:'
        finally:
            f.close()

        logger.info("ha_sync_ocfs2_cluster_conf: success")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_sync_ocfs2_cluster_conf: error. %s" % errmsg)
        return "failed"
