#!/usr/bin/env python
#^#
# OVSSiteMonitor.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import socket
import logging
from OVSServerProxy import ServerProxy
from OVSDB import db_get_vm, db_update_vm
from OVSDB import db_get_srv, db_update_srv, db_get_real_vm_path
from OVSCommons import exposed, fmt_err_msg, parse_result, chk_vm_path
from OVSCDS import CDS
from OVSServerSelection import collect_srv_stat, get_vm_host

@exposed
def get_hs_perf(hostname):
    """
    get host performance

    @return         "success:<string of host-info>" if success else "failed:<error message>"
    example of string of host-info: "cpu_idle=<cpu idle>;mem_free=<free mem>;hd_avail=<harddisk avail>"
    """
    logger = logging.getLogger("ovs.performance")
    try:
        srv = db_get_srv(hostname)
        sp = ServerProxy(srv['agt_url'])
        rs = sp.sys_perf_info()
        logger.info("get_hs_perf:srv('%s') => %s" % (hostname, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("get_hs_perf: error. srv('%s') => %s" % \
                (hostname, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_servers_perf(hostnames=''):
    """
    within this site, get servers performance data,
    hostnames is a list of hostname/ip delimited by ','

    @return         "success:<string of srvs-info>" if success else "failed:<error message>"
    example of string of srvs-info: "srv1=cpu_idle1:mem_free1:hd_avail1,srv2=cpu_idle2:mem_free2:hd_avail2"
    """
    logger = logging.getLogger("ovs.performance")
    try:
        if hostnames.strip():
            hostnames = hostnames.split(",")
        else:
            cds = CDS("srv")
            try:
                hostnames = cds.db_srv.keys()
            finally:
                cds.close()

        results = []
        for hostname in hostnames:
            try:
                srv = db_get_srv(hostname)
                sp = ServerProxy(srv['agt_url'])
                rs = sp.sys_perf_info()
                if rs.startswith("success"):
                    d = parse_result(rs)
                    results.append("%s=%s:%s:%s" % \
                            (hostname, d["cpu_idle"], d["mem_free"], d["hd_avail"]))
                    db_update_srv(hostname, {"status" : "RUNNING"})
                else:
                    logger.error("get_server_perfs: srv('%s') => %s" % (hostname, rs))
                    results.append("%s=" % (hostname, ))
                    db_update_srv(hostname, {"status" : "UNKNOWN"})
            except Exception, e:
                logger.error("get_servers_perf: srv('%s') => %s" % \
                        (hostname, fmt_err_msg(e)))
                db_update_srv(hostname, {"status" : "UNKNOWN"})

        rs = "success:%s" % ( ";".join(results), )
        logger.info("get_servers_perf: rs=%s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_servers_perf: hostnames=%s => %s" % \
                (hostnames, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_perf(vm_path):
    """
    get vm performance, please make sure the agent in the vm is running

    @return         "success:<string of vm-info>" if success else "failed:<error message>"
    example of string of vm-info: "cpu_idle=<cpu idle>;mem_free=<free mem>;hd_avail=<harddisk avail>"

    N.B.
    !!deprecated!! agent is not installed inside vm in this release
    """
    logger = logging.getLogger("ovs.performance")
    try:
        vm = db_get_vm(vm_path)
        sp = ServerProxy(vm)
        rs = sp.sys_perf_info()
        logger.info("get_vm_perf:vm('%s') => %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("get_vm_perf: error. vm('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vms_perf(vtech='xen'):
    """
    within this site, get vms performance data,

    @return         "success:<vm-info>" if success else "failed:<error message>"
                    e.g. vm-info "vm1=cpu_idle1:mem_free1:hd_avail1;vm2=cpu_idle2:mem_free2:hd_avail2"
    """
    logger = logging.getLogger("ovs.performance")
    try:
        cds = CDS("srv")
        hostnames = [ hostname for hostname, srv in cds.db_srv.items()
                if vtech in srv["srv_role"]]
        cds.close()

        results = []
        for hostname in hostnames:
            try:
                srv = db_get_srv(hostname)
                sp = ServerProxy(srv['agt_url'])
                rs = getattr(sp, "%s_top_info" % vtech)()
                if rs.startswith("success"):
                    vm_perfs = parse_result(rs)
                    if vm_perfs:
                        vm_perfs = ["%s=%s" % (vm_name, perf)
                                for vm_name, perf in vm_perfs.items()]
                        results.extend(vm_perfs)
                else:
                    logger.error("get_vms_perf: srv('%s') => %s" % (hostname, rs))
            except Exception, e:
                logger.warning("get_vms_perf: srv('%s') => %s" % \
                        (hostname, fmt_err_msg(e)))

        rs = "success:%s" %  ";".join(results)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("get_vms_perf: => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def get_host_info(hostname):
    """
    get architecture, cpu number, cpu speed (bogomips), memory size (in MB),

    @return     <refer to sys_os_info()>
    """
    logger = logging.getLogger("ovs.query")
    try:
        srv = db_get_srv(hostname)
        sp = ServerProxy(srv['agt_url'])
        rs = sp.sys_os_info()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs_caps = sp.xen_get_caps()
        if not rs_caps.startswith("success"):
            raise Exception(rs)
        rs = "%s;caps=%s" % (rs, parse_result(rs_caps)["caps"])
        logger.info("get_host_info: srv('%s') => %s" % (hostname, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("get_host_info: error. srv('%s') => %s" % \
                (hostname, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_host_arch(hostname):
    """
    get cpu architecture

    @return     refer to sys_cpu_info()
    N.B.: to check if the server is suitable to act as the vm migrating target.
    """
    logger = logging.getLogger("ovs.query")
    try:
        srv = db_get_srv(hostname)
        sp = ServerProxy(srv['agt_url'])
        rs = sp.sys_cpu_info()

        logger.info("get_host_arch: srv('%s') => %s" % (hostname, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("get_host_arch: error. srv('%s') => %s" % (hostname, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_xen_caps(hostname):
    """
    get xen caps

    @return     "success:caps=<caps>" if success else "failed:<error message>"
    N.B.: use it to check if the server support hvm and detect the bit mode
    """
    logger = logging.getLogger("ovs.query")
    try:
        srv = db_get_srv(hostname)
        sp = ServerProxy(srv['agt_url'])
        rs = sp.xen_get_caps()

        logger.info("get_xen_caps: srv('%s') => %s" % (hostname, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logging.error("get_xen_caps: failed. srv('%s') => %s" % (hostname, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_servers_vcaps():
    """
    get servers vt caps, where server's xend has hardware vitualization support

    @return "success:hvm=[server_name[,server_name]];xen=[server_name[,server_name]]"
    """
    logger = logging.getLogger("ovs.performance")
    try:
        hvm_srvs = []
        xen_srvs = []
        cds = CDS("srv")
        try:
            for (hostname, srv) in cds.db_srv.items():
                if "vcaps" in srv:
                    if "hvm" in srv["vcaps"]:
                        hvm_srvs.append(hostname)
                    elif "xen" in srv["vcaps"]:
                        xen_srvs.append(hostname)
        finally:
            cds.close()

        hvm = "hvm=%s" % ",".join(hvm_srvs)
        xen = "xen=%s" % ",".join(xen_srvs)
        rs = "%s;%s" % (hvm, xen)
        logger.info("get_servers_vcaps: success rs=>%s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_servers_vcaps: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def get_free_memory(hostnames=""):
    """
    @param hostnames    [<hostname>[,<hostname>]] if not hostnames.strip(), hostnames will be all the host in the pool
    @return     "success:memory=<memory>;hostname=<hostname>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        candidates = [hostname.strip() for hostname in hostnames.split(",") if hostname.strip()]
        if not candidates:
            cds = CDS("srv")
            try:
                candidates = cds.db_srv.keys()
            finally:
                cds.close()
        if not candidates: #no server registered in server pool
            raise Exception("no candidates to get memory!")

        collect_srv_stat(candidates)
        cds = CDS("srv")
        try:
            mem_max, tgt_host = 0, None
            for hostname in candidates:
                if hostname in cds.db_srv:
                    srv = cds.db_srv[hostname]
                    if srv["mem_free"] > mem_max:
                        mem_max, tgt_host = srv["mem_free"], hostname
        finally:
            cds.close()
        if not tgt_host: #rare case, no one got free memory
            raise Exception("no free memory on any of the servers.")

        rs = "success:memory=%d;hostname=%s" % (mem_max, tgt_host)
        logger.info("get_free_memory:hostnames=[%s] => %s" % (hostnames,rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_free_memory: failed. hostnames=[%s]=> %s" % (hostnames, errmsg))
        return "failed: %s" % errmsg

@exposed
def get_vm_uptime(vm_path):
    """Get uptime for a specific vm

    @param vm_path: to specify the VM
    @type vm_path:  string
    @return:        "success:uptime=%s"
                    "failed:<error message>" on failure
    @rtype:         string
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        (hostname, sp) = get_vm_host(vm_path)
        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_uptime" % vm["vtech"])(real_vm_path)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("get_vm_uptime: vm('%s'): rs=%s" % (vm_path, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vbd_qos: failed. vm('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg
