#!/usr/bin/env python
#^#
# OVSSiteVM.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#
import zlib
import re
import logging
from os.path import join, isdir
from OVSServerSelection import select_tgt_host, select_one_server, get_vm_host
from OVSServerProxy import ServerProxy
from OVSDB import db_get_vm, db_update_vm, db_del_vm, db_chkset_vm_status, db_get_real_vm_path
from OVSDB import db_get_srv, db_update_srv, db_del_srv
from OVSCommons import exposed, parse_result, fmt_err_msg, chk_vm_path
from OVSCDS import CDS
from OVSXUtility import utl_get_free_space
from OVSXXenVMConfig import xen_get_memory
from OVSException import *

@exposed
def start_vm(vm_path, dedicated_server=None):
    """
    start a vm
    dedicated_server => circumvent the balance mechanism

    @return         "success:ip=<ip>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("start_vm: vm('%s') dedicated_server=%s start..." % (vm_path, repr(dedicated_server)))
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("DOWN", "ERROR"), "PROCESSING")
        try:
            if dedicated_server:
                host_srv = dedicated_server
                srv = db_get_srv(host_srv)
                sp = ServerProxy(srv['agt_url'])
            else:
                host_srv = ""
                (host_srv, sp) = select_tgt_host(vm_path)
            logger.debug("start_vm: vm('%s') on srv('%s') ..." % (vm_path, host_srv))

            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            #start vm
            paras = (real_vm_path,)
            rs = getattr(sp, "%s_start_vm" % vm["vtech"])(*paras)
            logger.info("start_vm: vm('%s') on srv('%s') => %s" % (vm_path, host_srv, rs))
            #check result & update db
            if rs.startswith("success"):
                db_update_vm(vm_path, {"host_srv": host_srv, "status": "RUNNING", "manual_shutdown": False})
            else:
                raise Exception("return=>%s" % rs)

        except OVSException, e:
            if e.is_kind_of(ERR_SERVER_SELECTION):
                db_chkset_vm_status(vm_path, "", "DOWN")
            else:
                db_chkset_vm_status(vm_path, "", "ERROR")
            raise e
        except Exception, e:
            try:
                db_chkset_vm_status(vm_path, "", "ERROR")
                db_update_vm(vm_path, {"host_srv": host_srv})
            except:
                pass
            raise e

        logger.info("start_vm: success. vm('%s') ip='%s'" % (vm_path, vm["ip"]))
        return "success:ip=%s" % vm["ip"]
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("start_vm: failed! vm('%s') %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def stop_vm(vm_path):
    """
    stop a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("stop_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        #let DOWN in status_before, since it may out sync with ovs manager.
        #if it already DOWN, let it success anyway.
        db_chkset_vm_status(vm_path, ("DOWN", "RUNNING", "SUSPEND", "ERROR"), "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            rs = getattr(sp, "%s_stop_vm"% vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("stop_vm: vm('%s') on srv('%s') =>%s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"host_srv":"", "status": "DOWN", "manual_shutdown": True})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("stop_vm: vm('%s') failed=> %s" % (vm_path, errmsg))
        return "failed:%s" % (errmsg, )

@exposed
def reset_vm(vm_path):
    """
    reset a vm (reboot)

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("reset_vm: vm('%s')" % vm_path)
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("ERROR", "RUNNING"), "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            rs = getattr(sp, "%s_reboot_vm" % vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("reset_vm: vm('%s') on srv('%s') => %s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "RUNNING"})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("reset_vm: failed. vm('%s') %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def pause_vm(vm_path):
    """
    pause a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("pause_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("ERROR", "RUNNING"), "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            rs = getattr(sp, "%s_pause_vm"% vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("pause_vm: vm('%s') on srv('%s') => %s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "SUSPEND"})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("pause_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def unpause_vm(vm_path):
    """
    unpause a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("unpause_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("ERROR", "SUSPEND"), "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            rs = getattr(sp, "%s_unpause_vm"% vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("unpause_vm: vm('%s') on srv('%s') => %s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "RUNNING"})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unpause_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def save_vm(vm_path, save_file=None):
    """
    save a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("save_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)
        db_chkset_vm_status(vm_path, "RUNNING", "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)

            # check if there is enough space
            rs = utl_get_free_space(real_vm_path)
            if not rs.startswith("success"):
                raise Exception(rs)
            free_space = int(parse_result(rs)["free_space"])
            rs = xen_get_memory(real_vm_path)
            if not rs.startswith("success"):
                raise Exception(rs)
            memory = int(parse_result(rs)["memory"])
            if free_space < memory:
                raise Exception("There is not enough space.")

            if not save_file:
                save_file = join(real_vm_path, "vm.save")
            rs = getattr(sp, "%s_save_vm"% vm["vtech"])(real_vm_path, save_file)

            #check result & update db
            logger.info("save_vm: vm('%s') %s" % (vm_path, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "DOWN", "manual_shutdown": True})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("save_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def restore_vm(vm_path, dedicated_server=None, save_file=None):
    """
    restore a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("restore_vm: vm('%s') dedicated_server=%s start..." % (vm_path, repr(dedicated_server)))
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, "DOWN", "PROCESSING")
        try:
            if dedicated_server:
                host_srv = dedicated_server
                srv = db_get_srv(host_srv)
                sp = ServerProxy(srv['agt_url'])
            else:
                (host_srv, sp) = select_tgt_host(vm_path)

            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            if not save_file:
                save_file = join(real_vm_path, "vm.save")
            else:
                save_file = join(real_vm_path, save_file)
            rs = getattr(sp, "%s_restore_vm"% vm["vtech"])(save_file)

            #check result & update db
            logger.info("restore_vm: vm('%s') from '%s' on srv('%s') => %s" % \
                    (vm_path, save_file, host_srv ,rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "RUNNING", "host_srv": host_srv, "manual_shutdown": False})
            else:
                raise Exception(rs)

        except OVSException, e:
            if e.is_kind_of(ERR_SERVER_SELECTION):
                db_chkset_vm_status(vm_path, "", "DOWN")
            else:
                db_chkset_vm_status(vm_path, "", "ERROR")
            raise e
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("restore_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def select_migration_target(vm_path):
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("select_migration_target: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        vm = db_get_vm(vm_path)
        srv = db_get_srv(vm['host_srv'])
        sp = ServerProxy(srv['agt_url'])

        (tgt_host, tgt_sp) = select_tgt_host(vm_path,
                            any = False,
                            excludes = (vm['host_srv'], ),
                            with_attr = {"arch": srv["arch"]})

        logger.info("select_migration_target: vm('%s') from '%s'->'%s'" % \
                    (vm_path, vm["host_srv"], tgt_host))
        return "success:target_host=%s" % tgt_host
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("select_migration_target: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg


@exposed
def migrate_vm(vm_path, tgt_host=None, is_live=True):
    """
    migrate vm

    @return         "success" if success else "failed:<error message>
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("migrate_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        #a vm with status "ERROR" is also allowed to be migrated
        db_chkset_vm_status(vm_path, ("RUNNING", "ERROR"), "PROCESSING")
        try:
            vm = db_get_vm(vm_path)
            srv = db_get_srv(vm['host_srv'])
            sp = ServerProxy(srv['agt_url'])

            if not tgt_host:
                (tgt_host, tgt_sp) = select_tgt_host(vm_path,
                                    any = False,
                                    excludes = (vm['host_srv'], ),
                                    with_attr = {"arch": srv["arch"]})
            else:
                tgt_srv = db_get_srv(tgt_host)
                tgt_sp = ServerProxy(tgt_srv['agt_url'])

            #make sure the tgt host share the same view with current host server,
            rs = sp.utl_find_real_path(vm_path)
            if not rs.startswith("success"):
                raise Exception("host_srv=%s =>%s" % (vm['host_srv'], rs))
            rs = parse_result(rs)
            real_vm_path = rs["real_path"]

            rs = tgt_sp.utl_find_real_path(vm_path)
            if not rs.startswith("success"):
                raise Exception("tgt_host=%s =>%s" % (tgt_host, rs))
            rs = parse_result(rs)
            tgt_real_vm_path = rs["real_path"]

            if real_vm_path != tgt_real_vm_path:
                raise Exception("%s' != '%s'" % (real_vm_path, tgt_real_vm_path))

            #migrate finally
            rs = getattr(sp, "%s_migrate_vm" % vm["vtech"])(real_vm_path, tgt_host, is_live)

            #check result & update db
            logger.info("migrate_vm: vm('%s') from '%s'->'%s' => %s" % \
                    (vm_path, vm["host_srv"], tgt_host ,rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"host_srv": tgt_host, "status": "RUNNING"})
            else:
                raise Exception(rs)
        except OVSException, e:
            if e.is_kind_of(ERR_SERVER_SELECTION):
                db_chkset_vm_status(vm_path, "", "DOWN")
            else:
                db_chkset_vm_status(vm_path, "", "ERROR")
            raise e
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("migrate_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

