#!/usr/bin/env python
#^#
# OVSSiteVMConfig.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#
import logging
import re
import string
from os.path import join, basename, abspath
from OVSSiteUtility import get_vm_size, get_vdisk_size
from OVSServerSelection import select_one_server, get_utility_server, get_vm_host
from OVSDB import db_get_vm, db_get_real_vm_path
from OVSCommons import exposed, parse_result, fmt_err_msg, chk_vm_path, randomMAC
from OVSXUtility import find_real_path
from OVSXXenVMConfig import xen_unmarshall_cfg, xen_marshall_cfg, xen_get_vm_type, xen_set_vm_mem_vcpus_vncpasswd
from OVSXXenVMConfig import xen_optimize_os_profile, xen_get_avail_disk_slot, marshall_vif_cfg, unmarshall_vif_cfg
from OVSXXenVMConfig import xen_config_boot_sequence

@exposed
def get_vm_avail_disk_slot(vm_path):
    """
    get available vm disk slot -- there is limitation for disks can be attached to a vm
    @xref xen_get_avail_disk_slot()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        real_vm_path = db_get_real_vm_path(vm_path)
        rs = xen_get_avail_disk_slot(real_vm_path)

        if not rs.startswith("success"): raise Exception(rs)
        logger.info("get_vm_avail_disk_slot: vm('%s') => %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_avail_disk_slot: failed. vm('%s') => %s" % (vm_path, errmsg))
        return "failed: %s" % errmsg

@exposed
def add_vm_disk(vm_path, disk_path, sharable=False, fe_dev="auto", mode="w", scope="cfg", fe_dev_prefix="", position=-1):
    """
    attach a disk to a vm
    @xref xen_add_disk()
    @param position , int, specifies the position in the disk list, default=-1 means donot care the position and disk will go to the end.
    @return when succeed returns "success:fe_name=<front end name>", when fail returns "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        chk_vm_path(disk_path)

        logger.debug("add_vm_disk: vm('%s'), disk_path=%s, sharable=%s, fe_dev=%s, mode=%s, scope=%s" % \
            (vm_path, disk_path, repr(sharable), fe_dev, mode, scope))
        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        real_disk_path = find_real_path(disk_path)
        logger.debug("add_vm_disk:disk_path=%s, real_disk_path=%s" % (disk_path, real_disk_path))

        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_add_disk" % vm["vtech"])(real_vm_path, real_disk_path, sharable, fe_dev, mode, scope, fe_dev_prefix, position)
        else:
            rs =  getattr(sp, "%s_add_disk" % vm["vtech"])(real_vm_path, real_disk_path, sharable)

        logger.info("add_vm_disk: vm('%s') disk('%s'): scope=%s rs=%s" % \
                (vm_path, disk_path, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("add_vm_disk: failed. vm('%s') disk('%s') => %s" % \
                (vm_path, disk_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def remove_vm_disk(vm_path, disk_path, scope="cfg"):
    """
    detach a disk from a vm

    @xref   xen_remove_disk()
    @return when succeed returns "success:position=<disk original position>" failed returns "failed:<error message>", when returns "success:-1" means:
    1. disk had already been removed.
    2. scope='mem', in which case there is no position information.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        chk_vm_path(disk_path)

        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        try:
            real_disk_path = find_real_path(disk_path)
        except Exception, e:
            #if already deleted mannually should return success
            #rely on the start xen_correct_cfg() to remove the invalid entry
            logger.warning("remove_vm_disk: vm('%s') disk('%s') => %s" % \
                    (vm_path, disk_path, fmt_err_msg(e)))
            return "success:position=-1"

        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_remove_disk" % vm["vtech"])(real_vm_path, real_disk_path, scope)
        else:
            rs =  getattr(sp, "%s_remove_disk" % vm["vtech"])(real_vm_path, real_disk_path)

        logger.info("remove_vm_disk: vm('%s') disk('%s'): rs=%s" % (vm_path, disk_path, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("remove_vm_disk: failed. vm('%s') disk('%s') => %s" % \
                (vm_path, disk_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_memory(vm_path, memory, maxmem=-1, scope="cfg"):
    """
    set vm memory size in the configuration file, restart vm to take effect.

    @param memory   must > 256
    @xref xen_set_memory()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        memory, maxmem = int(memory), int(maxmem)

        try:
            if memory < 256:
                raise Exception("memory must >= 256.")
        except ValueError:
            raise Exception("memory must be integer.")

        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_set_memory" % vm["vtech"])(real_vm_path, memory, maxmem, scope)
        else:
            rs = getattr(sp, "%s_set_memory" % vm["vtech"])(real_vm_path, memory)

        logger.info("set_vm_memory: vm('%s') memory=%d maxmem=%d scope='%s': rs=%s" % \
                (vm_path, memory, maxmem, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_memory: failed. vm('%s') memory=%d maxmem=%d scope='%s' => %s" % \
                (vm_path, memory, maxmem, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_memory(vm_path, scope="cfg"):
    """
    get vm memory size

    @xref xen_get_memory()
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_get_memory" % vm["vtech"])(real_vm_path, scope)
        else:
            rs = getattr(sp, "%s_get_memory" % vm["vtech"])(real_vm_path)

        logger.info("get_vm_memory: vm('%s') scope=%s rs=%s" % \
                (vm_path, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_memory: failed. scope=%s vm('%s') => %s" % \
                (vm_path, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_vcpus(vm_path, vcpus, cpus="NONE", vcpu=-1, cpu=-1, scope="cfg"):
    """
    set vm cpu

    @xref xen_set_vcpus()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        vm = db_get_vm(vm_path)
        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        vcpus = int(vcpus)
        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_set_vcpus" % vm["vtech"])(real_vm_path, vcpus, cpus, vcpu, cpu, scope)
        else:
            rs = getattr(sp, "%s_set_vcpus" % vm["vtech"])(real_vm_path, vcpus)


        logger.info("set_vm_vcpus: vm('%s') vcpus=%d scope=%s => rs=%s" % \
                (vm_path, vcpus, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_vcpus: failed. vm('%s') vcpus=%d scope=%s => %s" % \
                (vm_path, vcpus, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_vcpus(vm_path, scope="cfg"):
    """
    get vm cpu

    @xref xen_get_vcpus()
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_get_vm(vm_path)
        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        #FIXME: why ignore scope?
        rs = getattr(sp, "%s_get_vcpus" % vm["vtech"])(real_vm_path)

        logger.info("get_vm_vcpus: vm('%s') scope=%s rs=%s" % \
                (vm_path, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vcpus: failed. vm('%s') scope=%s => %s" % \
                (vm_path, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_vifs(vm_path, vif_cfg, mode="setall", scope="cfg"):
    """
    set vm vif
    @xref  xen_set_vifs()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        vm = db_get_vm(vm_path)
        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_set_vifs" % vm["vtech"])(real_vm_path, vif_cfg, mode, scope)
        else:
            #make the parameter format compatible
            l = [",".join(vif_attr.split(",")[:2]) for vif_attr in vif_cfg.split(";")]
            vif_cfg = ";".join(["%s%s" % (s, "," * (1 - s.count(','))) for s in l])
            rs = getattr(sp, "%s_set_vifs" % vm["vtech"])(real_vm_path, vif_cfg)

        logger.info("set_vm_vifs: vm('%s') vif_cfg=%s scope=%s => rs=%s" % \
                (vm_path, vif_cfg, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_vifs: failed. vm('%s') vif_cfg=%s scope=%s => %s" % \
                (vm_path, vif_cfg, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_vifs(vm_path, scope="cfg"):
    """
    get vm vif
    @xref   xen_get_vifs()
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)

        vm = db_get_vm(vm_path)
        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_get_vifs" % vm["vtech"])(real_vm_path, scope)
        else:
            rs = getattr(sp, "%s_get_vifs" % vm["vtech"])(real_vm_path)

        logger.info("get_vm_vifs: vm('%s') scope=%s => rs=%s" % (vm_path, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vifs: failed. vm('%s') scope=%s => %s" % (vm_path, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def clear_vm_vif_mac(vm_path):
    """
    wrapper for xen_clear_vif_mac()

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        (_hostname, sp) = get_utility_server()
        real_vm_path = db_get_real_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        rs = getattr(sp, "%s_clear_vif_mac" % vm["vtech"])(real_vm_path)

        logger.info("clear_vm_vif_mac: vm('%s') => rs=%s" % (vm_path, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("clear_vm_vif_mac: failed. vm('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

#!!manager import function still use this function
@exposed
def get_vm_vif(vm_path):
    """
    get vm vif

    @return         "success:vif=<num of vif>" if success else "failed:<error message>"
    !!deprecated!!
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_get_vm(vm_path)
        #(hostname, sp) = select_one_server(vm["vtech"])
        (hostname, sp) = get_utility_server()
        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_vif_num" % vm["vtech"])(real_vm_path)

        logger.info("get_vm_vif: vm('%s') rs=%s" % \
                (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vif: failed. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_vdisks(vm_path, scope="cfg"):
    """
    get vm vdisks

    @xref   xen_get_vdisks()
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)

        vm = db_get_vm(vm_path)
        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        if scope in ("cfg", "both"):
            vm_cfg_file = join(real_vm_path, "vm.cfg")
            rs = getattr(sp, "%s_correct_cfg" % vm["vtech"])(vm_cfg_file, real_vm_path)
            if not rs.startswith("success"): raise Exception(rs)

        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_get_vdisks" % vm["vtech"])(real_vm_path, scope)
        else:
            rs = getattr(sp, "%s_get_vdisks" % vm["vtech"])(real_vm_path)
        if not rs.startswith("success"): raise Exception(rs)
        vdisks = parse_result(rs)["vdisks"].split(",")
        result = []
        for vdisk in vdisks:
            if not vdisk.strip(): continue
            rs_size = get_vdisk_size(vdisk)
            if not rs_size.startswith("success"):
                raise Exception(rs_size)
            size = parse_result(rs_size)["size"]
            result.append("%s:%s" % (vdisk, size))
        rs = ','.join(result)

        logger.info("get_vm_vdisks: success. vm('%s') scope=%s rs=%s" % (vm_path, scope, rs))
        return "success:vdisks=%s" % rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vdisks: failed. vm('%s') scope=%s => %s" % (vm_path, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_cfg_vdisks(vm_path):
    """
    get vm vdisks from the cfg file
    N.B.:
        the original reason to implement this function is that,
        when wget a vm, user want to first check if disks assosiated with this vm available.

    @return     "success:vdisks=<string of vdisks>" if success else "failed:<error message>"
    example of string of vdisks: "system.img,vmlinuz,initrd.gz"
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        #(hostname, sp) = select_one_server(vm["vtech"])
        (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_vdisks" % vm["vtech"])(real_vm_path)
        if not rs.startswith("success"): raise Exception(rs)
        vdisks_str = parse_result(rs)["vdisks"]

        if not vdisks_str:
            rs = "success:vdisks="
            logger.info("get_vm_cfg_vdisks: success. vm_path('%s') => %s" % \
                    (vm_path, rs))
            return rs

        vdisks = vdisks_str.split(",")
        vdisks_lst = []
        for vdisk in vdisks:
            vdisk = basename(vdisk)
            vdisks_lst.append(vdisk)
        vdisks_str = ",".join(vdisks_lst)

        rs = "success:vdisks=%s" % vdisks_str
        logger.info("get_vm_cfg_vdisks: success. vm_path('%s') => %s" % \
                (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_cfg_vdisks: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_memcpudisksize(vm_path):
    """
    get vm memory, vcpus, vdisks and vm size

    @return
    "success:memory=<mem size>;vcpus=<num of vcpus>;vdisks=<string of vdisks>;vm_size=<vm size>" if success
    else "failed:<error message>"
    examples of string of vdisks: "system.img:500,vmlinuz:10,initrd.gz:2", "system.raw:500"
    """
    logger = logging.getLogger("ovs.query")
    try:
        real_vm_path = db_get_real_vm_path(vm_path)

        rs_mem = get_vm_memory(real_vm_path)
        if not rs_mem.startswith('success:'): raise Exception(rs_mem)
        mem = int(parse_result(rs_mem)['memory'])

        rs_cpu = get_vm_vcpus(real_vm_path)
        if not rs_cpu.startswith('success:'): raise Exception(rs_cpu)
        cpu = int(parse_result(rs_cpu)['vcpus'])

        rs_disk = get_vm_vdisks(real_vm_path)
        if not rs_disk.startswith('success:'): raise Exception(rs_disk)
        disk = parse_result(rs_disk)['vdisks']

        rs_size = get_vm_size(real_vm_path)
        if not rs_size.startswith('success:'): raise Exception(rs_size)
        size = int(parse_result(rs_size)['size'])

        #compatible with ovs-agent-2.2-15
        #rs = "success:memory=%d;vcpus=%d;vdisks=%s;vm_size=%d" % \
        #    (mem, cpu, disk, size)
        rs = "success:memory=%d;vcpus=%d;vdisks=%s;vm_size=%d;mem_size=%d" % \
            (mem, cpu, disk, size, mem)

        logger.info("get_vm_memcpudisksize:success. vm('%s') => %s", (vm_path, rs))
        return rs

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_memcpudisksize:failed. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_type(vm_path):
    """
    get vm type

    @return     "success:type=<type>" if success else "failed:<error message>"
    type should be "hvm" or "para"
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        (hostname, sp) = select_one_server(vm["vtech"])
        logger.debug("get_vm_type: server '%s' is selected." % hostname)

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_vm_type" % vm["vtech"])(real_vm_path)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("get_vm_type: success. vm_path('%s') => %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_type: failed. vm_path('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def check_hvm_pvdriver(vm_path):
    """
    check whether a HVM's pvdriver is ready.
    @param      vm_path specifies the hvm machine
    @return     "success" if pvdriver is ready,else "failed:<reason>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        (hostname, sp) = get_vm_host(vm_path)
        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_check_hvm_pvdriver" % vm["vtech"])(real_vm_path)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("check_hvm_pvdriver: success. vm_path('%s') => %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("check_hvm_pvdriver: failed. vm_path('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_vnc_password(vm_path):
    """
    get vm vnc password from the cfg file

    @return     "success:vnc_password=<vnc_password>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_vm_vnc_password" % vm["vtech"])(real_vm_path)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("get_vm_vnc_password: success. vm_path('%s') => %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vnc_password: failed. vm_path('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_vnc_password(vm_path, vnc_password):
    """
    set vm vnc password

    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_set_vm_vnc_password" % vm["vtech"])(real_vm_path, vnc_password)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("set_vm_vnc_password: success. vm_path('%s') => %s" % \
                (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_vnc_password: failed. vm_path('%s'), => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_mem_vcpus_vncpasswd(vm_path, memory=-1, vcpus=-1, vncpasswd=None):
    """
    set vm memory, vcpus and vnc password

    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        real_vm_path = db_get_real_vm_path(vm_path)
        rs = xen_set_vm_mem_vcpus_vncpasswd(real_vm_path, memory, vcpus, vncpasswd)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("set_vm_mem_vcpus_vncpasswd: success. vm_path('%s'), memory('%s'), vcpus('%s')" % \
                (vm_path, memory, vcpus))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_mem_vcpus_vncpasswd: failed. vm_path('%s'), memory('%s'), vcpus('%s')" % \
                (vm_path, memory, vcpus))
        return "failed:%s" % errmsg

@exposed
def del_vm_vif(vm_path, macs, scope="cfg"):
    """
    del vifs

    @xref   xen_del_vif()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        vm = db_get_vm(vm_path)
        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            #(hostname, sp) = select_one_server(vm["vtech"])
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_del_vif" % vm["vtech"])(real_vm_path, macs, scope)
        else:
            rs = getattr(sp, "%s_del_vif" % vm["vtech"])(real_vm_path, macs)

        if not rs.startswith("success"): raise Exception(rs)

        logger.info("del_vm_vif: success. vm_path('%s'), macs('%s') scope=%s => %s" % \
                (vm_path, macs, scope, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("del_vm_vif: failed. vm_path('%s'), macs('%s') scope=%s => %s" % \
                (vm_path, macs, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def optimize_os_profile(vm_path, os_type, os_variant, os_arch=''):
    """
    @xref xen_optimize_os_profile()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm = db_get_vm(vm_path)
        real_vm_path = db_get_real_vm_path(vm_path)
        rs = xen_optimize_os_profile(real_vm_path, os_type, os_variant, os_arch)
        logger.info("optimize_os_profile: vm('%s') os_type=%s, os_variant=%s, os_arch=%s => %s" % \
                (vm_path, os_type, os_variant, os_arch, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("optimize_os_profile: failed. vm('%s'), os_type=%s, os_variant=%s, os_arch=%s => %s" % \
                (vm_path, os_type, os_variant, os_arch, errmsg))
        return "failed:%s" % errmsg

@exposed
def config_boot_sequence(vm_path, boot, cdrom=None):
    """
    @xref xen_config_boot_sequence()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm_path = abspath(vm_path)
        chk_vm_path(vm_path)

        real_vm_path = db_get_real_vm_path(vm_path)
        real_cdrom = None
        if cdrom:
            real_cdrom = find_real_path(cdrom)
        rs = xen_config_boot_sequence(real_vm_path, boot, real_cdrom)

        if not rs.startswith("success"): raise Exception(rs)
        logger.info("configure_boot_sequence: vm('%s') boot='%s' cdrom('%s')" % \
            (vm_path, boot, str(cdrom)))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("configure_boot_sequence: vm('%s') boot='%s' cdrom('%s')" % \
            (vm_path, boot, str(cdrom)))
        return "failed:%s" % errmsg

@exposed
def get_vm_keyboard_map(vm_path):
    """
    get vm keyboard map from the cfg file

    @return     "success:keyboard_map=<keyboard map>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_vm_keyboard_map" % vm["vtech"])(real_vm_path)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("get_vm_keyboard_map: success. vm_path('%s') => %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_keyboard_map: failed. vm_path('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_keyboard_map(vm_path, keyboard_map):
    """
    set vm keyboard map

    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)
        (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_set_vm_keyboard_map" % vm["vtech"])(real_vm_path, keyboard_map)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("set_vm_keyboard_map: success. vm_path('%s') => %s" % \
                (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_keyboard_map: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_cpu_pin(vm_path, phy_cpus, scope="cfg"):
    """
    pin the VM to only run on the specific physical CPUs
    @param phy_cpus: ['all'|'1,2'|'1-3'], get cpu information via: get_host_cpu_info
    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_set_vm_cpu_pin" % vm["vtech"])(real_vm_path, phy_cpus, scope)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("set_vm_cpu_pin: vm('%s') phy_cpus=%s scope='%s': rs=%s" % \
                (vm_path, phy_cpus, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_cpu_pin: failed. vm('%s') phy_cpus=%s scope='%s' => %s" % \
                (vm_path, phy_cpus, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_vif_qos(vm_path, mac, scope="cfg"):
    """Get QoS settings of the virtual network interface which has the
    specific MAC address.

    MAC address is used to identify a unique virtual interface.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string  
    @param mac:     MAC address of the virtual network interface
    @type mac:      string  
    @param scope:   in [ "cfg", "mem" ]
                    "cfg" - get qos configurations from vm config file
                    "mem" - get runtime qos configuration
    @type scope:    string  
    @return:        "success:rate=<rate>;latency=<latency>" on success
                    "failed:<error message>" on failure
    @rtype:         string

    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_vif_qos" % vm["vtech"])(real_vm_path, mac, scope)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("get_vm_vif_qos: vm('%s') mac=%s scope='%s': rs=%s" % \
                (vm_path, mac, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vif_qos: failed. vm('%s') mac=%s scope='%s' => %s" % \
                (vm_path, mac, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_vbd_qos(vm_path, backend_name, shared=0, scope="cfg"):
    """Get QoS settings of the virtual block device who's backend device is
    "backend_name".

    "backend_name, shared" are used to identify a unique virtual block device.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string  
    @param backend_name: backend device name of the virtual block device,e.g.: System.img
    @type backend_name:  string 
    @param shared: a boolean value to indicate whether the disk is shared
    @type shared:  integer [0:1]  
    @param scope:   in [ "cfg", "mem" ]
                    "cfg" - get qos configurations from vm config file
                    "mem" - get qos configurations from runtime
    @type scope:    string  
    @return:        "success:sched=<sched>;prio=<prio>"
                    "failed:<error message>" on failure
    @rtype:         string
    """
    logger = logging.getLogger("ovs.query")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_get_vbd_qos" % vm["vtech"])(real_vm_path, backend_name, shared, scope)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("get_vm_vbd_qos: vm('%s') backend_name=%s shared=%s scope='%s': rs=%s" % \
                (vm_path, backend_name, shared, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_vbd_qos: failed. vm('%s') backend_name=%s shared=%s scope='%s' => %s" % \
                (vm_path, backend_name, shared, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_vif_qos(vm_path, mac, rate, latency="50ms", scope="cfg"):
    """Set QoS settings of the virtual network interface which has the
    specific MAC address.

    MAC address is used to identify a unique virtual interface.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string  
    @param mac:     MAC address of the virtual network interface
    @type mac:      string  
    @param rate:    rate limit to the virtual network interface. Eg. "5mbit"
    @type rate:     string  
    @param latency: latency set to the virtual network interface.
                    Default: "50ms"
    @type latency:  string  
    @param scope:   in [ "cfg", "mem", "both" ]
                    "cfg" - set qos configurations on vm config file
                    "mem" - set qos configurations at runtime
                    "both" - do both on vm config file and runtime
    @type scope:    string  
    @return:        "success" on success
                    "failed:<error message>" on failure
    @rtype:         string
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_set_vif_qos" % vm["vtech"])(real_vm_path, mac, rate, latency, scope)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("set_vm_vif_qos: vm('%s') mac=%s, rate=%s, latency=%s, scope='%s': rs=%s" % \
                (vm_path, mac, rate, latency, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_vif_qos: failed. vm('%s') mac=%s, rate=%s, latency=%s, scope='%s' => %s" % \
                (vm_path, mac, rate, latency, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_vbd_qos(vm_path, backend_name, shared, sched, prio="4", scope="cfg"):
    """Set QoS settings of the virtual block device who's backend device is
    "backend_name".

    "backend_name, shared" are used to identify a unique virtual block device.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string 
    @param backend_name: backend device name of the virtual block device,e.g.: System.img
    @type backend_name:  string 
    @param shared: a boolean value to indicate whether the disk is shared
    @type shared:  integer [0:1] 
    @param sched:   scheduling algorithm.
                    in ["real-time", "rt", "best-effort", "idle"].
    @type sched:    string  
    @param prio:    priority for scheduling. from 0 to 7.
    @type prio:     string
    @param scope:   in [ "cfg", "mem", "both" ]
                    "cfg" - set qos configurations on vm config file
                    "mem" - set qos configuration at runtime
                    "both" - do both on vm config file and runtime
    @type scope:    string  
    @return:        "success" on success
                    "failed:<error message>" on failure
    @rtype:         string
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        vm = db_get_vm(vm_path)

        if scope in ("mem", "both"):
            (hostname, sp) = get_vm_host(vm_path)
        else:
            (hostname, sp) = get_utility_server()

        real_vm_path = db_get_real_vm_path(vm_path)
        rs = getattr(sp, "%s_set_vbd_qos" % vm["vtech"])(real_vm_path, backend_name, shared, sched, prio, scope)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("set_vm_vbd_qos: vm('%s') backend_name=%s, shared=%s, sched=%s, prio=%s, scope='%s': rs=%s" % \
                (vm_path, backend_name, shared, sched, prio, scope, rs))
        return rs
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_vbd_qos: failed. vm('%s') backend_name=%s, shared=%s, sched=%s, prio=%s, scope='%s' => %s" % \
                (vm_path, backend_name, shared, sched, prio,  scope, errmsg))
        return "failed:%s" % errmsg
