#!/usr/bin/env python
#^#
# OVSSiteVMInstall.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import logging
from OVSXUtility import get_storage_repos, assemble_real_path, find_real_path
from OVSSiteXenStore import get_vm_vnc_conn
from OVSSiteDB import register_vm
from OVSServerSelection import get_utility_server, select_one_server, select_tgt_host, get_vm_host
from OVSServerProxy import ServerProxy
from OVSDB import db_update_vm, db_get_vm, db_get_real_vm_path, db_get_srv
from OVSConstants import OVS_BASE
from OVSCommons import exposed, fmt_err_msg, chk_vm_path

#---------------------<> vm installation <>----
@exposed
def install_vm_hvm(vm_path, memory, vcpus, vif_cfg, disks, cdrom,
        vncpassword, vnclisten="0.0.0.0",
        dedicated_server=None,
        os_type='', os_variant='', keymap='en-us'):
    """
    install full virtualized vm from scratch, refer to xen_install_vm_hvm()

    @param  dedicated_server    specify server to process the installation
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        memory = int(memory)
        vcpus = int(vcpus)
        logger.debug("install_vm_hvm: vm_path=%s, memory=%d, vcpus=%d, vif_cfg=%s, disks=%s, cdrom=%s, dedicated_server=%s" % \
                (vm_path, memory, vcpus, vif_cfg, disks, cdrom, repr(dedicated_server)))

        disks_list = disks.split(",")
        disks_list = [disk.split(":") for disk in disks_list]
        vm_size = 0
        for disk_path, disk_size, sparse_option in disks_list:
            vm_size += float(disk_size) * 1024
        (repos, free_space) = get_storage_repos()
        if free_space < vm_size + 10:
            raise Exception("not enough free space")
        real_vm_path = assemble_real_path(vm_path, repos)

        vtech = "xen"
        register_vm(vm_path, vtech)
        db_update_vm(vm_path, {"vtype": "hvm"})

        if dedicated_server:
            hostname = dedicated_server
            srv = db_get_srv(hostname)
            sp = ServerProxy(srv['agt_url'])
            sp.agent_version = srv['agent_version']
        else:
            (hostname, sp) = select_tgt_host(vm_path, any=True)

        db_update_vm(vm_path, { "host_srv" : hostname, "real_path":real_vm_path })

        real_cdrom = find_real_path(cdrom, True)[0]
        disks = disks.replace(vm_path, real_vm_path)

        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_install_vm_hvm" % vtech)\
                (real_vm_path, int(memory), int(vcpus), vif_cfg, disks, real_cdrom, vncpassword, vnclisten, os_type, os_variant, keymap)
        else:
            rs = getattr(sp, "%s_install_vm_hvm" % vtech)\
                (real_vm_path, int(memory), int(vcpus), vif_cfg, disks, real_cdrom, vncpassword, vnclisten)

        if rs.startswith("success"):
            logger.info("install_vm_hvm: vm('%s') => %s" % (vm_path, rs))
            db_update_vm(vm_path, {"status": "RUNNING", "manual_shutdown": False})
        else: raise Exception(rs)

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("install_vm_hvm: vm('%s')=>%s " % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def change_vm_cdrom(vm_path, iso_path, scope="both"):
    """
    change vm cdrom, maily for multiple cd installation

    @xref   xen_change_vm_cdrom()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        if iso_path.strip():
            real_iso_path = find_real_path(iso_path, True)[0]
        else: #detach
            real_iso_path = iso_path
        vm = db_get_vm(vm_path)
        if scope in ["mem", "both"]:
            (hostname, sp) = get_vm_host(vm_path)
        else:
            (hostname, sp) = get_utility_server()
        real_vm_path = db_get_real_vm_path(vm_path)

        if sp.agent_version > (2, 2, 0, 15):
            rs = getattr(sp, "%s_change_vm_cdrom" % vm["vtech"])\
                    (real_vm_path, real_iso_path, scope)
        else:
            rs = getattr(sp, "%s_change_vm_cdrom" % vm["vtech"])\
                    (real_vm_path, real_iso_path)

        logger.info("change_vm_cdrom:vm('%s') iso('%s') => %s" % \
            (vm_path, iso_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("change_vm_cdrom:error. vm('%s') iso('%s') => %s " % \
            (vm_path, iso_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def install_vm_para(vm_path, memory, vcpus, vif_cfg, disks, location, vncpassword, vnclisten="0.0.0.0", dedicated_server=None, keymap='en-us'):
    """
    install para virtualized vm from scratch

    @xref   xen_install_vm_para()
    @param dedicated_server     if not specified, pool master will pickup one for you

    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        memory = int(memory)
        vcpus = int(vcpus)
        logger.debug("install_vm_para: vm_path=%s, memory=%d, vcpus=%d, vif_cfg=%s, disks=%s, location=%s, dedicated_server=%s, keymap=%s" % \
                (vm_path, memory, vcpus, vif_cfg, disks, location, repr(dedicated_server), keymap))

        disks_list = disks.split(",")
        disks_list = [disk.split(":") for disk in disks_list]
        vm_size = 0
        for disk_path, disk_size, sparse_option in disks_list:
            vm_size += float(disk_size) * 1024

        (repos, free_space) = get_storage_repos()
        if free_space < vm_size + 10:
            raise Exception("not enough free space")
        real_vm_path = assemble_real_path(vm_path, repos)

        vtech = "xen"
        register_vm(vm_path, vtech)
        db_update_vm(vm_path, {"vtype": "xen"})

        if dedicated_server:
            hostname = dedicated_server
            srv = db_get_srv(hostname)
            sp = ServerProxy(srv['agt_url'])
        else:
            (hostname, sp) = select_tgt_host(vm_path, any=True)

        db_update_vm(vm_path, {"host_srv" : hostname, "real_path": real_vm_path})

        disks = disks.replace(vm_path, real_vm_path)
        rs = getattr(sp, "%s_install_vm_para" % vtech)\
                    (real_vm_path, int(memory), int(vcpus), vif_cfg, disks, location, vncpassword, vnclisten, keymap)
        if rs.startswith("success"):
            logger.info("install_vm_para: vm('%s') => %s" % (vm_path, rs))
            db_update_vm(vm_path, {"status": "RUNNING", "manual_shutdown" : False})
        else: raise Exception(rs)

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("install_vm_para: vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def create_dummy_vm(vm_path, memory, vcpus, vif_cfg, disks, vtype, vncpassword, vnclisten="0.0.0.0", dedicated_server=None):
    """
    create a dummy vm

    @xref   xen_create_dummy_vm()
    @param  dedicated_server    specify a dedicated server to create this vm

    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        memory, vcpus = int(memory), int(vcpus)
        if vtype == "para": vtype = "pvm"
        if vtype not in ["hvm", "pvm"]:
            raise Exception("Wrong vtype: %s." % vtype)
        logger.debug("create_dummy_vm: vm_path=%s, memory=%d, vcpus=%d, vif_cfg=%s, disks=%s, vtype=%s dedicated_server=%s" % \
                (vm_path, memory, vcpus, vif_cfg, disks, vtype, repr(dedicated_server)))

        disks_list = disks.split(",")
        disks_list = [disk.split(":") for disk in disks_list]
        vm_size = 0
        for disk_path, disk_size, sparse_option in disks_list:
            vm_size += float(disk_size) * 1024

        (repos, free_space) = get_storage_repos()
        if free_space < vm_size + 10:
            raise Exception("not enough free space")
        real_vm_path = assemble_real_path(vm_path, repos)

        vtech = "xen"
        register_vm(vm_path, vtech)
        db_update_vm(vm_path, {"vtype": "xen"})

        if dedicated_server:
            hostname = dedicated_server
            srv = db_get_srv(hostname)
            sp = ServerProxy(srv['agt_url'])
            sp.agent_version = srv["agent_version"]
        else:
            (hostname, sp) = select_tgt_host(vm_path, any=True)

        if not sp.agent_version > (2, 2, 0, 15):
            raise Exception("srv('%s') does not support api: xen_create_dummy_vm()")

        db_update_vm(vm_path, {"host_srv" : hostname, "real_path": real_vm_path})

        disks = disks.replace(vm_path, real_vm_path)
        rs = getattr(sp, "%s_create_dummy_vm" % vtech)\
                    (real_vm_path, memory, vcpus, vif_cfg, disks, vtype, vncpassword, vnclisten)
        if rs.startswith("success"):
            logger.info("create_dummy_vm: vm('%s') => %s" % (vm_path, rs))
            db_update_vm(vm_path, {"status": "DOWN", "manual_shutdown": True})
        else: raise Exception(rs)

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("create_dummy_vm: vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg
