#!/usr/bin/env python
#^#
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import re
import os
import logging
from os.path import dirname, basename, join, isdir, isfile, exists
from glob import glob
from OVSServerProxy import ServerProxy
from OVSConstants import HA_TOUCH_PREFIX, HA_NFS_HEARTBEAT_FILE, OVS_BASE 
from OVSConstants import DLM_DOMAIN, AGENT_DIR, AGENT_GUARD_PROCESS, XEND_CONFIG, XEND_DLM_LOCK, DLM
from OVSCommons import fmt_err_msg, exposed, parse_result, marshall_string, unmarshall_string, run_cmd, run_cmd_bg, parse_shared_fs
from OVSCDS import CDS

@exposed
def ha_check_exist(path):
    logger = logging.getLogger("ovs.query")
    try:
        _dirname = dirname(path)
        filename = "%s%s" % (HA_TOUCH_PREFIX, basename(path))
        touch_file = join(_dirname, filename)
        if not isfile(touch_file):
            raise Exception("file touch path('%s') does not exists!" % touch_file)
        logger.info("ha_check_exist: success. file touch path('%s')." % touch_file)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_exist: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_mount_shared_storage(marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra=""):
    logger = logging.getLogger("ovs.operation")
    #---------------------------------------------------------------
    def _find_ocfs2_dev(uuid, dev_peer):
        # dev_peer: "dev/sdb1" -> ptn: "/dev/sd*1"
        v = list(dev_peer)
        idx = -1
        while v[idx] in "0123456789":
            idx -= 1
        v[idx] = "*"
        ptn = "".join(v)
        devs = glob(ptn)

        #find dev with same uuid
        tgt_dev = []
        for dev in devs:
            try:
                _uuid = run_cmd("debugfs.ocfs2 %s <<<stats 2>/dev/null | awk '/UUID/ {print $2;}'" % dev).splitlines()[-1]
                if _uuid == uuid:
                    tgt_dev.append(dev)
            except:
                pass

        return tgt_dev
    #---------------------------------------------------------------
    try:
        #base shared storage
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()

        master_shared_fs_line = unmarshall_string(marshalled_shared_fs_line)
        master_shared_fs, master_mount_point, master_shared_fs_type, master_mount_option = master_shared_fs_line.split()[0:4]

        if shared_fs and shared_fs_type == "ocfs2" and master_shared_fs_type == "ocfs2":
            #already mounted and is ocfs2 partition mounted by device name
            uuid = parse_result("success:%s" % shared_fs_extra)["uuid"]
            tgt_dev = _find_ocfs2_dev(uuid, master_shared_fs)
            if shared_fs not in tgt_dev:
                raise Exception("%s already mounted but is not the same as master node." % OVS_BASE)
        elif shared_fs and (master_shared_fs_type != shared_fs_type or master_shared_fs != shared_fs):
            #already mounted, but not the same as master node
                raise Exception("%s already mounted but is not the same as master node." % OVS_BASE)
        elif not shared_fs:
            #local filesytem, try to mount the shared storage
            if master_shared_fs_type == "ocfs2" and master_shared_fs.startswith("/dev"):
                uuid = parse_result("success:%s" % shared_fs_extra)["uuid"]
                tgt_dev = _find_ocfs2_dev(uuid, master_shared_fs)
                if tgt_dev:
                    shared_fs = tgt_dev[0]
                else:
                    raise Exception("cannot find the equivalent filesystem as master partition with uuid='%s'" % uuid)
            else:
                shared_fs = master_shared_fs
            cmd = "mount -t %s -o %s %s %s" % (master_shared_fs_type, master_mount_option, shared_fs, master_mount_point)
            run_cmd(cmd)

        #extended shared storage
        shared_repositories = unmarshall_string(marshalled_shared_repositories)
        f = open("/etc/ovs/repositories", "w")
        try:
            f.write(shared_repositories)
        finally:
            f.close()
        cmd = "service ovsrepositories start"
        run_cmd(cmd)

        logger.info("ha_mount_shared_storage: success")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_mount_shared_storage: error. =>%s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_setup_ocfs2_cluster(marshalled_cluster_config, shared_fs_type="ocfs2"):
    """
    *) reconfigure the o2cb service
    *) start the heartbeating
        /OVS is ocfs2, mount the partition
        /OVS is nfs, start heartbeat against the dmnfs heartbeat device
    """
    logger = logging.getLogger("ovs.operation")
    try:
        #configure & start cluster service
        if not isdir("/etc/ocfs2"):
            os.makedirs("/etc/ocfs2")
        f = open("/etc/ocfs2/cluster.conf", "w")
        try:
            f.write(unmarshall_string(marshalled_cluster_config))
        finally:
            f.close()

        logger.info("ha_setup_ocfs2_cluster: trying to reconfigure the cluster...")
        cmd = join(dirname(__file__), "utils/config_o2cb.sh")
        run_cmd(cmd)

        #start heartbeating
        if shared_fs_type == "ocfs2":
            #mount will let the server's start hearting, see ha_sync_shared_storage_mount()
            pass
        elif shared_fs_type == "nfs":
            cmd = "/usr/lib/ovs/ovs-cluster-check --alter-fstab"
            run_cmd(cmd)
        else:
            raise Exception("filesystem type '%s' not in [ocfs2, nfs]." % shared_fs_type)

        logger.info("ha_setup_ocfs2_cluster: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_setup_ocfs2_cluster: error. =>%s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_check_cluster_heartbeat():
    """
    check wheter the ocfs2 cluster heartbeat is alive
    """
    logger = logging.getLogger("ovs.query")
    try:
        rs = "failed"
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
        if shared_fs_type == "ocfs2":
            cmd = "ocfs2_hb_ctl -I -d %s" % shared_fs
            l = run_cmd(cmd).splitlines()
            if l[-1].split()[1] != "0":
                rs = "success"
        elif shared_fs_type == "nfs":
            cmd = "dmsetup table --target=nfs"
            l = run_cmd(cmd).splitlines()
            for line in l:
                parts = line.split()
                if len(parts) > 4 and parts[4] == HA_NFS_HEARTBEAT_FILE:
                    dev = parts[0].split(":")[0]
                    cmd = "ocfs2_hb_ctl -I -d /dev/mapper/%s" % dev
                    _l = run_cmd(cmd).splitlines()
                    if _l[-1].split()[1] != "0":
                        rs = "success"
                        break
        else:
            raise Exception("only [ocfs2, nfs] shared filesystem is supported.")
        
        logger.info("ha_check_cluster_heartbeat: shared_fs='%s' %s." % (shared_fs, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_cluster_heartbeat: error. %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def ha_join_dlm_domain(domain_name="ovm"):
    logger = logging.getLogger("ovs.operation")
    try:
        rs = ha_check_cluster_heartbeat()
        if not rs.startswith("success"):
            raise Exception("server is not heartbeating!")
        
        _dir = join(DLM, basename(domain_name))
        if not isdir(_dir):
            #find heartbeat device
            hb_dev = ""
            shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
            if shared_fs_type == "ocfs2":
                hb_dev = shared_fs
            elif shared_fs_type == "nfs":
                cmd = "dmsetup table --target=nfs"
                l = run_cmd(cmd).splitlines()
                for line in l:
                    parts = line.split()
                    if len(parts) > 4 and parts[4] == HA_NFS_HEARTBEAT_FILE:
                        hb_dev = join("/dev/mapper", parts[0].split(":")[0])
                        break
            else:
                raise Exception("currently only [ocfs2, nfs] shared filesystems are supported.")

            #add heartbeat reference
            cmd = "ocfs2_hb_ctl -S -d %s" % hb_dev
            run_cmd(cmd)

            #join domain
            os.mkdir(_dir)
            logger.info("ha_join_dlm_domain: success.")

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_join_dlm_domain: domain('%s') error. => %s" % (domain_name, errmsg))
        return "failed:%s" % errmsg

@exposed
def ha_set_dlm_lock(lock, name=""):
    """
    @param lock uuid of the vm
    @param name name of the vm
    @return     "success" if the lock successfully or "failed" with the error message
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not lock.strip():
            raise Exception("lock must be specified!")

        rs = ha_join_dlm_domain(DLM_DOMAIN)
        if not rs.startswith("success"):
            raise Exception(rs)
        
        #limit dlm lock name to 31 chars
        lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])

        if exists(lock_path):
            raise Exception("lock('%s') has already been set!" % lock_path)

        try:
            fd = os.open(lock_path, os.O_CREAT | os.O_NONBLOCK)
            os.close(fd)
        except Exception, e:
            try:
                os.remove(lock_path) #get rid of the side effect
            except:
                pass
            errmsg = fmt_err_msg(e)
            raise Exception("create lock('%s') failed. %s" % (lock_path, errmsg))

        #create the EX lock
        fd = os.open(lock_path, os.O_RDWR | os.O_NONBLOCK)
        #close it, or the agent process will hold the lock
        os.close(fd)

        logger.info("ha_set_dlm_lock: lock('%s') has been set for vm('%s')" % (lock, name))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_set_dlm_lock:failed. lock('%s') name('%s')=> %s" % (lock, name, errmsg))
        return "failed: %s" % errmsg

ha_accquire_lock = ha_set_dlm_lock

@exposed
def ha_release_dlm_lock(lock, name=""):
    """
    @param lock uuid of the vm
    @param name name of the vm
    @return     "success" if the lock successfully or "failed" with the error message
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not lock.strip():
            raise Exception("lock must be specified!")

        lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])
        if exists(lock_path):
            os.remove(lock_path)
        logger.info("ha_release_dlm_lock:lock('%s') name('%s') success." % (lock, name))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_release_dlm_lock:failed. lock('%s') name('%s') => %s" % (lock, name, errmsg))
        return "failed: %s" % errmsg

ha_release_lock = ha_release_dlm_lock

@exposed
def ha_config_xend_external_lock(operation="add"):
    """
    add '%s' to file '%s' and restart xend
    
    """ % (XEND_CONFIG, XEND_DLM_LOCK)
    logger = logging.getLogger("ovs.operation")
    try:
        if operation not in ("add", "remove"):
            raise Exception("only operation in ('add', 'remove') is acceptable.")

        f = open(XEND_CONFIG)
        try:
            lines = f.readlines()
        finally:
            f.close()
        ptn = re.compile(r"^\s*#?\s*\(\s*xend-domains-lock-path.*\)")
        cmt_ptn = re.compile(r"^\s*#")
        l1, l2 = [], [] #lock line & commented out lock line
        for idx, line in enumerate(lines):
            if  ptn.search(line):
                if cmt_ptn.match(line):
                    l2.append(idx)
                else:
                    l1.append(idx)

        xend_restart = True
        if operation == "add":
            if l1: #line already present
                idx = l1[-1]
                if lines[idx] != XEND_DLM_LOCK:
                    lines[idx] = XEND_DLM_LOCK
                else:
                    xend_restart = False
            elif l2: #comments are there
                idx = l2[-1] + 1
                lines.insert(idx, XEND_DLM_LOCK)
            else:
                lines.append(XEND_DLM_LOCK)
        if operation == "remove":
            if l1:
                for idx in l1:
                    if AGENT_DIR in lines[idx]:
                        del lines[idx]        
            else:
                xend_restart = False

        f = open(XEND_CONFIG, "w")
        try:
            f.writelines(lines)
        finally:
            f.close()

        #make the configuration take effect
        if xend_restart:
            cmd = "service xend restart"
            run_cmd(cmd)

        logger.info("ha_config_xend_external_lock: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_config_xend_external_lock: error. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def ha_start_agent_guard():
    logger = logging.getLogger("ovs.operation")
    try:
        run_cmd_bg(AGENT_GUARD_PROCESS)

        logger.info("ha_start_agent_guard: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_start_agent_guard: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_pull_cluster_setup():
    """
    when agent startup, pull cluster configuration from master node and set it up.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        master_agt_url = ""
        cds = CDS("master")
        try:
            master_agt_url = cds.db_master.get("master_agt_url")
        finally:
            cds.close()
        if not master_agt_url:
            logger.warning("ha_pull_cluster_setup: no master access infomation available!")
            return "success"

        sp = ServerProxy(master_agt_url)
        #shared storage configuration
        rs = sp.ha_get_shared_fs_conf()
        if not rs.startswith("success"):
            raise Exception(rs)
        marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra = rs[7:].strip().split("@@@")

        shared_fs_line = unmarshall_string(marshalled_shared_fs_line).strip()
        if not shared_fs_line:
            raise Exception("shared storage is not mounted on master server!")
        parts = shared_fs_line.split()
        shared_fs, shared_fs_type = parts[0], parts[2]

        #cluster configuration
        rs = sp.ha_get_cluster_conf()
        if not rs.startswith("success"):
            raise Exception(rs)
        marshalled_cluster_config = rs[8:] #strip "success:"

        #setup
        if shared_fs_type == "ocfs2":
            rs = ha_setup_ocfs2_cluster(marshalled_cluster_config, shared_fs_type)
            if not rs.startswith("success"):
                raise Exception(rs)

            rs = ha_mount_shared_storage(marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra)
            if not rs.startswith("success"):
                raise Exception(rs)

        elif shared_fs_type == "nfs":
            rs = ha_mount_shared_storage(marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra)
            if not rs.startswith("success"):
                raise Exception(rs)

            rs = ha_setup_ocfs2_cluster(marshalled_cluster_config, shared_fs_type)
            if not rs.startswith("success"):
                raise Exception(rs)
        else:
            raise Exception("%s not in supported shared filesystem [ocfs2, nfs]!" % shared_fs_type)
        
        rs = ha_start_agent_guard()
        if not rs.startswith("success"):
            raise Exception(rs)

        logger.info("ha_pull_cluster_setup:success")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_pull_cluster_setup:failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def ha_set_shutdown_mode(action="enter"):
    """
    1) tell the master the server is in shutdown mode
    2) create the shutdown mode flag
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("ha_set_shutdown_mode: inform master agent that i am %sing shutdown mode." % action)
        master_agt_url = ""
        cds = CDS("master")
        try:
            master_agt_url = cds.db_master.get("master_agt_url")
        finally:
            cds.close()
        if not master_agt_url:
            logger.warning("ha_set_shutdown_mode: no master access infomation available!")
        else:
            from OVSMonitorServer import get_my_ips
            my_ips = get_my_ips()
            my_ips = [ip for ip in my_ips if not ip.startswith("127")]
            sp = ServerProxy(master_agt_url)
            rs = "success"
            if action == "enter":
                rs = sp.set_shutdown_mode(",".join(my_ips), True)
            elif action == "leave":
                rs = sp.set_shutdown_mode(",".join(my_ips), False)

            if not rs.startswith("success"):
                raise rs
        
        from OVSConstants import AGENT_SHUTDOWN_MODE_FLAG
        if action == "enter":
            logger.info("ha_set_shutdown_mode: create shutdown mode flag.")
            open(AGENT_SHUTDOWN_MODE_FLAG, "w").close()
        elif action == "leave":
            logger.info("ha_set_shutdown_mode: remove shutdown mode flag.")
            if exists(AGENT_SHUTDOWN_MODE_FLAG):
                os.remove(AGENT_SHUTDOWN_MODE_FLAG)
            
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_set_shutdown_mode: failed. =>%s" % errmsg)
        return "failed: %s" % errmsg

