#!/usr/bin/env python

#^#
# OVSMonitorAgent.py
#
# ovs module, provides monitoring fuction, gether system information,
# performance data, etc
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import socket
import logging
from OVSTimer import Timer
from OVSCommons import fmt_err_msg, run_cmd, exposed, parse_result
from OVSConstants import OVS_BASE
from OVSXXenStore import xen_get_caps
from os.path import basename

@exposed
def sys_vmstat_info():
    """
    system stat data by vmstat
    """
    logger = logging.getLogger("ovs.performance")
    try:
        stdout = run_cmd("vmstat -n -S M")
        lines = stdout.splitlines()
        keys = lines[1].split()
        values = lines[2].split()
        d = dict(zip(keys, values))

        cpu_idle, mem_free = ("%.2f" % (float(d["id"])/100,),
                "%d" % (int(d["free"]) + int(d["buff"]) + int(d["cache"])))
        logger.info("sys_vmstat_info:cpu_idle=%s;mem_free=%s" % (cpu_idle, mem_free))
        return "success:cpu_idle=%s;mem_free=%s" % (cpu_idle, mem_free)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_vmstat_info: failed! => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def sys_perf_info(host_type="srv", df_path=OVS_BASE):
    """
    cpu_usage, mem_usage, hd_usage,
    hd_io (TODO), network_io (TODO)
    called by performance collector

    @param host_type: if is srv, then try xm info to get free memory
    """
    logger = logging.getLogger("ovs.performance")
    try:
        #vmstat
        stdout = run_cmd("vmstat -n -S M")
        lines = stdout.splitlines()
        keys = lines[1].split()
        values = lines[2].split()
        d = dict(zip(keys, values))
        cpu_idle, mem_free = ("%.2f" % (float(d["id"])/100,),
                "%d" % (int(d["free"]) + int(d["buff"]) + int(d["cache"])))

        if host_type == "srv":
            #xm info
            #mem_free from xend view  if available will be preferred
            try:
                cmd = "xm info | grep free_memory"
                stdout = run_cmd(cmd)
                mem_free_xen = int(stdout.split(":")[1].strip())
                #use the bigger mem_free
                if mem_free_xen > int(mem_free):
                    mem_free = "%d" % mem_free_xen
            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.info("sys_perf_info: xm info=>%s" % errmsg)
        #df
        stdout = run_cmd("df -B M %s" % df_path)
        lines = stdout.splitlines()
        idx0 = lines[0].index("Used") + 5
        idx1 = lines[0].index("Use%")
        hd_avail = lines[-1][idx0:idx1].strip(' M')

        rs = "cpu_idle=%s;mem_free=%s;hd_avail=%s" % (cpu_idle, mem_free, hd_avail)
        logger.info("sys_perf_info:%s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_perf_info: failed! => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def sys_os_info():
    """
    get architecture, cpu number, cpu speed (bogomips), memory size (in MB),

    @return     'success:arch=<val>;mem_total=<val><value from sys_cpu_info()>'
                    if success else 'failed:<error message>'
    """
    logger = logging.getLogger("ovs.query")
    try:
        #import platform
        #arch = platform.architecture()[0]
        rs = xen_get_caps()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)["caps"]
        if "x86_64" in rs:
            arch = "64bit"
        else:
            arch = "32bit"

        #memory
        f = open("/proc/meminfo")
        try:
            lines = f.readlines()
        finally:
            f.close()
        lines = filter(lambda line: line.find("MemTotal") != -1, lines)
        mem_total = int(lines[0].split()[1]) / 1024

        #in OVS the above one the the total memory available for domain0
        try:
            stdout = run_cmd("xm info | grep total_memory")
            mem_total_xen = int(stdout.split(":")[1].strip())
            if mem_total_xen > mem_total:
                mem_total = mem_total_xen
        except: pass

        rs = sys_cpu_info()
        if not rs.startswith("success"): raise Exception(rs)
        rs = rs[8:] #remove success:
        rs = "hostname=%s;arch=%s;mem_total=%s;%s" % (socket.gethostname(), arch, mem_total, rs)

        logger.info("sys_os_info:success. %s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_os_info: failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def sys_cpu_info():
    """
    get chip arch info

    return    "success:vendor_id=<val>;cpu_family=<val>;model=<val>;model_name=<val>;stepping=<val>;bogomips=<val>;cpu_num=<val>"
    """
#        vendor_id       : GenuineIntel
#        cpu family      : 6
#        model           : 15
#        model name      : Intel(R) Core(TM)2 CPU          6400  @ 2.13GHz
#        stepping        : 2

    logger = logging.getLogger("ovs.query")
    try:
        f = open("/proc/cpuinfo")
        try:
            d, count = {}, {}
            for line in f:
                try:
                    k, v = line.split(":")
                    k = "_".join(k.strip().split())
                    d[k] = v.strip()
                    if k in count: count[k] += 1
                    else:count[k] = 1
                except: pass
        finally:
            f.close()

        d_arch = {}
        for k in ("vendor_id", "cpu_family", "model", "model_name", "stepping", "bogomips"):
            if k in d:
                d_arch[k] = d[k]
            else:
                d_arch[k] = ""
        d_arch["cpu_num"] = str(count["bogomips"])
        rs = ";".join(sorted(["%s=%s" % (k, v) for (k, v) in d_arch.items()]))
        logger.info("sys_cpu_info:success. %s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_cpu_info:failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def xen_hypervisor_info():
    """
    get xm info
    return success:mem_free=<mem_free>;vcaps=<xen_caps>
    """
    logger = logging.getLogger("ovs.query")
    try:
        cmd = "xm info"
        output = run_cmd(cmd)
        d = {}
        for line in output.splitlines():
            try:
                parts = line.split(":", 1)
                d[parts[0].strip()] = parts[1].strip()
            except: pass

        rs = "success:mem_free=%s;vcaps=%s" % (d["free_memory"], d["xen_caps"])
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_hypervisor_info: failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def xen_get_uptime(vm_path):
    """
    get vm uptime
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_name = basename(vm_path)
        lines = run_cmd("xm uptime '%s'" % vm_name)
        lines = lines.split('\n')
        columns = lines[0]
        index = columns.find('Uptime')
        if index < 0:
            raise "Failed to locate 'uptime' column."
        line = lines[1]
        uptime = line[index:].strip()

        logger.info("xen_get_uptime: success. vm_path('%s') => %s" % (vm_path, uptime))
        return "success:uptime=%s" % uptime
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_uptime: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg
