#!/usr/bin/env python

#^#
# OVSXXenVM.py
#
# ovs agent module, provides operation like start,stop etc for vm
# on Xen host server.
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import re
import os
import logging
import commands
from shutil import copyfile
from os.path import join, isdir, isfile, islink, exists, dirname, basename, abspath, isabs
from OVSCommons import host_alive, parse_result, fmt_err_msg, run_cmd, chk_vm_path, exposed, check_shutdown_mode
from OVSTimer import Timer
from OVSXXenVMConfig import xen_correct_cfg, xen_update_cfg_nics

#----------------------------<> vm general operation <>----

@exposed
def xen_start_vm(vm_path, nics_enc=None):
    """
    start vm
    """
    vm_cfg_file = join(vm_path, "vm.cfg")
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        check_shutdown_mode()

        #since we cp a vm from seeds or other place
        xen_correct_cfg(vm_cfg_file, vm_path)

        if(nics_enc):
            #pass network configuration throu kernel cmdline
            xen_update_cfg_nics(vm_cfg_file, nics_enc)

        #try to start vm
        cmd = "xm create '%s'" % vm_cfg_file
        run_cmd(cmd)

        #bug#4996, remove vm.save after restore, when error occured in restore, clear it here
        try:
            #for the save_path, refer to save_vm()
            save_path = join(vm_path, "vm.save")
            if exists(save_path): os.remove(save_path)
        except Exception, e:
            logger.warning("xen_start_vm: delete dump file '%s' failed. => %s" % \
                    save_path, fmt_err_msg(e))

        logger.info("xen_start_vm: success. vm('%s')" % (vm_path, ))
        return "success"
    except Exception, e:
        logger.error("xen_start_vm:error! vm('%s')=>%s" % (vm_path, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

@exposed
def xen_stop_vm(vm_path):
    """
    stop vm
    """
    vm_path = abspath(vm_path)
    vm_name = basename(vm_path)
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        try:
            run_cmd("xm list '%s'" % vm_name)
        except Exception, e:
            logger.warning("xen_stop_vm: warning. vm('%s') does not exist in the list of active domains." \
                    % vm_path)
            return "success"

        try:
            run_cmd("xm shutdown '%s'" % vm_name)
            last, interval = 600, 5 #10 mins
            timer = Timer(last, interval)
            while timer.wait():
                try: run_cmd("xm list '%s'" % vm_name)
                except: break
            if timer.timeout():
                raise Exception("not shutdown after 10mins, destroy it!")
        except Exception, e:
            logger.warning("xen_stop_vm: xm shutdown -w '%s' => %s" % (vm_name, fmt_err_msg(e)))
            run_cmd("xm destroy '%s'" % vm_name)

        logger.info("xen_stop_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_stop_vm: failed. vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_reboot_vm(vm_path):
    """
    reboot vm
    """
    vm_path = abspath(vm_path)
    vm_name = basename(vm_path)
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        check_shutdown_mode()

        try: run_cmd("xm list '%s'" % vm_name)
        except: return xen_start_vm(vm_path)

        run_cmd("xm reboot -w '%s'" % vm_name)

        logger.info("xen_reboot_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_reboot_vm: failed. vm('%s')=>%s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_pause_vm(vm_path):
    """
    pause vm
    """
    vm_path = abspath(vm_path)
    vm_name = basename(vm_path)
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        run_cmd("xm pause '%s'" % vm_name)
        timer = Timer(60, 10)
        while timer.wait():
            stdout = run_cmd("xm list '%s'" % vm_name)
            lines = stdout.splitlines()
            stat = lines[-1].split()[-2]
            if stat == "--p---": break
        if timer.timeout(): raise Exception("pause vm operation time out!")

        logger.info("xen_pause_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_pause_vm: failed. vm('%s')=>%s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_unpause_vm(vm_path):
    """
    xm unpause vm
    """
    vm_path = abspath(vm_path)
    vm_name = basename(vm_path)
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        run_cmd("xm unpause '%s'" % vm_name)
        timer = Timer(60, 10)
        while timer.wait():
            stdout = run_cmd("xm list '%s'" % vm_name)
            lines = stdout.splitlines()
            stat = lines[-1].split()[-2]
            if stat == "-b----": break
            elif stat == "r-----": break
        if timer.timeout(): raise Exception("unpause vm operation time out!")

        logger.info("xen_unpause_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_unpause_vm: failed. vm('%s')=>%s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_save_vm(vm_path, save_path):
    """
    xm save vm save
    """
    vm_path = abspath(vm_path)
    vm_name = basename(vm_path)
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        chk_vm_path(save_path)

        run_cmd("xm save '%s' '%s'" % (vm_name, save_path))

        timer = Timer(60, 10)
        while timer.wait():
            try: run_cmd("xm list '%s'" % vm_name)
            except: break
        if timer.timeout(): raise Exception("save vm operation time out!")

        logger.info("xen_save_vm: success. vm('%s') save('%s')" % \
                (vm_path, save_path))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_save_vm: failed. vm('%s') save('%s')=>%s" % \
                (vm_path, save_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_restore_vm(save_path):
    """
    xm restore save
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(save_path)
        check_shutdown_mode()

        stdout = run_cmd("xm restore '%s'" % save_path)
        #bug#4996, remove vm.save after restore
        try:
            os.remove(save_path)
        except Exception, e:
            logger.warning("xen_restore_vm: delete dump file '%s' failed. => %s" % \
                    save_path, fmt_err_msg(e))

        logger.info("xen_restore_vm: success. save('%s') => %s" % (save_path, stdout))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_restore_vm: failed. save('%s') => %s" % (save_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_migrate_vm(vm_path, tgt_host, is_live=True):
    """
    xm migrate vm tgt_host
    """
    logger = logging.getLogger("ovs.operation")
    vm_path = abspath(vm_path)
    try:
        chk_vm_path(vm_path)

        vm_name = basename(vm_path)
        if is_live:
            cmd = "xm migrate --live --ssl '%s' '%s'" % (vm_name, tgt_host)
        else:
            cmd = "xm migrate --ssl '%s' '%s'" % (vm_name, tgt_host)

        (rc, output) = commands.getstatusoutput(cmd)
        if rc != 0:
            logger.info("xen_migrate_vm: migrate with ssl enabled failed, do failover(no ssl). vm('%s') -> tgt_srv('%s')" % \
                        (vm_path, tgt_host))
            if is_live:
                cmd = "xm migrate --live '%s' '%s'" % (vm_name, tgt_host)
            else:
                cmd = "xm migrate '%s' '%s'" % (vm_name, tgt_host)
            (rc, output) = commands.getstatusoutput(cmd)
            if rc != 0:
                raise Exception("xen_migrate_vm: migrate without ssl failed either.")

        logger.info("xen_migrate_vm: success. vm('%s') -> tgt_srv('%s')" % \
                (vm_path, tgt_host))
        return "success:target_host=%s" % tgt_host
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_migrate_vm: failed. vm('%s') -> tgt_srv('%s') =>%s" % \
                (vm_path, tgt_host, errmsg))
        return "failed:%s" % errmsg

xen_reset_vm = xen_reboot_vm

