#!/usr/bin/env python
#^#
# OVSXXenVMConfig.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import string
import re
import os
import logging
from shutil import copyfile
from random import randint, seed
from os.path import join, isdir, isfile, exists, dirname, basename, abspath, isabs
from OVSTimer import Timer
from OVSFileLock import FileLock
from OVSCommons import host_alive, parse_result, fmt_err_msg, run_cmd, exposed
from OVSConstants import HVM_IDE_MAX_NUM, HVM_SCSI_MAX_NUM

#-----------------------------------<>helper function<>----
def unmarshall_vif_cfg(vif_cfg):
    l = [e.split("=") for e in vif_cfg.split(",")]
    #vif_cfg='' -> l=[['',]]
    l = [map(string.strip, e) for e in l if len(e) == 2]
    if l: d = dict(l)
    else: d = {}
    return d

def marshall_vif_cfg(d):
    return ",".join(["=".join(t) for t in d.items()])

def xen_marshall_cfg(cfg, cfg_file):
    """
    write back vm cfg to config file
    """
    #-------------------------------------
    def _pretty_list(l):
        """
        return repr(l) with each element on a new line
        """
        if len(l) < 2: return repr(l)
        from StringIO import StringIO
        sio = StringIO()
        sio.write("[")
        for e in l:
            sio.write("%s,\n" % repr(e))
        sio.write("]")
        s = sio.getvalue()
        sio.close()
        return s
    #-------------------------------------
    lines = []
    for k, v in cfg.items():
        if v is None: continue
        if isinstance(v, list):
            lines.append("%s = %s\n" % (k, _pretty_list(v)))
        elif repr(v).startswith("<"): #e.g. import os
            continue
        elif isinstance(v, bool):
            if v: lines.append("%s = 1\n" % k)
            else: lines.append("%s = 0\n" % k)
        else:
            lines.append("%s = %s\n" % (k, repr(v)))
    lines.sort()

    #here a file lock is needed to avoid write contention
    lock_file = "%s.lock" % cfg_file
    fd = os.open(lock_file, os.O_CREAT | os.O_EXCL)
    os.close(fd)

    f = open(cfg_file, "w")
    f.writelines(lines)
    f.close()

    os.remove(lock_file)

def xen_unmarshall_cfg(cfg_file):
    """
    read in vm cfg file and parse it to a directory
    """
    if not isfile(cfg_file):
        raise Exception("vm config file ('%s') does not exist." % cfg_file)

    #FIXME: !!security!! and xen itself used execfile :(
    cfg = {}
    execfile(cfg_file, {}, cfg)

    return cfg


def xen_update_cfg_nics(vm_cfg_file, nics_enc):
    """
    update config file, point all the path to vm_path
    !!depreciated!! config nic will be removed
    """
    cfg = xen_unmarshall_cfg(vm_cfg_file)
    if "extra" in cfg:
        #the length limit for cmdline is?
        cfg["extra"] = re.sub(r"\s*ifcfgs=[^ ]*", "", cfg["extra"])
        cfg["extra"] += " ifcfgs=%s" % (nics_enc,)
    else:
        cfg["extra"] = "ifcfgs=%s" % (nics_enc,)
    xen_marshall_cfg(cfg, vm_cfg_file)

#----------------------------------------<> detect vm type (hvm, para) <>----
@exposed
def xen_get_vm_type(vm_path):
    """
    get vm type
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        if ("kernel" in cfg and "hvm" in cfg["kernel"]) or \
                ("builder" in cfg and cfg["builder"] == "hvm") or \
                ("device_model" in cfg and "qudomu" in cfg["device_model"]):
                    #hvm
                    type = "hvm"
        else:
            #para
            type = "para"

        logger.info("xen_get_vm_type: success. vm_path('%s') => type('%s')" % \
                (vm_path, type))
        return "success:type=%s" % type
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vm_type: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

from xen.lowlevel.xc import xc
from xen.xend.XendConstants import HVM_PARAM_CALLBACK_IRQ
@exposed
def xen_check_hvm_pvdriver(vm_path):
    """
    check whether a HVM's pvdriver  is ready.
    @return     "success" if pvdriver is ready,else "failed:<reason>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        #only for HVM
        rs = xen_get_vm_type(vm_path)
        if not rs.startswith("success"):
            raise Exception(rs)
        if parse_result(rs)["type"] != "hvm":
            raise Exception("%s is not a HVM machine" % vm_path)
        #only for a running HVM
        rs = xen_status_vm(vm_path)
        if not rs.startswith("success"):
            raise Exception("VM %s is not running" % vm_path)

        domid = xen_get_vm_id(vm_path)
        try:
            hvm_pvdrv = xc().hvm_get_param(int(domid), HVM_PARAM_CALLBACK_IRQ)
        except Exception, e:
            errmsg = fmt_err_msg(e)
            raise Exception("failed to get hvm pvdriver information: %s" % errmsg)
        if not hvm_pvdrv:
            raise Exception("Paravirtualized driver is not installed in this hvm guest.")

        logger.info("xen_check_hvm_pvdriver: success. vm_path('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_check_hvm_pvdriver: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg
#----------------------------------------<> update the config file <>-----
@exposed
def xen_config_boot_sequence(vm_path, boot, cdrom=None):
    """
    config the boot sequence to support mutiple boot systems

    @param boot     a string in ["c", "d", "n"]
                    c: hard disk, d: cdrom, n: network. man xmdomain.cfg for reference.
    @param cdrom     boot media if boot == "d" else ignored
    @return     "success" if success else "failed:<error message>"
    """
    #----------------------------------------------------

    logger = logging.getLogger("ovs.operation")
    try:
        boot = boot.lower()
        if not boot in ["c", "d", "n"]:
            raise Exception("boot system error!")

        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        # harddisk
        if boot == "c":
            cfg["boot"] = "c"
            cfg["bootloader"] = "/usr/bin/pygrub"
            if "bootargs" in cfg:
                del cfg["bootargs"]

        # cdrom
        elif boot == "d":
            rs = xen_get_vm_type(vm_path)
            if not rs.startswith("success"):
                raise Exception(rs)
            if parse_result(rs)["type"] != "hvm":
                raise Exception("attach cdrom to para vm is not supported yet!")
            cfg["boot"] = "d"

            if "disk" not in cfg: cfg["disk"] = []
            disks = []
            find_cdrom = False
            for disk in cfg["disk"]:
                backend_dev, frontend_dev, mode = disk.split(",")
                if "cdrom" in frontend_dev:
                    find_cdrom = True
                    if cdrom:
                        disk = "file:%s,%s,%s" % (cdrom, frontend_dev, mode)
                disks.append(disk)
            if not find_cdrom:
                if cdrom:
                    disks.append("file:%s,hdc:cdrom,r" % cdrom)
                else:#append an empty cdrom. we can config boot from cdrom even it is empty.
                    disks.append(",hdc:cdrom,r")

            cfg["disk"] = disks

        # PXE
        # for hvm pxeboot, if there's no ioemued vif, it will fail.
        # for pvm pxeboot, if there's a vif without mac address, random mac will be added to enable pxeboot.
        else:
            rs = xen_get_vm_type(vm_path)
            if not rs.startswith("success"):
                raise Exception(rs)
            # HVM
            if parse_result(rs)["type"] == "hvm":
                cfg["boot"] = "n"
                # make sure there is at least one vif without "type=netfront"
                if "vif" in cfg and len(cfg["vif"]) > 0:
                    pxebootable = False
                    for vif in cfg["vif"]:
                        d = unmarshall_vif_cfg(vif)
                        if "type" not in d or d["type"] == "ioemu":
                            pxebootable = True
                            break
                    if not pxebootable:
                        raise Exception("no pxe bootable vif found.")
                else:
                        raise Exception("no vif found.")
            # PVM
            else:
                if "boot" in cfg: del cfg["boot"]
                cfg["bootloader"] = "/usr/bin/pypxeboot"
                if "vif" in cfg and len(cfg["vif"]) > 0:
                    pxebootable = False
                    _idx = -1
                    for idx, vif in enumerate(cfg["vif"]):
                        d = unmarshall_vif_cfg(vif)
                        if "mac" in d:
                            pxebootable = True
                            _idx = idx
                            break
                    if not pxebootable:
                        d = marshall_vif_cfg(cfg["vif"][0])
                        d["mac"] = randomMAC()
                        cfg["vif"][0] = unmarshall_vif_cfg(d)
                        _idx = 0
                    cfg["bootargs"] = cfg["vif"][_idx]
                else:
                    raise Exception("no vif found.")

        xen_marshall_cfg(cfg, vm_cfg_file)
        logger.info("xen_config_boot_sequence: success. vm('%s'), boot='%s', cdrom('%s')" % \
                (vm_path, boot, cdrom))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_config_boot_sequence: failed. vm('%s'), boot='%s', cdrom('%s') => %s" % \
                (vm_path, boot, cdrom, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_correct_cfg(vm_cfg_file, vm_path):
    """
    update config file, point all the path to vm_path
    """
    logger = logging.getLogger("ovs.operation")
    vm_path = abspath(vm_path)
    try:
        #backup
        vm_cfg_file_orig = "%s.orig" % vm_cfg_file
        if not exists(vm_cfg_file_orig):
            copyfile(vm_cfg_file, vm_cfg_file_orig)

        cfg = xen_unmarshall_cfg(vm_cfg_file)

        if "name" in cfg:
            cfg["name"] = basename(vm_path)

        if "kernel" in cfg:
            if exists(join(vm_path, basename(cfg["kernel"]))):
                cfg["kernel"] = join(vm_path, basename(cfg["kernel"]))

        if "ramdisk" in cfg:
            if exists(join(vm_path, basename(cfg["ramdisk"]))):
                cfg["ramdisk"] = join(vm_path, basename(cfg["ramdisk"]))

        if "disk" in cfg:
            disks = []
            cdrom_disk = ""
            for disk in cfg["disk"]:
                #"backend-dev,frontend-dev,mode"
                backend_dev, frontend_dev, mode = disk.split(",")
                if "cdrom" in frontend_dev: #will deal with cdrom later
                    cdrom_disk = disk; continue
                if not backend_dev: #donnot know how to deal with it
                    continue
                dev_type, dev_id = backend_dev.rsplit(":", 1)
                if dev_type.strip() in ("file", "tap:aio"):
                    disk_path_orig = dev_id
                    if "shareddisk" in disk_path_orig.lower(): #share disk
                        disk_path = disk_path_orig
                        if exists(disk_path):
                            disks.append(disk)
                    else: #normal file based disk
                        disk_path = join(vm_path, basename(disk_path_orig))
                        disk = disk.replace(disk_path_orig, disk_path)
                        if exists(disk_path):
                            disks.append(disk)
                else: #other device => phy, etc
                    if exists(dev_id):
                        disks.append(disk)
            #deal with cdrom
            try:
                backend_dev, frontend_dev, mode = cdrom_disk.split(",")
                dev_type, dev_id = backend_dev.split(":")
                if not exists(dev_id):
                    raise Exception("Does not exist.")
                disks.append(cdrom_disk)
            except: #add default cdrom for hvm, see bug#4785
                rs = xen_get_vm_type(vm_path)
                if rs.startswith("success") and parse_result(rs)["type"] == "hvm":
                    #4 ide devices available due to historical limitation
                    ide = ["hd%s" % c for c in "abcd"]
                    l = [disk.split(",") for disk in disks]
                    for idx, (backend_dev, frontend_dev, mode) in enumerate(l):
                        if frontend_dev in ide:
                            ide.remove(frontend_dev)
                    if ide: #free IDE slot still available
                        if "hdc" in ide: #prefer hdc
                            disks.append(",hdc:cdrom,r")
                        else:
                            l.append(("", "%s:cdrom" % ide[0], "r"))
                            disks = [",".join(t) for t in l]
                    #else: no IDE slot for cdrom, cannot attach any cdrom device.

            cfg["disk"] = disks

        #bug#3853,3829, remove uuid, and let xend generate one for it
        #will now generate a uuid for the vm when register the vm to the pool
        #FIXME: one problem left, what if user manually change the uuid?
        if "uuid" not in cfg:
            from xen.xend import uuid as genuuid
            from random import seed
            seed()
            cfg["uuid"] = genuuid.createString()
        xen_marshall_cfg(cfg, vm_cfg_file)
        xen_correct_qos_cfg(vm_cfg_file)

        #bug#5406, config bootargs for PVM PXE boot
        if "bootloader" in cfg and cfg["bootloader"] == "/usr/bin/pypxeboot":
            rs = xen_config_boot_sequence(vm_path, 'n')
            if not rs.startswith("success"):
                raise Exception(rs)

        logger.info("xen_correct_cfg: success. vm('%s') => cfg_file('%s')" %\
                (vm_path, vm_cfg_file))
        return "success"
    except Exception, e:
        logger.error("xen_correct_cfg: failed. vm('%s'), cfg_file('%s') => %s" %\
                (vm_path, vm_cfg_file, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

@exposed
def xen_update_uuid(vm_cfg_file):
    logger = logging.getLogger("ovs.operation")
    try:
        cfg = xen_unmarshall_cfg(vm_cfg_file)
        from xen.xend import uuid as genuuid
        from random import seed
        seed()
        uuid = genuuid.createString()
        cfg["uuid"] = uuid
        xen_marshall_cfg(cfg, vm_cfg_file)

        logger.info("xen_update_uuid: success. vm('%s') uuid=%s" % (vm_cfg_file, uuid))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_update_uuid: failed. vm('%s')=> %s" % (vm_cfg_file, errmsg))
        return "failed:%s" % errmsg

def xen_correct_qos_cfg(vm_cfg_file):
    logger = logging.getLogger("ovs.operation")
    try:
        vm_cfg = xen_unmarshall_cfg(vm_cfg_file)
        #for the vbd
        if "disk_other_config" in vm_cfg and len(vm_cfg["disk_other_config"]) > 0:
            fronts = []
            if "disk" in vm_cfg and len(vm_cfg["disk"]) > 0:
                fronts = [disk.split(",")[1].strip() for disk in vm_cfg["disk"]]
            for disk in vm_cfg["disk_other_config"]:
                if len(disk) == 0 or disk[0] not in fronts:
                    vm_cfg["disk_other_config"].remove(disk)
         #for the vif
        if "vif_other_config" in vm_cfg and len(vm_cfg["vif_other_config"]) > 0:
            macs = []
            if "vif" in vm_cfg and len(vm_cfg["vif"]) > 0:
                for vif in vm_cfg["vif"]:
                    d = unmarshall_vif_cfg(vif)
                    if "mac" in d:
                        macs.append(d["mac"].lower())
            for vif in vm_cfg["vif_other_config"]:
                if len(vif) == 0 or vif[0].lower() not in macs:
                    vm_cfg["vif_other_config"].remove(vif)
        xen_marshall_cfg(vm_cfg, vm_cfg_file)
        logger.info("xen_correct_qos_cfg: vm_cfg('%s')=>success." % (vm_cfg_file))
    except Exception, e:
        logger.error("xen_correct_qos_cfg:vm_cfg('%s')=>%s" % (vm_cfg_file, fmt_err_msg(e)))
        #raise

#----------------------------------------<> vm memory <>-----
@exposed
def xen_set_memory(vm_path, memory=-1, maxmem=-1, scope="cfg"):
    """
    set vm memory
    @param memory   memory in M, if specified as -1, it will be ignored.
    @param maxmem   max memory in M, if specified as -1, it will be ignored.
    @param scope    in [cfg, mem, both]
                    cfg, persist config to file
                    mem, config using xm command
                    both, cfg & mem
    """
    logger = logging.getLogger("ovs.operation")
    vm_cfg_file = join(vm_path, "vm.cfg")
    try:
        vm_name = basename(vm_path)
        memory, maxmem = int(memory), int(maxmem)

        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))

        if scope_cfg:
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            if memory != -1:
                cfg["memory"] = memory
            if maxmem != -1:
                cfg["maxmem"] = maxmem
            xen_marshall_cfg(cfg, vm_cfg_file)
        if scope_mem:
            try:
                if memory != -1:
                    cmd = "xm mem-set %s %d" % (vm_name, memory)
                    run_cmd(cmd)
                if maxmem != -1:
                    cmd = "xm mem-max %s %d" % (vm_name, maxmem)
                    run_cmd(cmd)
            except Exception, e:
                if scope_cfg: #ok, if cannot set dynamically
                    logger.warning("xen_set_memory:vm('%s') dynamic set memory:memory=%d maxmem=%d failed. =>%s" % \
                        (vm_path, memory, maxmem, fmt_err_msg(e)))
                else: raise e

        logger.info("xen_set_memory:vm('%s') => memory=%dM, maxmem=%dM" % \
            (vm_path, memory, maxmem))
        #compatible with ovs-agent-2.2-15
        #return "success:memory=%d;maxmem=%d" % (memory, maxmem)
        return "success:memory=%d;maxmem=%d;mem_size=%d" % (memory, maxmem, memory)
    except Exception, e:
        logger.error("xen_set_memory:vm('%s')=>%s" % (vm_path, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

@exposed
def xen_get_memory(vm_path, scope='cfg'):
    """
    get vm memory (in MB)
    """
    logger = logging.getLogger("ovs.query")
    vm_cfg_file = join(vm_path, "vm.cfg")
    try:
        vm_name = basename(vm_path)

        if scope == 'cfg':
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            memory = int(cfg["memory"])
            maxmem = memory
            if "maxmem" in cfg:
                maxmem = int(cfg["maxmem"])
        elif scope == 'mem':
            cmd = "xm list --long %s | grep -o -P '(maxmem|(?<!_)memory)\s+\d+'" % vm_name
            output = run_cmd(cmd)
            d = {}
            for line in output.splitlines():
                parts = line.split()
                d[parts[0]] = int(parts[1])
            memory, maxmem = d["memory"], d["maxmem"]

        logger.info("xen_get_memory:vm('%s') => memory=%dM maxmem=%dM" % (vm_path, memory, maxmem))
        #version compatible
        #return "success:memory=%d;maxmem=%d" % (memory, maxmem)
        return "success:memory=%d;maxmem=%d;mem_size=%d" % (memory, maxmem, memory)
    except Exception, e:
        logger.error("xen_get_memory:vm('%s')=>%s" % (vm_path, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

#----------------------------------------<> vm vcpus <>-----
@exposed
def xen_set_vcpus(vm_path, vcpus=-1, cpus="NONE", vcpu=-1, cpu=-1, scope="cfg"):
    """
    set VM vcpus, cpus, cpu, refer to 'man xmdomain.cfg'
    @param vcpus    if -1, will be ignored.
    @param cpus     if "NONE", will be ignored, if "" or "any", will leave xen to pick
    @param vcpu     specify the <vcpu>th vcpu for pin, if -1, will be ignored.
    @param cpu      if -1, will be ignored.
    @param scope    in [cfg, mem, both],
                    cfg, persist setting to cfg file
                    mem, only change the setting dynamically
                    both, cfg & mem
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vcpus, cpus, vcpu, cpu = int(vcpus), str(cpus), int(vcpu), int(cpu)

        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))

        if scope_cfg:
            cfg_file = join(vm_path, "vm.cfg")
            cfg = xen_unmarshall_cfg(cfg_file)
            if vcpus != -1: cfg["vcpus"] = vcpus
            if cpus != "NONE":
                if cpus == "any": cpus = ""
                cfg["cpus"] = cpus
            if cpu != -1: cfg["cpu"] = cpu
            xen_marshall_cfg(cfg, cfg_file)

        if scope_mem:
            vm_name = basename(vm_path)
            if vcpus != -1:
                #NOTE: can only decrease cpu number, kind of disable a cpu
                cmd = "xm vcpu-set %s %d" % (vm_name, vcpus)
                run_cmd(cmd)
            if vcpu != -1: #the <vcpu>'th vcpu
                cmd = "xm vcpu-pin %d %s" % (vm_name, vcpu, cpus)
                run_cmd(cmd)

        logger.info("xen_set_vcpus:success. vm('%s') => vcpus=%d, cpus=%s, vcpu=%d, cpu=%d" % \
                (vm_path, vcpus, cpus, vcpu, cpu))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vcpus:failed. vm('%s') vcpus=%s, cpus=%s, vcpu=%s, cpu=%s =>%s" % \
            (vm_path, repr(vcpus), repr(cpus), repr(vcpu), repr(cpu), errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_get_vcpus(vm_path, scope="cfg"):
    """
    get vm vcpu number and vcpu to cpu mapping
    @param scope    in [cfg, mem]
                    cfg, get vcpu configuration from cfg file
                    mem, get runtime vcpu configuration
    @return "success:vcpus=<vcpu num>;vcpu_list=<vcpu cpu mapping>" if success else "error message"
            vcpu cpu mapping, e.g.
            vcpus_list=<vcpu>:<cpu> <vcpu>:<cpu> ...
            vcpus_list=0:0 1:1 3:0,1
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_name = basename(vm_path)

        if scope == "cfg":
            vm_cfg_file = join(vm_path, "vm.cfg")
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            if "vcpus" in cfg:
                vcpus = int(cfg["vcpus"])
            else: #default
                vcpus = 1

        vcpu_list = []
        if scope == "mem":
            cmd = "xm vcpu-list %s" % vm_name
            output = run_cmd(cmd)
            lines = output.splitlines()
            #Name                                ID  VCPU   CPU State   Time(s) CPU Affinity
            #titles = lines[0].split()
            idx_vcpu, idx_cpu = 2, 6
            vcpus = 0
            for line in lines[1:]:
                parts = line.split()
                vcpu_list.append("%s:%s" % (parts[idx_vcpu], parts[idx_cpu]))
                vcpus += 1
        vcpu_list = " ".join(vcpu_list)

        logger.info("xen_get_vcpus:vm('%s') => vcpus=%d vcpu_list=%s" % (vm_path, vcpus, vcpu_list))
        return "success:vcpus=%d;vcpu_list=%s" % (vcpus, vcpu_list)
    except Exception, e:
        logger.error("xen_get_vcpus:vm('%s')=>%s" % (vm_path, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

#----------------------------<> virtual disk <>----
@exposed
def xen_get_vdisks(vm_path, scope="cfg"):
    """
    get vm vdisks from config file
    @return 'success:vdisks=<path_to_file1>,<path_to_file2>' if success else 'failed:<error message>'
    """
    logger = logging.getLogger("ovs.query")
    vm_cfg_file = join(vm_path, "vm.cfg")
    try:
        vm_name = basename(vm_path)

        files = []
        if scope == "cfg":
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            vdisks = cfg["disk"]
            for vdisk in vdisks:
                filename = vdisk.split(",")[0].split(":")[-1].strip()
                if filename: #possible ',hdc:cdrom,r'
                    files.append(filename)

        if scope == "mem":
            cmd = "xm block-list %s" % vm_name
            output = run_cmd(cmd)
            for line in output.splitlines():
                path = line.split()[-1]
                if path.startswith("/"):
                    cmd = "xenstore-read %s/params" % path
                    v = run_cmd(cmd)
                    filename = v.strip("\"'\n").split(":")[-1]
                    files.append(filename)

        file_str = ','.join(files)
        logger.info("xen_get_vdisks:vm('%s') => vdisks=%s scope=%s" % (vm_path, file_str, scope))
        return "success:vdisks=%s" % file_str
    except Exception, e:
        logger.error("xen_get_vdisks:vm('%s') => %s" % (vm_path, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

@exposed
def xen_get_avail_disk_slot(vm_path):
    """
    @return     "success:free_slot=<val>[;free_ide_slot=<val>;free_scsi_slot=<val>]"
                if pvm, free_slot = -1. no hd or sd will be returned
                if hvm, free_slot = free_ide_slot + free_scsi_slot
    """
    logger = logging.getLogger("ovs.operation")
    try:
        rs = xen_get_vm_type(vm_path)
        if rs.startswith("success"):
            d = parse_result(rs)
        else:
            raise Exception(rs)
        if d["type"] == "para":
            rs = "success:free_slot=-1"
        else: #hvm
            cfg_file = join(vm_path, "vm.cfg")
            cfg = xen_unmarshall_cfg(cfg_file)
            if "disk" in cfg:
                fe_devs = [ disk.split(",")[1] for disk in cfg["disk"] ]
                #reserve one slot for cdrom
                free_ide_slot = HVM_IDE_MAX_NUM - len([fe for fe in fe_devs if fe.startswith("hd") and not "cdrom" in fe]) - 1
                free_scsi_slot = HVM_SCSI_MAX_NUM - len([fe for fe in fe_devs if fe.startswith("sd")])
            else:
                free_ide_slot = HVM_IDE_MAX_NUM - 1 #one reserved for cdrom
                free_scsi_slot = HVM_SCSI_MAX_NUM
            free_slot = free_ide_slot + free_scsi_slot
            rs = "success:free_slot=%d;free_ide_num=%d;free_scsi_num=%d" % (free_slot, free_ide_slot, free_scsi_slot)

        logger.info("xen_get_avail_disk_slot: vm('%s') => %s" % (vm_path, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_avail_disk_slot: failed. vm('%s') => %s" % (vm_path, errmsg))
        return "failed: %s" % errmsg

@exposed
def xen_add_disk(vm_path, disk_path, sharable=False, fe_dev="auto", mode="w", scope="cfg", fe_dev_prefix="", position=-1):
    """
    add disk to vm config file
    @param fe_dev           if "auto", choose automatically by agent
    @param fe_dev_prefix    in [hd, sd, xvd, ...]. based on disk type: IDE->hd, SCSI->sd, Xen Virtual Block->xvd ...
    @param position         prefered position in the disk list, ie. you can maintain the sequence.
    @return when succeed returns "success:fe_name=<front end name>", when fail returns "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    # --------------------------------------------------------
    def next_fe_dev(fe_devs, fe_dev, fe_dev_prefix):
    #def next_fe_dev(fe_devs, fe_dev):
    #NOTE: why fe_dev_prefix is necessary here?
    #http://docs.python.org/ref/naming.html
    #   "If a name is bound in a block, it is a local variable of that block.
    #If a name is bound at the module level, it is a global variable.
    #(The variables of the module code block are local and global.)
    #If a variable is used in a code block but not defined there, it is a free variable.
    #When a name is not found at all, a NameError exception is raised.
    #If the name refers to a local variable that has not been bound, a UnboundLocalError exception is raised."
        rs = xen_get_vm_type(vm_path)
        if not rs.startswith("success"):
            raise Exception("Failed to get VM type for %s: %s" % (vm_path, rs))
        is_hvm =  (parse_result(rs)["type"] == "hvm")

        if fe_dev != 'auto':
            if fe_dev in fe_devs:
                raise Exception("Front end device name '%s' has been taken." % fe_dev)
            elif is_hvm and not fe_dev.startswith('hd') and not fe_dev.startswith('sd'):
                raise Exception("HVM front end device name '%s' is invalid, must start with 'hd' or 'sd'." % fe_dev)

        #http://wiki.us.oracle.com/calpg/OracleVM/Limitations
        #hvm currently only support at most: 4 IDE  + 7 SCSI
        #is the same with HVM + PV driver?
        if is_hvm:
            ide_num = len([ fe for fe in fe_devs if fe.startswith('hd') and not fe.startswith('hdc')])
            scsi_num = len([ fe for fe in fe_devs if fe.startswith('sd') ])
            no_free_ide = (ide_num >= HVM_IDE_MAX_NUM - 1) #one reserved for cdrom
            no_free_scsi = (scsi_num >= HVM_SCSI_MAX_NUM)
            if fe_dev != 'auto':
                if fe_dev.startswith('hd') and no_free_ide:
                    raise Exception("currently only up to %d IDE disks can be attached to a hvm vm." % \
                            (HVM_IDE_MAX_NUM -1, ))
                elif fe_dev.startswith('sd') and no_free_scsi:
                    raise Exception("currently only up to %d SCSI disks can be attached to a hvm vm." % \
                            (HVM_SCSI_MAX_NUM -1, ))
            else: #choose an available disk slot
                if not fe_dev_prefix:
                    if not no_free_ide:
                        fe_dev_prefix = 'hd'
                    elif not no_free_scsi:
                        fe_dev_prefix = 'sd'
                    else:
                        raise Exception("currently only up to %d IDE disks and %d SCSI disks can be attached to a hvm vm." % \
                            (HVM_IDE_MAX_NUM -1, HVM_SCSI_MAX_NUM))
                else: #user specified the disk type
                    if (fe_dev_prefix == "hd" and no_free_ide) or (fe_dev_prefix == "sd" and no_free_scsi):
                        raise Exception("currently only up to %d IDE disks and %d SCSI disks can be attached to a hvm vm." % \
                            (HVM_IDE_MAX_NUM -1, HVM_SCSI_MAX_NUM))
                for c in string.ascii_lowercase:
                    fe_dev = "%s%s" % (fe_dev_prefix, c)
                    #prefer to leave hdc for cdrom
                    if fe_dev not in fe_devs and fe_dev != 'hdc':
                        break
        else: #para vm
            if fe_dev == "auto":
                if not fe_dev_prefix:
                    if not fe_devs:
                        fe_dev_prefix = "xvd"
                    else:
                        last_dev = re.sub(r"[0-9]+$", "", fe_devs[-1]) # e.g. xvda1 -> xvda
                        fe_dev_prefix = last_dev[:-1] #e.g. xvda -> xvd
                for c in string.ascii_lowercase:
                    fe_dev = "%s%s" % (fe_dev_prefix, c)
                    if fe_dev not in fe_devs:
                        break
                #FIXME:since always add new disk not new partition, so in fact no more than 26 disks can be attached
                if fe_dev in fe_devs:
                    raise Exception("cannot find disk slot for the disk.")

        return fe_dev
    # --------------------------------------------------------

    vm_cfg_file = join(vm_path, "vm.cfg")
    try:
        vm_name = basename(vm_path)

        logger.debug("xen_add_disk: vm('%s'), disk_path=%s, sharable=%s, fe_dev=%s, mode=%s, scope=%s" % \
            (vm_path, disk_path, repr(sharable), fe_dev, mode, scope))

        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))

        if scope_cfg:
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            #check duplicate
            for disk in cfg["disk"]:
                path = disk.split(",")[0].split(":")[-1]
                if disk_path == path:
                    raise Exception("disk already attached.")

            #file:/path/to/file.iso,hdc:cdrom,r -> hdc
            #file:/path/to/file.img,hda,w -> hda
            fe_devs = [ disk.split(",")[1].split(":")[0]
                    for disk in cfg["disk"] ]
            tgt_fe_dev = next_fe_dev(fe_devs, fe_dev, fe_dev_prefix)
            #use file io, tap:aio is not stable enough yet
            be_dev = "file:%s" % disk_path
            disk = "%s,%s,%s" % (be_dev, tgt_fe_dev, mode)
            if sharable:
                disk = "%s!" % disk
            if position >= 0 and position < len(cfg["disk"]):
                cfg["disk"].insert(position, disk)
            else:
                cfg["disk"].append(disk)

            xen_marshall_cfg(cfg, vm_cfg_file)
            logger.info("xen_add_disk: success. vm('%s') disk('%s') scope=cfg." % \
                (vm_path, disk_path))

        if scope_mem:
            files = []
            fe_devs = []
            cmd = "xm block-list %s" % vm_name
            output = run_cmd(cmd)
            for line in output.splitlines():
                path = line.split()[-1]
                if path.startswith("/"):
                    #file
                    cmd = "xenstore-read %s/params" % path
                    v = run_cmd(cmd)
                    filename = v.strip("\"'\n").split(":")[-1]
                    files.append(filename)
                    #dev
                    cmd = "xenstore-read %s/dev" % path
                    v = run_cmd(cmd)
                    dev = v.strip("\"'\n")
                    fe_devs.append(dev)

            if disk_path in files:
                raise Exception("disk already attached.")
            tgt_fe_dev = next_fe_dev(fe_devs, fe_dev, fe_dev_prefix)
            be_dev = "file:%s" % disk_path
            cmd = "xm block-attach %s %s %s %s" % (vm_name, be_dev, tgt_fe_dev, mode)
            logger.debug("xen_add_disk: %s" % cmd)
            run_cmd(cmd)
            logger.info("xen_add_disk: success. vm('%s') disk('%s') scope=mem." % \
                (vm_path, disk_path))

        return "success:fe_name=%s" % tgt_fe_dev
    except Exception, e:
        logger.error("xen_add_disk: error. vm('%s') disk('%s')=>%s" % \
                (vm_path, disk_path, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

@exposed
def xen_remove_disk(vm_path, disk_path, scope="cfg"):
    """
    remove disk from config file
    @return when succeed returns "success:position=<disk original position>" failed returns "failed:<error message>", when returns "success:-1" means:
    1. disk had already been removed.
    2. scope='mem', in which case there is no position information.
    """
    logger = logging.getLogger("ovs.operation")
    vm_cfg_file = join(vm_path, "vm.cfg")
    try:
        vm_name = basename(vm_path)

        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))
        tgt_idx = -1

        if scope_cfg:
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            for (idx, disk) in enumerate(cfg["disk"]):
                path = disk.split(":")[-1].split(",")[0]
                if disk_path == path:
                    tgt_idx = idx
                    break
            if tgt_idx != -1:
                del cfg["disk"][tgt_idx]
                xen_marshall_cfg(cfg, vm_cfg_file)
                xen_correct_qos_cfg(vm_cfg_file)
                logger.info("xen_remove_disk: success. vm('%s') disk('%s') scope=%s." % \
                        (vm_path, disk_path, scope))
            else:
                logger.warning("xen_remove_disk: success. warning! disk('%s') does not exist in vm('%s') config." % \
                        (disk_path, vm_path, ))

        if scope_mem:
            vbd_id = "-1"
            cmd = "xm block-list %s" % vm_name
            output = run_cmd(cmd)
            for line in output.splitlines():
                path = line.split()[-1]
                if path.startswith("/"):
                    cmd = "xenstore-read %s/params" % path
                    v = run_cmd(cmd)
                    filename = v.strip("\"\n'").split(":")[-1]
                    if filename == disk_path:
                        vbd_id = path.split("/")[-1]

            if vbd_id != "-1":
                #FIXME:
                #--force may cause wired behavior, for e.g.
                #*) vbd_id still exists in xenstore xenstore-ls /local/domain/0 | grep <vbd_id>
                #*) xm create/shutdown fail for this vm
                #*) cannot attach this disk again ...
                # in a word, use online detach with caution
                #cmd = "xm block-detach %s %s --force" % (vm_name, vbd_id)
                cmd = "xm block-detach %s %s" % (vm_name, vbd_id)
                run_cmd(cmd)
                logger.info("xen_remove_disk: success. vm('%s') disk('%s') scope=%s." % \
                        (vm_path, disk_path, scope))
            else:
                logger.warning("xen_remove_disk: success. warning! disk('%s') does not exist in xenstore." % disk_path )

        return "success:position=%d" % tgt_idx
    except Exception, e:
        logger.error("xen_remove_disk: error. vm('%s') disk('%s') => %s" % \
            (vm_path, disk_path, fmt_err_msg(e)))
        return "failed:%s" % fmt_err_msg(e)

@exposed
def xen_remove_shared_disk(vm_path):
    """
    remove shared disk from config file.

    mainly used when deploy a vm and donnot want to have the shared disk in the dest cfg file
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        disks = []
        for disk in cfg["disk"]:
            try:
                be_dev, fe_dev, mode = disk.split(",")
                if "!" in mode: continue
            except Exception, e:
                logger.warning("xen_remove_shared_disk: vm('%s') disk('%s')=> %s" % \
                        (vm_path, disk, fmt_err_msg(e)))
            disks.append(disk)
        cfg["disk"] = disks

        xen_marshall_cfg(cfg, vm_cfg_file)
        xen_correct_qos_cfg(vm_cfg_file)
        logger.info("xen_remove_shared_disk: success. vm('%s')." %  vm_path)

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_remove_shared_disk: error. vm('%s') => %s" % \
            (vm_path, errmsg))
        return "failed:%s" % errmsg


#----------------------------<> virtual network interface <>----
@exposed
def xen_get_vif_num(vm_path):
    """
    get vif num
    @return "success:vif_num=<val>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)
        if "vif" in cfg: vif = len(cfg["vif"])
        else: vif = 0
        logger.info("xen_get_vif_num: success. vm('%s') get vif_num=%d" % (vm_path, vif))
        return "success:vif=%d" % vif
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vif_num: failed. vm('%s') =>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_del_vif(vm_path, macs, scope="cfg"):
    """
    del vif from the cfg file

    @param macs     mac address delimited by ',',
                    if is "" then remove all vifs
    @return "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm_name = basename(vm_path)
        macs_lst = macs.upper().split(",")

        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))

        if scope_cfg and "vif" in cfg:
            vm_cfg_file = join(vm_path, "vm.cfg")
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            vifs = []
            if not macs_list: #clear vif entries if no mac specified
                cfg["vif"] = vifs
                logger.info("xen_del_vif: remove all vif for vm('%s')" % vm_path)

            for vif in cfg["vif"]:
                to_remove = False
                for mac in macs_lst:
                    if vif.upper().strip().find(mac.uppper().strip()) != -1:
                        to_remove = True
                        break
                if not to_remove: vifs.append(vif)
            cfg["vif"] = vifs
            xen_marshall_cfg(cfg, vm_cfg_file)
            xen_correct_qos_cfg(vm_cfg_file)

        if scope_mem:
            cmd = "xm network-list %s" % vm_name
            output = run_cmd(cmd)
            idx_idx, idx_mac = 0, 2
            vifs = {}
            for line in output.splitlines()[1:]:
                parts = line.split()
                idx, mac = parts[idx_idx], parts[idx_mac]
                vifs[mac.upper()] = idx
            for _mac in macs_lst:
                for mac in vifs:
                    if _mac == mac:
                        #FIXME: when to use/not to use --force?
                        #cmd = "xm network-detach --force %s %s" % (vm_name, vifs[mac])
                        cmd = "xm network-detach --force %s %s" % (vm_name, vifs[mac])
                        run_cmd(cmd)
                        break

        logger.info("xen_del_vif: success. vm_path('%s'), macs('%s')" % \
                (vm_path, macs))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_del_vif: failed. vm_path('%s'), macs('%s') => %s" % \
                (vm_path, macs, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_set_vifs(vm_path, vif_cfg, mode="setall", scope="cfg"):
    """
    set vif to vm config file

    @param vif_cfg  vif_cfg ::= vif_attr{;vif_attr}
                    vif_attr ::= [<mac>,[<bridge>,[<type>]]]
                    type in [netfront, ioemu, ALL], if type=ALL, type will be cleared
                    if type='', type will be default to what already there
                    e.g.
                    "00:15:60:C3:07:A1,xenbr0;00:15:60:C3:07:A2,xenbr2;,xenbr1,netfront"
    @param mode     in [setall, remove, append, update]
                    setall, remove original
                    remove, remove by mac, bridge & type will be ignored
                    append, if not already attached, append; else update
                    update, if already attaced, append; else append
                    remove & update is based on mac, so mac in vif_cfg is mandatory
    @return "success" if success else "failed:<error message>"
    """
    #FIXME: should we split this function to 4? [set|add|remove|update]_vif()?
    logger = logging.getLogger("ovs.operation")

    #-------------------------------------------------------------------
    def restore_vif_extra_attr(vif_new, vif_orig):
        #inplace update vif_new
        for idx, vif in enumerate(vif_new):
            d = unmarshall_vif_cfg(vif)
            if len(vif_orig) > idx:
                _d = unmarshall_vif_cfg(vif_orig[idx])
                _d.update(d)
                vif_new[idx] = marshall_vif_cfg(_d)

    def get_mac_orig_new_map(vif_orig, vif_new):
        rs = {}
        for idx, vif in enumerate(vif_orig):
            mac_orig, mac_new = "", ""
            d = unmarshall_vif_cfg(vif)
            if "mac" in d:
                mac_orig = d["mac"]
            if len(vif_new) > idx:
                d = unmarshall_vif_cfg(vif_new[idx])
                if "mac" in d:
                    mac_new = d["mac"]
            if mac_orig and mac_new:
                rs[mac_orig.upper()] = mac_new
        return rs

    def remap_vif_other_config(vif_other_config, mac_orig_new_map):
        #inplace update vif_other_config
        l = vif_other_config[:]
        #clear vif_other_config
        while vif_other_config:
            vif_other_config.pop()
        for (mac_orig, key, val) in l:
            mac_new = mac_orig_new_map.get(mac_orig.upper())
            if mac_new:
                vif_other_config.append([mac_new, key, val])
    #-------------------------------------------------------------------
    try:
        vm_name = basename(vm_path)

        l_vif_cfg = vif_cfg.split(";")
        if len(l_vif_cfg) > 8:
            #http://wiki.xensource.com/xenwiki/XenFaq#head-9896478cf65a16f43ab4fb066f74c0e0d67a16ac
            raise Exception("cannot add more than 8 interfaces.")
        l_vif_cfg = [("%s%s" % (s, "," * (2 - s.count(',')))).split(",") for s in l_vif_cfg]
        l_vif_cfg = [vif_attr[:3] for vif_attr in l_vif_cfg] #so that future field won't break it
        l_vif_cfg = [(mac.strip().upper(), bridge.strip(), _type.strip()) for (mac, bridge, _type) in l_vif_cfg]

        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))

        if scope_cfg:
            vm_cfg_file = join(vm_path, "vm.cfg")
            cfg = xen_unmarshall_cfg(vm_cfg_file)

            if "vif" not in cfg:
                cfg["vif"] = []

            #copy of original setting
            vif_orig = cfg["vif"][:]
            if "vif_other_config" not in cfg:
                cfg["vif_other_config"] = []
            vif_other_config_orig = cfg["vif_other_config"]

            #find device type and save in new_type
            orig_type = ""
            if mode in ["append", "update", "remove"]:
                for vif in cfg["vif"]:
                    if not vif.strip(): continue
                    lst = [map(string.strip, attr.split("=")) for attr in vif.split(",")]
                    d = dict(lst)
                    if orig_type == "" and "type" in d and d["type"] != "":
                        orig_type = d["type"]
                        break
            new_type = ""
            if mode in ["append", "update", "setall"]:
                for vif in l_vif_cfg:
                    if vif[2].strip() != "":
                        new_type = vif[2]
                if new_type == "" and orig_type != "":
                    new_type = orig_type
            else: #mode = 'remove'
                new_type = orig_type

            if mode == "setall":
                cfg["vif"] = []
                for (mac, bridge, _type) in l_vif_cfg:
                    vif = {}
                    if mac: vif["mac"] = mac
                    if bridge: vif["bridge"] = bridge
                    if new_type: vif["type"] = new_type #not using _type, for driver type must be the same
                    vif_val = ",".join([ "%s=%s" % (attr, value) for (attr, value) in vif.items()])
                    cfg["vif"].append(vif_val)
                restore_vif_extra_attr(cfg["vif"], vif_orig)
                _map = get_mac_orig_new_map(vif_orig, cfg["vif"])
                remap_vif_other_config(cfg["vif_other_config"], _map)

            if mode == "append":
                for (mac, bridge, _type) in l_vif_cfg:
                    vif = {}
                    if bridge: vif["bridge"] = bridge
                    if new_type: vif["type"] = new_type #not using _type, for driver type must be the same
                    if mac:
                        vif["mac"] = mac
                        _idx = -1
                        for (idx, _vif) in enumerate(cfg["vif"]):
                            if mac.upper() in _vif.upper():
                                _idx = idx
                                break
                        vif_val = ",".join([ "%s=%s" % (attr, value) for (attr, value) in vif.items()])
                        if _idx != -1: #if already there, append -> update
                            cfg["vif"][_idx] = vif_val
                        else:
                            cfg["vif"].append(vif_val)
                    else: #no mac is specified
                        cfg["vif"].append(vif_val)

            if mode == "update":
                for (mac, bridge, _type) in l_vif_cfg:
                    vif = {}
                    if bridge: vif["bridge"] = bridge
                    if new_type: vif["type"] = new_type #not using _type, for driver type must be the same
                    if mac: #update based on mac, no mac no update
                        vif["mac"] = mac
                        _idx = -1
                        for (idx, _vif) in enumerate(cfg["vif"]):
                            if mac.upper() in _vif.upper():
                                _idx = idx
                                break
                        vif_val = ",".join([ "%s=%s" % (attr, value) for (attr, value) in vif.items()])
                        if _idx != -1:
                            cfg["vif"][_idx] = vif_val
                        else: #if not already there, update -> append
                            cfg["vif"].append(vif_val)
                restore_vif_extra_attr(cfg["vif"], vif_orig)
                _map = get_mac_orig_new_map(vif_orig, cfg["vif"])
                remap_vif_other_config(cfg["vif_other_config"], _map)

            if mode == "remove":
                for (mac, bridge, _type) in l_vif_cfg:
                    if mac: #remove based on mac, no mac no remove
                        _idx = -1
                        for (idx, _vif) in enumerate(cfg["vif"]):
                            if mac.upper() in _vif.upper():
                                _idx = idx
                                break
                        if _idx != -1:
                            del cfg["vif"][_idx]

            #make driver type consistent
            if new_type:
                for idx, vif_val in enumerate(cfg["vif"]):
                    lst = [map(string.strip, attr.split("=")) for attr in vif_val.split(",")]
                    try: vif = dict(lst)
                    except: vif = {}
                    if new_type != 'ALL':
                        vif["type"] = new_type
                    else: #clear type attr if "ALL" is specified
                        if "type" in vif: del vif["type"]
                    vif_val = ",".join([ "%s=%s" % (attr, value) for (attr, value) in vif.items()])
                    cfg["vif"][idx] = vif_val

            xen_marshall_cfg(cfg, vm_cfg_file)
            xen_correct_qos_cfg(vm_cfg_file)
            logger.info("xen_set_vifs: success. vm('%s') set vif_cfg=%s, mode=%s, scope=%s" % \
                    (vm_path, vif_cfg, mode, scope))

        if scope_mem:
#xm network-list cannot get mac
#            cmd = "xm network-list %s" % vm_name
#            output = run_cmd(cmd)
#            idx_idx, idx_mac = 0, 2
#            vifs = {}
#            for line in output.splitlines()[1:]:
#                parts = line.split()
#                idx, mac = parts[idx_idx], parts[idx_mac]
#                vifs[mac.upper()] = idx

            #get current attached vif list
            vifs = {}
            try:
                cmd = "xm domid %s" % vm_name
                domid = run_cmd(cmd)
                cmd = "xenstore-ls /local/domain/0/backend/vif/%s" % domid
                output = run_cmd(cmd)
                for line in output.splitlines():
                    if not line.startswith(" "):
                        vif_id = line.split("=")[0].strip()
                    if line.startswith(" mac"):
                        mac = line.split("=")[1].strip(' "\'').upper()
                        vifs[mac] = vif_id
            except Exception, e: #if no vif, it will exception
                logger.warning("xen_set_vifs: get vm('%s') vifs from xenstore => %s" % \
                        (vm_name, fmt_err_msg(e)))

            if mode == "setall":
                for mac, bridge, _type in l_vif_cfg: #if not attached, attach it
                    cmd = "xm network-attach %s" % vm_name
                    if mac:
                        if mac.upper() in vifs: continue #already attached
                        cmd = "%s mac=%s" % (cmd, mac)
                    if bridge: cmd = "%s bridge=%s" % (cmd, bridge)
                    if _type: cmd = "%s type=%s" % (cmd, _type)
                    logger.debug("xen_set_vifs: mode=%s cmd=%s" % (mode, cmd))
                    run_cmd(cmd)
                for _mac in vifs: #if not in vif_cfg, remove it, so mac MUST be specified in vif_cfg
                    _remove = True
                    for mac, bridge, _type in l_vif_cfg:
                        if _mac == mac:
                            _remove = False
                            break
                    if _remove:
                        #FIXME: when to use/ not to use --force
                        cmd = "xm network-detach %s %s --force" % (vm_name, vifs[_mac])
                        logger.debug("xen_set_vifs: mode=%s cmd=%s" % (mode, cmd))
                        run_cmd(cmd)

            if mode == "append":
                for mac, bridge, _type in l_vif_cfg:
                    cmd = "xm network-attach %s" % vm_name
                    if mac:
                        if mac.upper() in vifs: continue #already attached no append
                        cmd = "%s mac=%s" % (cmd, mac)
                    if bridge: cmd = "%s bridge=%s" % (cmd, bridge)
                    if _type: cmd = "%s type=%s" % (cmd, _type)
                    logger.debug("xen_set_vifs: mode=%s cmd=%s" % (mode, cmd))
                    run_cmd(cmd)

            if mode == "update":
                #currently not support for update already attached vif
                logger.warning("xen_set_vifs: online upate is not support yet!")

            if mode == "remove":
                for mac, bridge, _type in l_vif_cfg:
                    _remove = False
                    for _mac in vifs: #remove only if _mac in vif_cfg
                        if _mac == mac:
                            _remove = True
                            break
                    if _remove:
                        #FIXME: when to use/ not to use --force
                        cmd = "xm network-detach %s %s --force" % (vm_name, vifs[_mac])
                        logger.debug("xen_set_vifs: mode=%s cmd=%s" % (mode, cmd))
                        run_cmd(cmd)

            logger.info("xen_set_vifs: success. vm('%s') set vif_cfg=%s, mode=%s, scope=%s" % \
                    (vm_path, vif_cfg, mode, scope))

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vifs: error. vm('%s') vif_cfg=%s, mode=%s, scope=%s =>%s" % \
                (vm_path, vif_cfg, mode, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_get_vifs(vm_path, scope="cfg"):
    """
    get vif from vm config file
    @return    vif_cfg ::= vif_attr{;vif_attr}
               vif_attr ::= <vif>=<mac>,<bridge>,<type>
               e.g.
                "vif0=00:15:60:C3:07:A1,xenbr0,netfront;vif1=00:15:60:C3:07:A2,xenbr2,;vif3=,xenbr1,;vif4=,,"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_name = basename(vm_path)

        rs = []
        if scope == "cfg":
            vm_cfg_file = join(vm_path, "vm.cfg")
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            if "vif" not in cfg: cfg["vif"] = []
            for idx, vif in enumerate(cfg["vif"]):
                d = {"mac":"", "bridge":"", "type": ""}
                pairs = [map(string.strip, part.split("="))
                                for part in vif.split(",") if part.strip()]
                d.update(dict(pairs))
                rs.append("vif%d=%s,%s,%s" % (idx, d["mac"], d["bridge"], d["type"]))

        if scope == "mem":
            #FIXME: how to get bridge name and device type?
#            cmd = "xm network-list %s" % vm_name
#            output = run_cmd(cmd)
#            idx_mac = 2
#            for idx, line in enumerate(output.splitlines()[1:]):
#                mac = line.split()[idx_mac]

#xm network-list el5_hvm
#Idx BE     MAC Addr.     handle state evt-ch tx-/rx-ring-ref BE-path
#0   0  ??                   0     1      -1    -1   /-1      /local/domain/0/backend/vif/92/0
#1   0  00:16:3E:77:DC:B3    1     1      -1    -1   /-1      /local/domain/0/backend/vif/92/1
#                if mac == "??": mac = ""
            cmd = "xm domid %s" % vm_name
            domid = run_cmd(cmd)
            cmd = "xenstore-ls /local/domain/0/backend/vif/%s" % domid
            output = run_cmd(cmd)
            vifs = {}
            for line in output.splitlines():
                if not line.startswith(" "):
                    vif_id = line.split("=")[0].strip()
                if line.startswith(" mac"):
                    mac = line.split("=")[1].strip(' "\'').upper()
                    vifs[mac] = vif_id
                    rs.append("vif%s=%s,," % (vif_id, mac))

        rs = ";".join(rs)
        logger.info("xen_get_vifs: success. vm('%s') scope=%s rs=>%s" % \
            (vm_path, scope, rs))
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vifs: error. vm('%s') scope=%s =>%s" % \
                (vm_path, scope, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_clear_vif_mac(vm_path):
    """
    clear vif mac from vm configure file, used by copy vm, etc
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        if "vif" in cfg:
            vifs = []
            for vif in cfg["vif"]:
                attrs = [attr.strip() for attr in vif.split(",")
                        if not re.match(r"^mac\s*=", attr.strip())]
                vifs.append(",".join(attrs))
            cfg["vif"] = vifs

        xen_marshall_cfg(cfg, vm_cfg_file)
        xen_correct_qos_cfg(vm_cfg_file)

        logger.info("xen_clear_vif_mac: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_clear_vif_mac: error. vm('%s') =>%s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

#!!manager still use get_vm_vif() which will call xen_get_vif_num()
@exposed
def xen_get_vif_num(vm_path):
    """
    get vif num
    @return "success:vif_num=<val>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)
        if "vif" in cfg: vif = len(cfg["vif"])
        else: vif = 0
        logger.info("xen_get_vif_num: success. vm('%s') get vif_num=%d" % (vm_path, vif))
        return "success:vif=%d" % vif
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vif_num: failed. vm('%s') =>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

#----------------------------------------<> vnc password <>----
@exposed
def xen_get_vm_vnc_password(vm_path):
    """
    get vm vnc passoword
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        rs = xen_get_vm_type(vm_path)
        if rs.startswith("success"):
            type = parse_result(rs)["type"]
            if "hvm" == type:
                #hvm
                if "vncpasswd" in cfg:
                    vnc_password = cfg["vncpasswd"]
                else:
                    raise Exception("can not get vnc password => type('%s')" % type)
            elif "para" == type:
                #para
                if "vfb" in cfg:
                    vfb = cfg["vfb"]
                    vfb_lst = vfb[0].split(',')

                    vnc_password = None
                    for item in vfb_lst:
                        if item.split('=')[0].strip() == "vncpasswd":
                            vnc_password = item.split('=')[1].strip()
                            break
                    if None == vnc_password:
                        raise Exception("Can not get vnc password => type('%s')" % type)
                else:
                    raise Exception("Can not get vnc password => type('%s')" % type)
            else:
                raise Exception("Error getting VM type => type('%s')" % type)
        else:
            raise Exception("Error getting VM type")

        logger.info("xen_get_vm_vnc_password: success. vm_path('%s')" % vm_path)
        return "success:vnc_password=%s" % vnc_password
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vm_vnc_password: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_set_vm_vnc_password(vm_path, vnc_password, vnclisten="0.0.0.0"):
    """
    set vm vnc password
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if re.sub(r"[0-9a-zA-Z_]", "", vnc_password) != "":
            raise Exception("only alphanumeric chars and '_' are allowed in vnc password!")

        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        rs = xen_get_vm_type(vm_path)
        if rs.startswith("success"):
            type = parse_result(rs)["type"]
            if "hvm" == type:
                #hvm
                cfg["vnc"] = 1
                cfg["vncunused"] = 1
                cfg["vnclisten"] = vnclisten
                cfg["vncconsole"] = 1
                cfg["vncpasswd"] = vnc_password
            elif "para" == type:
                #para
                vnc_cfg = "type=vnc,vncunused=1,vnclisten=%s,vncpasswd=%s" % (vnclisten, vnc_password)
                cfg["vfb"] = []
                cfg["vfb"].append(vnc_cfg)
            else:
                raise Exception("Error getting VM type('%s')" % type)
            xen_marshall_cfg(cfg, vm_cfg_file)
        else:
            raise Exception("Error getting VM type")

        logger.info("xen_set_vm_vnc_password: success. vm_path('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vm_vnc_password: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_set_vm_mem_vcpus_vncpasswd(vm_path, memory=-1, vcpus=-1, vncpasswd=None):
    """
    set vm memory, cpu and vnc password
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if memory != -1:
            rs = xen_set_memory(vm_path, memory)
            if not rs.startswith("success"): raise Exception(rs)
        if vcpus != -1:
            rs = xen_set_vcpus(vm_path, vcpus)
            if not rs.startswith("success"): raise Exception(rs)
        if vncpasswd != None:
            rs = xen_set_vm_vnc_password(vm_path, vncpasswd)
            if not rs.startswith("success"): raise Exception(rs)

        logger.info("xen_set_vm_mem_vcpus_vncpasswd: success. vm_path('%s'), memory('%s'), vcpus('%s')" % \
                (vm_path, memory, vcpus))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vm_mem_vcpus_vncpasswd: failed. vm_path('%s'), memory('%s'), vcpus('%s') => %s" % \
                (vm_path, memory, vcpus, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_optimize_os_profile(vm_path, os_type, os_variant, os_arch=''):
    """
    optimize vm.cfg according to os type/variant/arch

    @param os_type      in python -c "import virtinst; print virtinst.FullVirtGuest.list_os_types()"
    @param os_variant   in python -c "import virtinst; print virtinst.FullVirtGuest.list_os_variants(os_type)"
    @param os_arch      ... x86, x86_64, etc
    @return success if success else "failed:<error message>"
    """
    #FIXME: currently os_arch is only a placeholder, wait virtinst implementation
    logger = logging.getLogger("ovs.operation")
    try:
        import virtinst

        rs = xen_get_vm_type(vm_path)
        if not rs.startswith("success"): raise Exception(rs)
        rs = parse_result(rs)
        if rs["type"] != "hvm":
            logger.warning("xen_optimize_os_profile: vm('%s') type=%s => optimization only applies to hvm!" % \
                    (vm_path, rs['type']))
            return "success"

        os_type = os_type.strip().lower()
        os_variant = os_variant.strip().lower()
        try:
            os_types = virtinst.FullVirtGuest.list_os_types()
            if os_type not in os_types:
                raise Exception("no match os type!")
            os_variants = virtinst.FullVirtGuest.list_os_variants(os_type)
            if os_variant not in os_variants:
                raise Exception("no match os variant!")
        except Exception, e: #no match, return success
            logger.warning("xen_optimize_os_profile: no match profile. vm('%s') os_type=%s os_variant=%s => %s" % \
                (vm_path, os_type, os_variant, fmt_err_msg(e)))
            return "success"

        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)
        _os = virtinst.FullVirtGuest()
        _os.set_os_type_parameters(os_type, os_variant)
        cfg.update(_os.features)
        t = _os.get_input_device()
        if t[0] == "tablet" and t[1] == "usb":
            cfg["usbdevice"] = "tablet"
        elif  "usbdevice" in cfg:
            del cfg["usbdevice"]
        xen_marshall_cfg(cfg, vm_cfg_file)

        logger.info("xen_optimize_os_profile: success. vm('%s') os_type=%s os_variant=%s" % \
                (vm_path, os_type, os_variant))
        return "success"

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_optimize_os_profile: failed. vm('%s') os_type=%s os_variant=%s => %s" % \
            (vm_path, os_type, os_variant, errmsg))
        return "failed:%s" % errmsg

#----------------------------------------<> keyboard map  <>----
@exposed
def xen_get_vm_keyboard_map(vm_path):
    """
    get vm keyboard map
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        keyboard_map = "en-us"
        if "keymap" in cfg:
            keyboard_map = cfg["keymap"]

        logger.info("xen_get_vm_keyboard_map: success. vm_path('%s') => %s" % (vm_path, keyboard_map))
        return "success:keyboard_map=%s" % keyboard_map
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vm_keyboard_map: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_set_vm_keyboard_map(vm_path, keyboard_map):
    """
    set vm keyboard map
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm_cfg_file = join(vm_path, "vm.cfg")
        cfg = xen_unmarshall_cfg(vm_cfg_file)

        cfg["keymap"] = keyboard_map

        xen_marshall_cfg(cfg, vm_cfg_file)
        logger.info("xen_set_vm_keyboard_map: success. vm_path('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vm_keyboard_map: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_set_vm_cpu_pin(vm_path, phy_cpus, scope="cfg"):
    """
    pin the VM to only run on the specific physical CPUs
    @param phy_cpus: ['all'|'1,2'|'1-3'], get cpu information via: get_host_cpu_info
    @return     "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    vm_cfg_file = join(vm_path, "vm.cfg")
    try:
        vm_name = basename(vm_path)
        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))
        if scope_cfg:
            cfg = xen_unmarshall_cfg(vm_cfg_file)
            if phy_cpus == 'all':
                cfg.pop('cpus', None)
            else:
                cfg['cpus'] = phy_cpus
            xen_marshall_cfg(cfg, vm_cfg_file)
        if scope_mem:
            try:
                cmd = "xm vcpu-pin '%s' all '%s'" % (vm_name, phy_cpus)
                run_cmd(cmd)
            except Exception, e:
                if scope_cfg:
                    logger.warning("xen_set_vm_cpu_pin:vm('%s') dynamic pin cpu failed: phy_cpus='%s' failed. =>%s" % \
                                                    (vm_path, phy_cpus, fmt_err_msg(e)))
                else:
                    raise e

        logger.info("xen_set_vm_cpu_pin: vm('%s') phy_cpus='%s' scope='%s'" % \
                (vm_path, phy_cpus, scope))
        return "success"
    except Exception,e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vm_cpu_pin: failed. vm('%s') phy_cpus='%s' scope='%s' => %s" % \
                (vm_path, phy_cpus, scope, errmsg))
        return "failed:%s" % errmsg


#----------------------------------------<> vbd/vif QoS setting <>----
from OVSXXenStore import xen_get_vm_id,xen_status_vm
def mac_exists(vifs_config, mac):
    exist = False
    for vifid in range(len(vifs_config)):
        d = unmarshall_vif_cfg(vifs_config[vifid])
        if "mac" in d and d["mac"].lower() == mac.lower():
            exist = True
            break
    return exist

def xs_get_vifid(domid, mac):
    devices = run_cmd("""xenstore-ls /local/domain/0/backend/vif/%s | grep '^[0-9]\+ = ""'""" % domid)
    vifids = [ dev.split('=')[0].strip() for dev in devices.splitlines()]
    mac_addr = ''
    for vifid in vifids:
        try:
            mac_addr = run_cmd("xenstore-read /local/domain/0/backend/vif/%s/%s/mac" % (domid, vifid)).strip()
        except:
            pass
        else:
            if mac_addr.lower() == mac.lower():
                break
    if mac_addr.lower() != mac.lower():
        raise Exception("faild:no vif with mac %s found in xenstore." % mac)
    else:
        return vifid

@exposed
def xen_get_vif_qos(vm_path, mac, scope="cfg"):
    """Get QoS settings of the virtual network interface which has the
    specific MAC address.

    MAC address is used to identify a unique virtual interface.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string
    @param mac:     MAC address of the virtual network interface
    @type mac:      string
    @param scope:   in [ "cfg", "mem" ]
                    "cfg" - get qos configurations from vm config file
                    "mem" - get runtime qos configuration
    @type scope:    string
    @return:        "success:rate=<rate>,latency=<latency>" on success
                    "failed:<error message>" on failure
    @rtype:         string

    """
    logger = logging.getLogger("ovs.query")
    try:
        if scope == 'cfg':
            vm_cfg_file = join(vm_path, "vm.cfg")
            vm_cfg = xen_unmarshall_cfg(vm_cfg_file)
            # no device, return failure
            if "vif" not in vm_cfg or len(vm_cfg["vif"]) == 0 or not mac_exists(vm_cfg['vif'], mac):
                return "failed:cannot identify the vif by the specific mac address: %s" % mac
            # no extra config, return failure
            if "vif_other_config" not in vm_cfg or len(vm_cfg["vif_other_config"]) == 0:
                return "failed:no qos configuration for vif device with mac: %s" % mac
            for vif_config in vm_cfg["vif_other_config"]:
                if len(vif_config) == 0 or vif_config[0].lower() != mac.lower():
                    continue
                if len(vif_config) != 3:
                    return "failed: '%s' is not the correct QoS setting for vif." % repr(vif_config)
                _, qos_type, qos_params = vif_config[0:3]
                if qos_type != 'tbf':
                    return "failed:'tbf' is required as the qos_algorithm_type, but %s found." % qos_type
                return "success:%s" % qos_params
            return "failed:no qos configuration for vif device with mac: %s" % mac

        elif scope == 'mem':
            # only do it when the vm is running
            rc = xen_status_vm(vm_path)
            if not rc.startswith("success"):
                return "failed:vm not running"
            #get domid
            domid = xen_get_vm_id(vm_path)
            #get vifid
            vifid = xs_get_vifid(domid, mac)
            cmd = "xenstore-read /local/domain/0/backend/vif/%s/%s/qos_algorithm_type" % (domid, vifid)
            qos_type = run_cmd(cmd).strip()
            if qos_type == '':
                return "failed:failed to get qos_algorithm_type, command '%s' returned nothing." % cmd
            if qos_type != 'tbf':
                return "failed:'tbf' is required as the qos_algorithm_type, but %s found." % qos_type
            cmd = "xenstore-read /local/domain/0/backend/vif/%s/%s/qos_algorithm_params" % (domid, vifid)
            qos_params = run_cmd(cmd).strip()
            if qos_params == '':
                return "failed:failed to get qos_algorithm_params, command '%s' returned nothing." % cmd
            return "success:%s" % qos_params

        else:
            raise Exception("Unrecognized scope: %s" % scope)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vif_qos:vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_set_vif_qos(vm_path, mac, rate, latency="50ms", scope="cfg"):
    """Set QoS settings of the virtual network interface which has the
    specific MAC address.

    MAC address is used to identify a unique virtual interface.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string
    @param mac:     MAC address of the virtual network interface
    @type mac:      string
    @param rate:    rate limit to the virtual network interface. Eg. "5mbit"
    @type rate:     string
    @param latency: latency set to the virtual network interface.
                    Default: "50ms"
    @type latency:  string
    @param scope:   in [ "cfg", "mem", "both" ]
                    "cfg" - set qos configurations on vm config file
                    "mem" - set qos configurations at runtime
                    "both" - do both on vm config file and runtime
    @type scope:    string
    @return:        "success" on success
                    "failed:<error message>" on failure
    @rtype:         string
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if scope in ['cfg', 'both']:
            RATE_CONVERTS={'mbit':'Mb', 'gbit':'Gb', 'kbit':'Kb'}
            vm_cfg_file = join(vm_path, "vm.cfg")
            vm_cfg = xen_unmarshall_cfg(vm_cfg_file)
            # no device, return failure
            if "vif" not in vm_cfg or len(vm_cfg["vif"]) == 0 or not mac_exists(vm_cfg['vif'], mac):
                return "failed:cannot identify the vif by the specific mac address: %s" % mac
            #simply check the params, rate must end with 'mbit','gbit','kbit'
            if rate.lower() != 'default' and not (rate[-4:] in RATE_CONVERTS.keys()):
                raise Exception("rate must end within %s" % repr(RATE_CONVERTS.keys()))
            # no extra config, create an empty one
            if "vif_other_config" not in vm_cfg:
                vm_cfg["vif_other_config"] = []
            #remove the vif_other_config for the specific mac
            vif_other_config = vm_cfg["vif_other_config"]
            for vif in vif_other_config:
                if len(vif) == 0 or vif[0].lower() == mac.lower():
                    vif_other_config.remove(vif)
            #add the config, but not for rate='default' ,which means to clear
            if rate.lower() != "default":
                vif_other_config.append([mac.lower(), 'tbf', 'rate=%s,latency=%s' % (rate, latency)])
            #outbound
            for vifid in range(len(vm_cfg['vif'])):
                d = unmarshall_vif_cfg(vm_cfg['vif'][vifid])
                if "mac" in d and d["mac"].lower() == mac.lower():
                    vif = vm_cfg['vif'][vifid]
                    params = [ x.strip() for x in vif.split(',') ]
                    for p in params:
                        if p.startswith('rate'):
                            params.remove(p)
                    if rate.lower() != "default":
                        params.append('rate=%s%s/s@%s' % (rate[:-4],RATE_CONVERTS[rate[-4:]], latency))
                    vif = ','.join(params)
                    vm_cfg['vif'][vifid] = vif
                    break
            xen_marshall_cfg(vm_cfg, vm_cfg_file)

        if scope in ['mem', 'both']:
            # only do it when the vm is running
            rc = xen_status_vm(vm_path)
            if not rc.startswith("success"):
                return "failed:vm not running"
            #get dom id
            domid = xen_get_vm_id(vm_path)
            #get vifid
            vifid = xs_get_vifid(domid, mac)
            run_cmd("xenstore-write /local/domain/0/backend/vif/%s/%s/qos_algorithm_type '%s'" % (domid, vifid, 'tbf'))
            if rate.lower() == "default":
                run_cmd("xenstore-write /local/domain/0/backend/vif/%s/%s/qos_algorithm_params ''" % (domid, vifid))
            else:
                run_cmd("xenstore-write /local/domain/0/backend/vif/%s/%s/qos_algorithm_params '%s'" % (domid, vifid, 'rate=%s,latency=%s' % (rate, latency)))
            logger.warning("xen_set_vif_qos: vm('%s')=>%s" %(vm_path, 'vif inbound Qos setting done, but not able to set outbound for scope=mem'))

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vif_qos:vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

def getVbdFrontEndName(disks, backend):
    backend = 'file:%s' % backend
    for disk in disks:
        back, front, params = disk.split(',')[0:3]
        if back.strip() == backend.strip():
            return front.strip()
    raise Exception('failed to find the front device for the specific end device: %s' % backend)

#copy from OVSXUtility.py, since we cannot directly call it due to the intercross-reference
from OVSConstants import STORAGE_REPO_CONFIG, OVS_BASE
def parse_storage_repos():
    #ovs-makerepo created repos
    d = {}
    try:
        f = file(STORAGE_REPO_CONFIG)
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            try:
                uuid, fs = string.split(line, maxsplit=1)
                d[uuid] = fs
            except: continue
        f.close()
    except:#possible file does not exist
        pass

    return d

def find_real_path(vpath, list_all=False):
    """
    find the real location of path, which could be inside any shared dir,
    created by ovs-makerepo.
    for e.g.:
    '/OVS/running_pool/a' could be one and only one of these paths:
    /OVS/local/running_pool/a
    /OVS/<uuid1>/running_pool/a
    /OVS/<uuid2>/running_pool/a

    @list_all   if True will return list if multiple paths found
                else will raise exception.
    ...
    """
    logger = logging.getLogger("ovs.query")

    vpath = abspath(vpath)
    possible_paths = [vpath]

    ptn = re.compile("^%s" % OVS_BASE)
    possible_paths.append(ptn.sub(join(OVS_BASE, "local"), vpath))

    d = parse_storage_repos()
    possible_paths.extend([ptn.sub(join(OVS_BASE, uuid), vpath) for uuid in d])
    real_paths = [path for path in possible_paths if exists(path)]
    real_paths = list(set(real_paths))
    if real_paths:
        if not list_all:
            if len(real_paths) > 1:
                raise Exception("Got more than one file/dir with the same name: '%s'" % repr(real_paths))
            return real_paths[0]
        else:
            return real_paths
    else:
        raise Exception("Could not find '%s' in any storage repository." % vpath)

def xs_get_vbdid(domid, disk_path):
    devices = run_cmd("""xenstore-ls /local/domain/0/backend/vbd/%s | grep '^[0-9]\+ = ""'""" % domid)
    vbdids = [ dev.split('=')[0].strip() for dev in devices.splitlines()]
    backend_path = ''
    for vbdid in vbdids:
        try:
            backend_path = run_cmd("xenstore-read /local/domain/0/backend/vbd/%s/%s/params" % (domid, vbdid)).strip()
        except:
            pass
        else:
            if backend_path == disk_path:
                break
    if backend_path != disk_path:
        raise Exception("faild:no vbd with backend %s found in xenstore." % disk_path)
    else:
        return vbdid

@exposed
def xen_get_vbd_qos(vm_path, backend_name, shared=0, scope="cfg"):
    """Get QoS settings of the virtual block device who's backend device is
    "backend_name".

    "backend_name, shared" are used to identify a unique virtual block device.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string
    @param backend_name: backend device name of the virtual block device,e.g.: System.img
    @type backend_name:  string
    @param shared: a boolean value to indicate whether the disk is shared
    @type shared:  integer [0:1]
    @param scope:   in [ "cfg", "mem" ]
                    "cfg" - get qos configurations from vm config file
                    "mem" - get qos configurations from runtime
    @type scope:    string
    @return:        "success:sched=<sched>,prio=<prio>"
                    "failed:<error message>" on failure
    @rtype:         string
    """
    logger = logging.getLogger("ovs.query")
    try:
        if shared == 0:
            disk_path = join(vm_path, backend_name)
        else:
            disk_path = join("/OVS/sharedDisk", backend_name)
        disk_path = find_real_path(disk_path)

        if scope == "cfg":
            vm_cfg_file = join(vm_path, "vm.cfg")
            vm_cfg = xen_unmarshall_cfg(vm_cfg_file)
            # no device, return failure
            if "disk" not in vm_cfg or len(vm_cfg["disk"]) == 0:
                return "failed:cannot identify the vbd by the specific backend: %s" % disk_path
            front = getVbdFrontEndName(vm_cfg["disk"], disk_path)
            # no extra config, return failure
            if "disk_other_config" not in vm_cfg or len(vm_cfg["disk_other_config"]) == 0:
                return "failed:no configuration for vbd device with backend: %s" % disk_path
            for disk_config in vm_cfg["disk_other_config"]:
                if len(disk_config) == 0 or disk_config[0] != front:
                    continue
                if len(disk_config) != 3:
                    return "failed: '%s' is not the correct QoS setting for vbd." % repr(disk_config)
                _, qos_type, qos_params = disk_config[0:3]
                if qos_type != 'ionice':
                    return "failed:'ionice' is required as the qos_algorithm_type, but %s found." % qos_type
                return "success:%s" % qos_params
            return "failed:no configuration for vbd device with backend: %s" % disk_path

        elif scope == 'mem':
            # only do it when the vm is running
            rc = xen_status_vm(vm_path)
            if not rc.startswith("success"):
                return "failed:vm not running"
            #get domid
            domid = xen_get_vm_id(vm_path)
            #get vbdid
            vbdid = xs_get_vbdid(domid, disk_path)
            cmd = "xenstore-read /local/domain/0/backend/vbd/%s/%s/qos_algorithm_type" % (domid, vbdid)
            qos_type = run_cmd(cmd).strip()
            if qos_type == '':
                return "failed:failed to get qos_algorithm_type, command '%s' returned nothing." % cmd
            if qos_type != 'ionice':
                return "failed:'ionice' is required as the qos_algorithm_type, but %s found." % qos_type
            cmd = "xenstore-read /local/domain/0/backend/vbd/%s/%s/qos_algorithm_params" % (domid, vbdid)
            qos_params = run_cmd(cmd).strip()
            if qos_params == '':
                return "failed:failed to get qos_algorithm_params, command '%s' returned nothing." % cmd
            return "success:%s" % qos_params

        else:
            raise Exception("Unrecognized scope: %s" % scope)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_disk_qos:vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_set_vbd_qos(vm_path, backend_name, shared=0, sched="real-time", prio="4", scope="cfg"):
    """Set QoS settings of the virtual block device who's frontend device is
    "front". Internal use only.
    "front" is used to identify a unique virtual block device.

    @param vm_path: virtual machine config file and disk image file path
    @type vm_path:  string
    @param backend_name: backend device name of the virtual block device,e.g.: System.img
    @type backend_name:  string
    @param shared:  a boolean value to indicate whether the disk is shared
    @type shared:   integer [0:1]
    @param sched:   scheduling algorithm.
                    in ["real-time", "rt", "best-effort", "idle"].
    @type sched:    string
    @param prio:    priority for scheduling. from 0 to 7.
    @type prio:     string
    @param scope:   in [ "cfg", "mem", "both" ]
                    "cfg" - set qos configurations on vm config file
                    "mem" - set qos configuration at runtime
                    "both" - do both on vm config file and runtime
    @type scope:    string
    @return:        "success" on success
                    "failed:<error message>" on failure
    @rtype:         string
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if shared == 0:
            disk_path = join(vm_path, backend_name)
        else:
            disk_path = join("/OVS/sharedDisk", backend_name)
        disk_path = find_real_path(disk_path)

        if scope in ['cfg', 'both']:
            vm_cfg_file = join(vm_path, "vm.cfg")
            vm_cfg = xen_unmarshall_cfg(vm_cfg_file)
            # no device, return failure
            if "disk" not in vm_cfg or len(vm_cfg["disk"]) == 0:
                return "failed:cannot identify the vbd by the specific backend: %s" % disk_path
            front = getVbdFrontEndName(vm_cfg["disk"], disk_path)
            # no extra config, create an empty one
            if "disk_other_config" not in vm_cfg:
                vm_cfg["disk_other_config"] = []
            #remove the disk_other_config for the specific device
            disk_other_config = vm_cfg["disk_other_config"]
            for disk in disk_other_config:
                if len(disk) == 0 or disk[0] == front:
                    disk_other_config.remove(disk)
            #add the qos config for the device, but not for sched=default which means to clear
            if sched.lower() != "default":
                disk_other_config.append([front, 'ionice', 'sched=%s,prio=%s' % (sched, prio)])
            xen_marshall_cfg(vm_cfg, vm_cfg_file)

        if scope in ['mem', 'both']:
            # only do it when the vm is running
            rc = xen_status_vm(vm_path)
            if not rc.startswith("success"):
                return "failed:vm not running"
            #get dom id
            domid = xen_get_vm_id(vm_path)
            #get vbdid
            vbdid = xs_get_vbdid(domid, disk_path)
            run_cmd("xenstore-write /local/domain/0/backend/vbd/%s/%s/qos_algorithm_type '%s'" % (domid, vbdid, 'ionice'))
            if sched.lower() == "default":
                run_cmd("xenstore-write /local/domain/0/backend/vbd/%s/%s/qos_algorithm_params ''" % (domid, vbdid))
            else:
                run_cmd("xenstore-write /local/domain/0/backend/vbd/%s/%s/qos_algorithm_params '%s'" % (domid, vbdid, 'sched=%s,prio=%s' % (sched, prio)))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_set_vbd_qos:vm('%s')=>%s" % (vm_path, errmsg))
        return "failed:%s" % errmsg
