#!/usr/bin/env python
#^#
# OVSXXenVMInstall.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#
import string
import re
import os
import logging
from tempfile import NamedTemporaryFile
from shutil import copyfile
from os.path import join, isdir, isfile, basename, abspath, dirname
from OVSXXenVMConfig import xen_unmarshall_cfg, xen_marshall_cfg, xen_set_vm_vnc_password, xen_add_disk, xen_set_vifs
from OVSXXenVMConfig import xen_set_vm_vnc_password, xen_set_memory, xen_set_vcpus
from OVSXXenStore import xen_mem_free, xen_get_caps
from OVSXUtility import utl_exists, xen_create_disk
from OVSCommons import fmt_err_msg, run_cmd, exposed, parse_result, check_shutdown_mode

#----------------------------<> helper function <>----
def parse_vif_cfg_param(vif_cfg, vtype="hvm"):
    d = {}
    macs, bridges = [], []
    for vif in vif_cfg.split(";"):
        try:
            #virt-install does not accept device type yet
            l = map(string.strip, vif.split(","))
            l += [''] * (3 - len(l))
            mac, bridge, _type = l
            if mac: macs.append(mac)
            if bridge: bridges.append(bridge)
            if _type: d["vif_type"] = _type
        except: pass
    if len(macs) == len(bridges):
        d["mac_bridge"] = zip(macs, bridges)
    else:
        d["bridge"] = bridges

    vif_param = ""
    if "bridge" in d:
        for bridge in d["bridge"]:
            vif_param += " --bridge=%s " % bridge
    if "mac_bridge" in d:
        for mac, bridge in d["mac_bridge"]:
            vif_param += " --mac=%s --bridge=%s " % (mac, bridge)
    if ("vif_type" in d) and (vtype == "hvm"):
        vif_param += " --vif-type=%s " % d["vif_type"]

    return vif_param

#----------------------------<> install from scratch <>----
@exposed
def xen_install_vm_hvm(vm_path, memory, vcpus, vif_cfg, disks, cdrom,
        vncpassword, vnclisten="0.0.0.0",
        os_type='', os_variant='', keymap='en-us'):
    """
    install full virtualized vm from scratch
    @param memory   size in MB
    @param vcpus    cpu number
    @param vif_cfg      with format '<mac>,<bridge>,<type>[;<mac>,<bridge>,<type>]'
                        mac in this param follow all or none convention, i.e.
                        "<mac>,<bridge>,<type>;<mac>,<bridge>,<type>" or ",<bridge>,<type>;,<bridge>,<type>"
    @param disks    disk string, in format "<disk_path>:<disk_size>:<is_sparse>[,<disk_path>:<disk_size>:<is_sparse>]" e.g.
                    system.raw:2:False
                    disk path should be relative to vm_path
                    disk_size is counted in GB
                    is_sparse should be True or False
    @param cdrom    path to the install media iso file
    @param os_type      in python -c 'import virtinst; print virtinst.FullVirtGuest.list_os_types()'
                        if '' will not specified as virt-install option
    @param os_variant   in python -c 'import virtinst; print virtinst.FullVirtGuest.list_os_variants(os_type)'
                        if '' will not specified as virt-install option
    @param keymap	the keyboard map simulated by the QEMU
    @return             "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    #-------------------------------------------------------------------------------
    def get_os_profile_option(os_type, os_variant):
        option = ''

        os_type = os_type.strip().lower()
        os_variant = os_variant.strip().lower()
        if not os_type or not os_variant: return option

        try:
            import virtinst
            os_types = virtinst.FullVirtGuest.list_os_types()
            if os_type in os_types:
                os_variants = virtinst.FullVirtGuest.list_os_variants(os_type)
                if os_variant in os_variants:
                    option = " --os-type=%s --os-variant=%s" % (os_type, os_variant)
        except Exception, e:
            logger.warning("xen_install_vm_hvm: get_os_profile_option() => %s" % fmt_err_msg(e))
        return option
    #-------------------------------------------------------------------------------
    vm_path = abspath(vm_path)
    try:
        check_shutdown_mode()

        rs = xen_get_caps()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)
        caps = rs["caps"]
        if not "hvm" in caps:
            raise Exception("Fully virtualized VM is not supported on the host. CPU with Intet-VT or AMD-V needed.")

        vm_name = basename(vm_path)
        if not re.match(r"^[0-9a-zA-Z_]+$", vm_name):
            raise Exception("Domain name must be alphanumeric or _")
        if len(vm_name) > 64:
            raise Exception("Length of Domain name must be less than 64")
        if not isdir(vm_path):
            os.makedirs(vm_path)

        rs = utl_exists(cdrom)
        if not rs.startswith("success"):
            raise Exception("cdrom('%s') does not exist." % cdrom)

        rs = xen_mem_free()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)
        mem_free = int(rs["mem_free"])
        if mem_free < memory:
            raise Exception("not enough memory")

        disks = disks.split(",")
        disks = [disk.split(":") for disk in disks]
        disk_param = ""
        for disk_path, disk_size, sparse_option in disks:
            disk_param += "--file %s --file-size=%s " % \
                    (join(vm_path, disk_path), disk_size)
            if sparse_option.lower() == "false":
                disk_param += " --nonsparse "

        vif_param = parse_vif_cfg_param(vif_cfg, "hvm")

        f = NamedTemporaryFile()
        try:
            f.write(vncpassword)
            f.flush()

            cmd = """virt-install --hvm --name=%s --ram=%d --vcpus=%d %s %s \
                    --vnc --vnclisten=%s --vncpasswd=%s --noautoconsole --keep-cdrom --cdrom=%s""" % \
                    (vm_name, int(memory), int(vcpus), vif_param, disk_param, vnclisten, f.name, cdrom)
            extra_option = get_os_profile_option(os_type, os_variant)
            if extra_option: cmd = "%s %s" % (cmd, extra_option)
            cmd = "%s --keymap=%s" % (cmd, keymap)

            logger.debug("xen_install_vm_hvm: cmd=%s" % cmd)
            run_cmd(cmd)
        finally: f.close()

        vm_cfg_file = join(vm_path, "vm.cfg")
        copyfile("/etc/xen/%s" % vm_name, vm_cfg_file)
        cfg = xen_unmarshall_cfg(vm_cfg_file)
        if "uuid" in cfg: del cfg["uuid"]
        if "disk" in cfg:
            disks = []
            for disk in cfg["disk"]:
                backend_dev, frontend_dev, mode = disk.split(",")
                if "cdrom" in frontend_dev:
                    disk = "file:%s,%s,%s" % (cdrom, frontend_dev, mode)
                disks.append(disk)
            cfg["disk"] = disks
        else:
            raise Exception("invalid config file: no disk found!")

        xen_marshall_cfg(cfg, vm_cfg_file)
        xen_set_vm_vnc_password(vm_path, vncpassword)

        logger.info("xen_install_vm_hvm:success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_install_vm_hvm: error. vm('%s') %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_change_vm_cdrom(vm_path, iso_path, scope="mem"):
    """
    change vm cdrom
    @param  iso_path abspath to iso file. if iso_path=='', remove any existing cdrom
    @return "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    vm_path = abspath(vm_path)
    vm_cfg_file = join(vm_path, "vm.cfg")
    try:
        vm_name = basename(vm_path)

        cfg = xen_unmarshall_cfg(vm_cfg_file)
        cdrom_idx, cdrom_dev, cdrom_mode = -1, "", "r"
        for idx, vdisk in enumerate(cfg["disk"]):
            be_dev, fe_dev, mode = vdisk.split(",")
            if "cdrom" in fe_dev:
                cdrom_idx, cdrom_dev, cdrom_mode = idx, fe_dev, mode

        scope_cfg = (scope in ("cfg", "both"))
        scope_mem = (scope in ("mem", "both"))

        if scope_mem:
            be_dev = "file:%s" % iso_path
            fe_dev = cdrom_dev or "hdc:cdrom"
            mode = "r"
            cmd = "xm block-configure %s %s %s %s" % \
                    (vm_name, be_dev, fe_dev, mode)
            run_cmd(cmd)

        if scope_cfg:
            if cdrom_dev and cdrom_idx != -1:
                if not iso_path.strip():
                    be_dev = "" #empty cdrom
                else:
                    be_dev = "file:%s" % iso_path
                cfg["disk"][cdrom_idx] = "%s,%s,%s" % (be_dev, cdrom_dev, cdrom_mode)
                xen_marshall_cfg(cfg, vm_cfg_file)

        logger.info("xen_change_vm_cdrom:success. vm('%s') iso('%s')" % \
                (vm_path, iso_path))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_change_vm_cdrom:failed. vm('%s') iso('%s')=>%s" % \
                (vm_path, iso_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_install_vm_para(vm_path, memory, vcpus, vif_cfg, disks, location, vncpassword, vnclisten="0.0.0.0", keymap='en-us'):
    """
    install para virtualized vm from scratch
    @param memory       size in MB
    @param vcpus        cpu number
    @param vif_cfg      refer to xen_set_vifs()
    @param disks        disk string, in format "disk_path:disk_size:is_sparse" e.g.
                        system.raw:2:False
                    disk path should be relative to vm_path
                    disk_size is counted in GB
                    is_sparse should be True or False
    @param location     location of the install directory e.g.
                        nfs:host:path
                        http and ftp protocols are also supported e.g.
                        http://host/path
                        ftp://host/path
    @param keymap	keyboard map simulated
    @return             "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    vm_path = abspath(vm_path)
    try:
        check_shutdown_mode()

        vm_name = basename(vm_path)
        if not re.match(r"^[0-9a-zA-Z_]+$", vm_name):
            raise Exception("Domain name must be alphanumeric or _")
        if len(vm_name) > 64:
            raise Exception("Length of Domain name must be less than 64")
        if not isdir(vm_path):
            os.makedirs(vm_path)

        rs = xen_mem_free()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)
        mem_free = int(rs["mem_free"])
        if mem_free < memory:
            raise Exception("not enough memory")

        disks = disks.split(",")
        disks = [disk.split(":") for disk in disks]
        disk_param = ""
        for disk_path, disk_size, sparse_option in disks:
            disk_param += "--file %s --file-size=%s " % \
                    (join(vm_path, disk_path), disk_size)
            if sparse_option.lower() == "false":
                disk_param += " --nonsparse "

        vif_param = parse_vif_cfg_param(vif_cfg, "para")

        f = NamedTemporaryFile()
        try:
            f.write(vncpassword)
            f.flush()
            cmd = """virt-install --paravirt --name=%s --ram=%d --vcpus=%d %s %s --vnc \
                    --vnclisten=%s --vncpasswd=%s --noautoconsole --location=%s --keymap=%s""" % \
                    (vm_name, int(memory), int(vcpus), vif_param, disk_param, vnclisten, f.name, location, keymap)

            logger.debug("xen_install_vm_para: cmd=%s" % cmd)
            run_cmd(cmd)
        finally: f.close()

        vm_cfg_file = join(vm_path, "vm.cfg")
        copyfile("/etc/xen/%s" % vm_name, vm_cfg_file)
        cfg = xen_unmarshall_cfg(vm_cfg_file)
        if "uuid" in cfg: del cfg["uuid"]
        xen_marshall_cfg(cfg, vm_cfg_file)
        xen_set_vm_vnc_password(vm_path, vncpassword)

        logger.info("xen_install_vm_para:success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_install_vm_para: error. vm('%s') %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_create_dummy_vm(vm_path, memory, vcpus, vif_cfg, disks, vtype, vncpassword, vnclisten="0.0.0.0"):
    """
    create a dummy vm, (situation to use this dummy vm are 1) pxe boot 2) boot from cdrom, etc)

    @param memory       size in MB
    @param vcpus        cpu number
    @param vif_cfg      refer to xen_set_vifs()
    @param disks        disk string, in format "disk_path:disk_size:is_sparse" e.g.
                        system.raw:2:False
                        disk path should be relative to vm_path
                        disk_size is counted in GB
                        is_sparse should be True or False
    @param vtype        "hvm" or "pvm"
    @return             "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    vm_path = abspath(vm_path)
    try:
        vm_name = basename(vm_path)
        if not re.match(r"^[0-9a-zA-Z_]+$", vm_name):
            raise Exception("Domain name must be alphanumeric or _")
        if len(vm_name) > 64:
            raise Exception("Length of Domain name must be less than 64")
        if not isdir(vm_path):
            os.makedirs(vm_path)

        vm_cfg_file = join(vm_path, "vm.cfg")
        copyfile("%s/conf/PXE_%s.cfg.default" % (dirname(__file__), vtype), vm_cfg_file)

        # set disk
        disks = disks.split(",")
        disks = [disk.split(":") for disk in disks]
        disk_param = ""
        for disk_path, disk_size, sparse_option in disks:
            if sparse_option.lower() == "false":
                sparse = False
            else:
                sparse = True
            disk_size = float(disk_size)
            full_disk_path = join(vm_path, basename(disk_path))
            rs = xen_create_disk(full_disk_path, disk_size * 1024, sparse)
            if not rs.startswith("success"): raise Exception(rs)
            rs = xen_add_disk(vm_path, full_disk_path)
            if not rs.startswith("success"): raise Exception(rs)

        # set vif
        rs = xen_set_vifs(vm_path, vif_cfg)
        if not rs.startswith("success"): raise Exception(rs)

        # set vnc
        rs = xen_set_vm_vnc_password(vm_path, vncpassword, vnclisten)
        if not rs.startswith("success"): raise Exception(rs)

        # set memory
        rs = xen_set_memory(vm_path, memory)
        if not rs.startswith("success"): raise Exception(rs)

        # set vcpus
        rs = xen_set_vcpus(vm_path, vcpus)
        if not rs.startswith("success"): raise Exception(rs)

        logger.info("xen_create_dummy_vm:success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_create_dummy_vm: error. vm('%s') %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg
