#!/usr/bin/env python
#^#
# agent_config.py
#
# /etc/init.d/ovs-agent configure
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#
from os.path import dirname, join
import sys
sys.path.append(join(dirname(__file__), ".."))

import string
import sha
import re
import random
from shutil import copyfile
from os.path import join, dirname, normpath, abspath, isabs, isfile, isdir, getsize, exists
from os import makedirs, remove, kill
from getpass import getpass
from OVSConstants import CONFIG_DIR, PID_DIR
from ConfigParser import ConfigParser
from signal import SIGUSR1
from time import sleep

#----------------------------<> utils function <>----
def unmarshall_passwdfile(passwd_file):
    d = {}
    if not exists(passwd_file):
        return d
    f = file(passwd_file, "r")
    for line in f:
        try:
            username, password = line.strip().split(":{SHA}")
            d[username] = password
        except: continue
    f.close()
    return d

def marshall_passwdfile(d, passwd_file):
    path=dirname(passwd_file);
    path=abspath(path);
    if not exists(path):
        makedirs(path);

    f = open(passwd_file, "w")
    try:
        for username, passwd_digest in d.items():
            f.write("%s:{SHA}%s\n" % (username, passwd_digest))
    finally:
        f.close()

def update_ovsagent_user(passwd_file, username, passwd):
    d = unmarshall_passwdfile(passwd_file)
    if username in d: rs = "passwd digest for %s updated!" % username
    else: rs = "passwd digest for %s added!" % username
    d[username] = sha.new(passwd).digest().encode("base64").replace("\n", "")
    marshall_passwdfile(d, passwd_file)
    return rs

def remove_ovsagent_user(passwd_file, username):
    d = unmarshall_passwdfile(passwd_file)
    if username in d:
        del d[username]
        marshall_passwdfile(d, passwd_file)
        rs = "passwd digest for %s removed!" % username
    else: rs = "passwd digest for %s not exists!" % username
    return rs

def list_ovsagent_user(passwd_file):
    print "\novs agent users list:\n\t",
    d = unmarshall_passwdfile(passwd_file)
    usernames = d.keys()
    #donnot list admin user
    usernames.remove("admin")
    print " ".join(d.keys())

def gen_passwd(length=8):
    chars = "%s%s" %(string.digits, string.letters)
    return "".join([random.choice(chars) for i in range(0, length)])

def notify_agent():
    pid_file = join(PID_DIR, "OVSAgentServer.pid")
    if not exists(pid_file):
        return
    pid = None
    f = open(pid_file, 'r')
    try:
        pid = f.read().strip()
    finally:
        f.close()
    cmdline_file = '/proc/%s/cmdline' % pid
    if not pid or not exists(cmdline_file):
        return
    f = open(cmdline_file)
    try:
        cmdline = f.readline()
    finally:
        f.close()
    if cmdline.find('OVSAgentServer.py') < 0:
        return

    print "Tell agent to reload cache ..."
    failed = True
    if pid:
        dirty = join(CONFIG_DIR, "dirtycache")
        #agent removes dirty file after reloading the cache
        open(dirty, 'w').close()
        #signal is not safe, may get lost, for try 3 times ...
        for i in range(3):
            try:
                kill(int(pid), SIGUSR1)
            except:
                break
            sleep(1)
            if not exists(dirty):
                failed = False
                break
    if failed:
        try:
            remove(dirty)
        except:
            pass
        print "Failed to reload the agent cache,if the agent is running, you need to restart the agent to make the new config effective."
    else:
        print "Agent cache reloaded."

#----------------------------<> main <>----
from atexit import register
register(notify_agent)

agent_dir = normpath("%s/.." % dirname(__file__))
from OVSConfig import get_config
current_config = get_config()

cfg_file_default = join(dirname(__file__), "../conf/agent.ini.default")
default_config = ConfigParser()
default_config.read(cfg_file_default)

f = open(cfg_file_default)
lines = f.readlines()
f.close()
lines = [line.strip() for line in lines]

#header
header = []
for line in lines:
    if line.startswith(";") or line == "":
        header.append(line)
    else:break

#tail
tail = []
lines.reverse()
for line in lines:
    if line.startswith(";") or line=="":
        tail.append(line)
    else: break
tail.reverse()
lines.reverse()

lines=lines[len(header):]
if tail: lines=lines[:-len(tail)]

sections = []
section = None
item = None
for line in lines:
    if line.startswith("["):
        if section: sections.append(section)
        section = {"section":line, "items":[]}
        item = {"comment":[], "keyval":""}
    elif line.startswith(";"):
        item["comment"].append(line)
    elif line != "":
        item["keyval"] = line
        section["items"].append(item)
        item = {"comment":[], "keyval":""}
sections.append(section)

#user configuration
for (s_idx, section) in enumerate(sections):
    for (i_idx, item) in enumerate(section["items"]):
        if not item["comment"] or not re.match(r";\s*\(HIDDEN\)", item["comment"][-1]):
            if item["comment"]: print "\n".join(item["comment"])
            k, v = item["keyval"].split("=")
            #preserve user's customization
            try:
                if current_config:
                    section_name = section["section"].strip(" []")
                    v = current_config.get(section_name, k)
            except: pass
            try:
                input = raw_input("%s=%s\n%s=" %(k,v,k))
            except Exception, e:
                print "\nconfiguration canceled."
                sys.exit(1)
            #type a space to clear previous value
            if input != "":
                sections[s_idx]["items"][i_idx]["keyval"] = "%s=%s" % (k, input.strip())
            print "now %s\n" % sections[s_idx]["items"][i_idx]["keyval"]

#username/passwd for agent authentication
passwd_file = current_config.get("security", "passwdfile")
if not isabs(passwd_file):
    passwd_file = abspath(join(agent_dir, passwd_file))

#normal user to access ovs agent
try:
    yn = raw_input("would you like to modify password to communicate with agent (local)?[y/N]")
except Exception, e:
    print "\nconfiguration canceled."
    sys.exit(1)
if yn.lower() in ("y", "yes", "ok"):
    username = "oracle"
    while True:
        try:
            passwd = getpass("password:").strip()
            passwd_2 = getpass("again:").strip()

            if not passwd: raise Exception("password cannot be empty!")
            if passwd != passwd_2: raise Exception("2 inputs do not match!")
            if not passwd.isalnum(): raise Exception("only alphanumeric char allowed!")
            if len(passwd) < 6: raise Exception("minimum password length is 6!")

            update_ovsagent_user(passwd_file, username, passwd)
            break
        except KeyboardInterrupt, e:
            print "\nconfiguration canceled."
            sys.exit(1)
        except Exception, e:
            print "\n%s try again or ^C to cancel." % e.args[0]

if default_config.get("ha", "remaster").lower().startswith("enable"):
    #shared passwdfile
    passwd_file_shared = default_config.get("ha", "passwdfile.shared")
    try:
        passwd_file_shared = current_config.get("ha", "passwdfile.shared")
    except: pass
    try:
        yn = raw_input("would you like to modify password to communicate with agent (master)?[y/N]")
    except Exception, e:
        print "\nconfiguration canceled."
        sys.exit(1)
    if yn.lower() in ("y", "yes", "ok"):
        username = "oracle"
        while True:
            try:
                passwd = getpass("password:").strip()
                passwd_2 = getpass("again:").strip()

                if not passwd: raise Exception("password cannot be empty!")
                if passwd != passwd_2: raise Exception("2 inputs do not match!")
                if not passwd.isalnum(): raise Exception("only alphanumeric char allowed!")
                if len(passwd) < 6: raise Exception("minimum password length is 6!")

                update_ovsagent_user(passwd_file_shared, username, passwd)
                break
            except KeyboardInterrupt, e:
                print "\nconfiguration canceled."
                sys.exit(1)
            except Exception, e:
                print "\n%s try again or ^C to cancel." % e.args[0]

#update admin passwd
admin_passwd = gen_passwd()
update_ovsagent_user(passwd_file, "admin", admin_passwd)

#write back config file
cfg_file = join(CONFIG_DIR, "agent.ini")
f = open(cfg_file, "w")
try:
    f.write("%s\n" % "\n".join(header))
    for section in sections:
        f.write("%s\n" % section["section"])
        for item in section["items"]:
            if item["comment"]:
                f.write("%s\n" % "\n".join(item["comment"]))
            #spetial case for admin.password
            if item["keyval"].strip().startswith("admin.password"):
                item["keyval"] = "admin.password=%s" % admin_passwd
            f.write("%s\n" % item["keyval"])
    f.write("%s\n" % "\n".join(tail))
finally:
    f.close()

print "\nOVS Agent configuration done.\n"
