#!/usr/bin/env python
# Copyright (C) 2008, Oracle
# v2v.py
# OracleVM 2.1.2
# Virtual to virtual conversion
import os
import statvfs
import signal
from subprocess import Popen
from subprocess import PIPE
import threading
import tempfile
import vmx_parser
import fcntl
import select
import logging
import time
import urllib
from stat import *
import errno

from procname import setprocname

from v2vstatus import get_status, set_status, save_utl_srv, get_utl_srv

import virtinst.util # for randomMAC

from OVSCommons import parse_result, exposed, chk_vm_path, fmt_err_msg, run_cmd
from OVSServerSelection import select_one_server, get_utility_server
from OVSXUtility import find_real_path, utl_find_real_path, assemble_real_path, utl_get_free_space, get_storage_repos
from OVSConstants import PID_DIR
from OVSServerProxy import ServerProxy
from OVSCDS import Timeout

import vmx_parser
import vmdk_parser
import ovs_uuid

debug = 0
unit_test = 0

_supported_src_type = ['vmx']
_supported_dest_type = ['raw']

def _pid_filename(pid):
    filename = os.path.splitext(os.path.basename(__file__))[0]
    return os.path.join(PID_DIR, '%s%d.pid' % (filename, pid))

def _create_pid_file(pid=None):
    if not pid:
        pid = os.getpid()
    f = open(_pid_filename(pid), 'w')
    try:
        f.write('%d' % pid)
    finally:
        f.close()

def _remove_pid_file(pid=None):
    if not pid:
        pid = os.getpid()
    try:
        os.remove(_pid_filename(pid))
    except:
        pass

def _setprocname(name):
    setprocname('python OVSAgentServer %s' % name)

def _vmx2raw(src_file_path, dest_dir_path):
    global debug

    p = vmx_parser.vmx_parser(src_file_path)
    if not p.is_validate_vmx():
        return "failed:Invalidated vmx config file"

    disks = p.get_vmdk_list()
    if len(disks) == 0:
        raise Exception("_vmx2raw: There is no disk in the vmx config file.")

    src_dir = os.path.dirname(src_file_path)
    root, ext = os.path.splitext(src_file_path)
    root_base = os.path.basename(root)

    __, tmpname = tempfile.mkstemp("_v2v_disks")
    if debug:
        print "tmpname:", tmpname
    try:
        f = open(tmpname, "w")
        for v in disks:
            if not v.startswith("/"):
                v = os.path.join(src_dir, v)
            f.write(v + '\n')
    finally:
        f.close()
    p = Popen([os.path.join(os.path.dirname(__file__), "utils/v2v_disk_convert.sh"), tmpname, dest_dir_path], stdout=PIPE, stderr=PIPE) 
    return p

def _create_vmcfg(src_file_path, dest_dir_path):
    """ Create vm.cfg from vmx config
        
        src_file_path: local path of vmx config
        dest_dir_path: target directory of vm.cfg

        Return:
            failed:  "failed:%s" % errmsg
                     "failed:vmware-vdiskmanager doesn't found in the path."
        Example:
            _create_vmcfg("/OVS/running_pool/test/test.vmx", "/tmp")
    """
    global debug

    logger = logging.getLogger("ovs.operation")
    try:

        logger.info('_create_vmcfg: "%s" "%s"' % (src_file_path, dest_dir_path))
        basename = os.path.basename(src_file_path)
        basename = os.path.splitext(basename)[0]
        p = vmx_parser.vmx_parser(src_file_path)
        name = p.get_guestOS()
        if not name:
            name = basename + "_v2v"
        disk = "["
        disk_label = 'a'
        for v in p.get_vmdk_list():
            v = os.path.splitext(v)[0] + '.img'
            disk = disk + " 'file:%s,hd%s,w'," % (os.path.join(dest_dir_path, v), disk_label)
            disk_label = chr(ord(disk_label) + 1)
        disk = disk + " ]"
        vif = "["
        i = 0
        while i < p.get_ethnum():
            vif = vif + " 'type=ioemu, mac=%s, bridge=xenbr0'," % virtinst.util.randomMAC()
            i = i + 1
        vif = vif + " ]"

        f = open(os.path.join(dest_dir_path, 'vm.cfg'), 'w')
        try:
            f.write('# Automatically created by OracleVM V2V\n')
            f.write('name = "%s"\n' % name)
            f.write('builder = "hvm"\n')
            f.write('memory = "%s"\n' % p.get_memsize())
            f.write('disk = %s\n' % disk)
            f.write('vif = %s\n' % vif)
            f.write('uuid = "%s"\n' % ovs_uuid.uuid())
            f.write('device_model = "/usr/lib/xen/bin/qemu-dm"\n')
            f.write('kernel = "/usr/lib/xen/boot/hvmloader"\n')
            f.write('vnc=1\n')
            f.write('vncunused=1\n')
            f.write("vnclisten='0.0.0.0'\n")
            f.write('timer_mode=0\n')
            f.write('apic=1\n')
            f.write('acpi=1\n')
            f.write('pae=1\n')
            f.write('\n')
            f.write('vcpus=%d\n' % p.get_vcpusnum())
            f.write('serial = "pty"\n')
            f.write("on_reboot = 'restart'\n")
            f.write("on_crash = 'restart'\n")
        finally:
            f.close()

        return "success:"
    except Exception,e:
        errmsg = fmt_err_msg(e)
        if debug:
            print e
        logger.error('_create_vmcfg: failed. =>%s'  % errmsg)
        return "failed:%s" % errmsg

def _run_as_daemon():
    try:
        pid = os.fork()
    except OSError, e:
        raise Exception("%s [%d]" % (e.strerror, e.errno))
    if pid == 0: # child
        os.setsid()
        try:
            pid = os.fork()
        except OSError, e:
            raise Exception("%s [%d]" % (e.strerror, e.errno))
        if pid == 0: # second child
            os.chdir("/")
            os.umask(0)
        else:
            os._exit(0) # Exit first child

@exposed
def create_vmcfg(src_file_path, dest_dir_path):
    """ Create vm.cfg from vmx config
        
        src_file_path: local path of vmx config
        dest_dir_path: target directory of vm.cfg

        Return:
            failed:  "failed:%s" % errmsg
                     "failed:vmware-vdiskmanager doesn't found in the path."
        Example:
            create_vmcfg("/OVS/running_pool/test/test.vmx", "/tmp")
    """

    logger = logging.getLogger("ovs.operation")
    try:
        logger.info('create_vmcfg: "%s" "%s"' % (src_file_path, dest_dir_path))
        real_src_file_path = find_real_path(src_file_path, True)[0]
        real_dest_dir_path = find_real_path(dest_dir_path, True)[0]
        return  _create_vmcfg(real_src_file_path, real_dest_dir_path)
    except Exception,e:
        errmsg = fmt_err_msg(e)
        if debug:
            print e
        logger.error('create_vmcfg: failed. =>%s'  % errmsg)
        return "failed:%s" % errmsg

###############################################################################
# Utility Interfaces

@exposed
def utl_v2v_start(src_file_path, dest_dir_path, src_type=None, dest_type=None):
    """ Start a v2v session

        Return:
            success: "success:%d"%session_id
            failed:  "failed:%s"%error_message
                     "failed:vmware-vdiskmanager doesn't found in the path."
    """
    _setprocname('utl_v2v_start %s' % os.path.basename(src_file_path))

    logger = logging.getLogger("ovs.operation")
    logger.info("utl_v2v_start(%d): %s, %s, %s, %s" % (os.getpid(), src_file_path, dest_dir_path, src_type, dest_type))
    global debug
    try:
        if not os.path.isfile(src_file_path):
            raise ValueError('failed:Invalidated source file')
        if not os.path.isdir(dest_dir_path):
            raise ValueError('failed:Invalidated destination directory')
        if src_type == None or src_type == "":
            src_type = 'vmx'
        if not src_type in _supported_src_type:
            raise ValueError('Unsupported source type', src_type)
        if dest_type == None or dest_type == "":
            dest_type = 'raw'
        if not dest_type in _supported_dest_type:
            raise ValueError('Unsupported destination type', dest_type)
    except ValueError,e:
        errmsg = fmt_err_msg(e)
        if debug:
            print e
        logger.error("utl_v2v_start(%d): failed. =>%s" % (os.getpid(), errmsg))
        return "failed:" % errmsg

    try:
        # create a daemon
        try:
            pid = os.fork()
        except OSError, e:
            raise Exception("%s [%d]" % (e.strerror, e.errno))
        if pid == 0: # child
            os.setsid()
            try:
                pid = os.fork()
            except OSError, e:
                raise Exception("%s [%d]" % (e.strerror, e.errno))
            if pid == 0: # second child
                os.chdir("/")
                os.umask(0)
            else:
                set_status(pid, "Starting convert")
                _create_pid_file(pid)
                return "success:%d" % pid
        else:
            os._exit(0)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_v2v_start(%d): failed. =>%s" % (os.getpid(), errmsg))
        return "failed:" % errmsg

    try:
        _setprocname('utl_v2v_start daemon %s' % os.path.basename(src_file_path))

        pid = os.getpid()
        logger.debug("utl_v2v_start daemon pid=%d" % pid)

        if src_type == 'vmx' and dest_type == 'raw':

            logger.info("utl_v2v_start(%d): _vmx2raw" % os.getpid())
            p = _vmx2raw(src_file_path, dest_dir_path)

            # nonblock read
            fl = fcntl.fcntl(p.stdout.fileno(), fcntl.F_GETFL)
            fcntl.fcntl(p.stdout.fileno(), fcntl.F_SETFL, fl | os.O_NDELAY)

            # update status according output of shell command
            preline = None
            while 1:
                try:
                    line = p.stdout.readline()
                    line = line.strip()

                    if debug:
                        print "line=", line
                    logger.debug('utl_v2v_start(%d): line=%s' % (os.getpid(), line))

                    # handle the case that keep reading empty line from pipe
                    if preline == line and line == "":
                        logger.debug('utl_v2v_start(%d): read empty lines from pipe of v2v_disk_convert.sh.' % os.getpid())
                        break

                    if line:
                        set_status(pid, line)
                    
                    if not line and not p.poll():
                        break

                    preline = line

                except IOError, e:
                    if e.errno != errno.EAGAIN:
                        raise e

                    time.sleep(1)

            ret = p.wait()
            if ret != 0:
                raise Exception("v2v_disk_convert: error %d. %s\n%s" % (ret, p.stdout.read(), p.stderr.read()))

            #logger.info("utl_v2v_start: create_vmcfg")
            _create_vmcfg(src_file_path, dest_dir_path)

            logger.info('utl_v2v_start(%d): done' % os.getpid())
            _remove_pid_file()

            os._exit(0)
        else:
            raise Exception("Unsupportted types (%s, %s)" % (src_type, dest_type))

    # FIXME KeyboardInterrupt and SystemExit don't inherit from Exception
    # anymore in version 2.5
    # http://docs.python.org/lib/module-exceptions.html
    except (KeyboardInterrupt, SystemExit):
        logger.info("utl_v2v_start(%d) canceled." % os.getpid())
        set_status(os.getpid(), "Error:canceled")
        _remove_pid_file()
        os._exit(0)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_v2v_start(%d) failed. =>%s" % (os.getpid(), errmsg))
        set_status(os.getpid(), "Error:%s" % errmsg)
        _remove_pid_file()
        os._exit(0)


@exposed
def utl_v2v_status(session_id):
    """Get the status of the v2v sessions

       If the session_id is null, return the a list containing all the sessions

       Return:
           success: "success:%d,%s;..."%(session_id, status)...
           failed:  "failed:%s"%error_message
    """
    if session_id and type(session_id) == type(''):
        session_id = int(session_id)

    _setprocname('utl_v2v_status %d' % session_id)

    try:
        ret = get_status(session_id)
    except Timeout,e:
        return 'CDS_lock_timeout'

    try:
        os.getpgid(session_id)
    except OSError, oserr:
        ret = "Nonexisted:%s" % ret

    return ret

@exposed
def utl_v2v_cancel(session_id):
    """ Cancel a v2v session

        Return:
            success: "success:%d" % session_id
            failed:  "failed:%s" % error_message
    """
    if session_id and type(session_id) == type(''):
        session_id = int(session_id)

    logger = logging.getLogger("ovs.operation")
    logger.info("utl_v2v_cancel %d." % session_id)

    _setprocname('utl_v2v_cancel %d' % session_id)

    try:
        os.kill(session_id, signal.SIGINT)
    except OSError, oserr:
        if oserr.errno == errno.ESRCH:
            return "success:%d,Nonexisted" % session_id
    except Exception, e:
        errmsg = fmt_err_msg(e)
        return "failed:%s" % errmsg

    return "success:%d" % session_id

@exposed
def freespace(path):
    """ Get free space of the path in M
    """
    try:
        os.mkdir(path)
    except:
        pass
    s = os.statvfs(path)
    return s[statvfs.F_BAVAIL] * long(s[statvfs.F_BSIZE]) / 1024 / 1024

###############################################################################
# Pool Master Interfaces
@exposed
def v2v_start(src_file_path, dest_dir_path, src_type=None, dest_type=None):
    """ Start a v2v session
        
        src_file_path:
            external: with proxy: proxy://[username[:password]@]proxyserver/URL
            internal: config file path

        Return:
            success: "success:%d"%session_id
            failed:  "failed:%s"%error_message
    """
    _setprocname('v2v_start %s' % os.path.basename(src_file_path))

    global debug
    logger = logging.getLogger("ovs.operation")
    files_needed_remove = []
    utl_session_id = None
    try:
        lower_src_file_path = src_file_path.lower()

        logger.info("v2v_start(%d): %s, %s, %s, %s" % (os.getpid(), src_file_path, dest_dir_path, src_type, dest_type))
        utl_srv_name, sp = get_utility_server()
        utl_srv_url = sp.agt_url
        logger.info("v2v_start(%d): selected utility server => %s. sp => %s" % (os.getpid(), utl_srv_name, sp))

        found = False
        try:
                real_dest_dir_path = find_real_path(dest_dir_path, True)[0]
                found = True
        except: #if not exists will create it later
                pass
        if not found:
            rs = sp.utl_get_storage_repos()
            if not rs.startswith("success"): raise Exception(rs)
            rs = parse_result(rs)
            repos = rs["repos"]
            real_dest_dir_path = assemble_real_path(dest_dir_path, repos)

        logger.debug("v2v_start(%d): real_dest_dir_path => %s" % (os.getpid(), real_dest_dir_path))

        proxy_server = ""
        proxy_user   = ""
        proxy_password = ""
        proxy_str_len = 0
        if lower_src_file_path.startswith("proxy://"):
            proxy_str = lower_src_file_path[len("proxy://"):]
            proxy_str = proxy_str.split('/')[0]
            proxy_str_len = len(proxy_str)
            lower_src_file_path = lower_src_file_path[len("proxy://") + len(proxy_str) + 1:]
            src_file_path = src_file_path[len("proxy://") + len(proxy_str) + 1:]
            proxy_str = proxy_str.split('@')
            if len(proxy_str) > 1:
                proxy_server = proxy_str[1]
                proxy_user = proxy_str[0]
                proxy_user = proxy_user.split(':')
                if len(proxy_user) > 1:
                        proxy_user = proxy_user[0]
                        proxy_password = proxy_user[1]
            else:
                proxy_server = proxy_str[0]

            logger.debug("v2v_start(%d): proxy_server = %s, proxy_user = %s, proxy_password = %s" % (os.getpid(), proxy_server, proxy_user, proxy_password))

        external = 0
        if lower_src_file_path.startswith("http://"):
            proto = "http://"
            external = 1
        elif lower_src_file_path.startswith("ftp://"):
            proto = "ftp://"
            external = 1
        elif lower_src_file_path.startswith("https://"):
            proto = "https://"
            external = 1

        # create a daemon
        try:
            pid = os.fork()
        except OSError, e:
            raise Exception("%s [%d]" % (e.strerror, e.errno))
        if pid == 0: # child
            os.setsid()
            try:
                pid = os.fork()
            except OSError, e:
                raise Exception("%s [%d]" % (e.strerror, e.errno))
            if pid == 0: # second child
                os.chdir("/")
                os.umask(0)
            else:
                set_status(pid, "Starting")
                _create_pid_file(pid)
                return "success:%d" % pid
        else:
            os._exit(0)

    except Exception,e:
        errmsg = fmt_err_msg(e)
        if debug:
            print e
        logger.error("v2v_start(%d): failed! => %s" % (os.getpid(), errmsg))
        set_status(os.getpid(), "Error:%s" % errmsg)
        return "failed:%s" % errmsg

    try:
        _setprocname('v2v_start daemon %s' % os.path.basename(src_file_path))

        pid = os.getpid()
        logger.debug("v2v_start daemon pid=%d" % pid)

        if external:
            #TODO move the download part to a separate function
            src_file_path = src_file_path[len(proto):]
            cfg_file_name = src_file_path.split('/')
            if debug:
                logger.info("cfg_file_name: %s" % cfg_file_name)
            cfg_file_name = cfg_file_name[len(cfg_file_name) - 1]
            if debug:
                logger.info("cfg_file_name: %s" % cfg_file_name)
            base_url      = src_file_path[:len(src_file_path) - len(cfg_file_name)]
            if debug:
                logger.info("base_url: %s" % base_url)

            # Download vmx file
            hostname = src_file_path.split('/')[0]
            src_file_path = src_file_path[len(hostname):]
            encode_url = proto + hostname + urllib.quote(src_file_path)
            basename = os.path.basename(src_file_path)

            set_status(pid, "Downloading %s" % basename)

            ret = sp.utl_wget(encode_url, real_dest_dir_path, False, "", proxy_server, proxy_user, proxy_password)
            if not ret.startswith("success"):
                raise Exception(ret)

            files_needed_remove.append(os.path.join(real_dest_dir_path, basename))

            ret = sp.get_vmx_configs(os.path.join(real_dest_dir_path, cfg_file_name))
            if not ret.startswith("success"):
                raise Exception(ret)

            # get vmdk index files
            vmdk_indexs = ret.split(':')[1].split(',')[1].split('=')[1].split(';')
            if debug:
                logger.info("vmdk_indexs:", vmdk_indexs)

            # download vmdk index files
            for vmdk_index in vmdk_indexs:
                if len(vmdk_index) == 0:    # skip empty string
                    continue

                url = base_url + vmdk_index
                hostname = url.split('/')[0]
                url = url[len(hostname):]
                encode_url = proto + hostname + urllib.quote(url)

                set_status(pid, "Downloading %s" % vmdk_index)

                ret = sp.utl_wget(encode_url, real_dest_dir_path, False, "", proxy_server, proxy_user, proxy_password)
                if not ret.startswith("success"):
                    raise Exception(ret)

                files_needed_remove.append(os.path.join(real_dest_dir_path, vmdk_index))

            # size of vmdk disk
            disksize = 0
            for vmdk_index in vmdk_indexs:
                if len(vmdk_index) == 0:    # skip empty string
                    continue

                p = vmdk_parser.vmdk_parser(os.path.join(real_dest_dir_path, vmdk_index))
                s = p.get_disk_size()
                if s == -1: # unknown size
                    s = 2048
                disksize += s

            # check free space for downloading and converting
            ret = sp.freespace(real_dest_dir_path)
            freespace_needed = 2 * disksize #Mb
            if freespace_needed >= ret:
                raise Exception("No enough free space in %s (needed:%d, free:%d)" % (real_dest_dir_path, freespace_needed, ret))
            
            # download vmdk files
            for vmdk_index in vmdk_indexs:
                if len(vmdk_index) == 0:    # skip empty string
                    continue

                ret = sp.get_vmdk_list(os.path.join(real_dest_dir_path, vmdk_index))
                if not ret.startswith("success"):
                    raise Exception(ret)

                vmdk_list = ret.split(':')[1].split(';')
                for vmdk in vmdk_list:
                    if len(vmdk) == 0:  # skip empty string
                        continue

                    url = base_url + vmdk
                    hostname = url.split('/')[0]
                    url = url[len(hostname):]
                    encode_url = proto + hostname + urllib.quote(url)

                    set_status(pid, "Downloading %s" % vmdk)

                    ret = sp.utl_wget(encode_url, real_dest_dir_path, False, "", proxy_server, proxy_user, proxy_password)
                    if not ret.startswith("success"):
                        raise Exception(ret)

                    files_needed_remove.append(os.path.join(real_dest_dir_path, vmdk))

            # set the vmx path to local path
            src_file_path = os.path.join(real_dest_dir_path, cfg_file_name)

        # local convert
        real_src_file_path = find_real_path(src_file_path, True)[0]
        real_src_base_dir = os.path.dirname(real_src_file_path)
        ret = sp.get_vmx_configs(real_src_file_path)
        if not ret.startswith("success"):
            raise Exception(ret)

        # get vmdk index files
        vmdk_indexs = ret.split(':')[1].split(',')[1].split('=')[1].split(';')
        if debug:
            logger.info("vmdk_indexs:", vmdk_indexs)

        # size of vmdk disks
        disksize = 0
        for vmdk_index in vmdk_indexs:
            if len(vmdk_index) == 0:    # skip empty string
                continue

            p = vmdk_parser.vmdk_parser(os.path.join(real_src_base_dir, vmdk_index))
            s = p.get_disk_size()
            if s == -1: # unknown size
                s = 2048
            disksize += s

        # check free space for converting
        ret = sp.freespace(real_dest_dir_path)
        freespace_needed = disksize
        if freespace_needed > 2048:
            freespace_needed = 2048
        if freespace_needed >= ret:
            raise Exception("No enough free space in %s (needed:%d, free:%d)" % (real_dest_dir_path, freespace_needed, ret))

        if src_type == None: src_type = ""
        if dest_type == None: dest_type = ""

        ret = sp.utl_v2v_start(real_src_file_path, real_dest_dir_path, src_type, dest_type)
        if not ret.startswith("success"):
            raise Exception(ret)

        utl_session_id = ret[len("success:"):]
        sts = "Converting %s" % utl_session_id
        logger.debug("v2v_start(%d) %s" % (os.getpid(), sts))
        save_utl_srv(pid, utl_srv_url)
        set_status(pid, sts)

        id = ret.split(':')[1]
        timeout_cnt = 0
        while 1:
            status = sp.utl_v2v_status(id)
            if debug:
                print "status: %s" % status
            if status.startswith('Done'):
                break
            elif status.startswith('Error:'):
                raise Exception(status[len('Error:'):])
            elif status.startswith('Nonexisted:'):
                status = status[len('Nonexisted:'):]
                if status.startswith('Done'):
                    break
                elif status.startswith('Error:'):
                    raise Exception(status[len('Error:'):])
                else:
                    raise Exception(status)
            elif status == 'CDS_lock_timeout':   # CDS timeout, wait more
                # prevent dead loop
                if timeout_cnt > 3:
                    raise Timeout('CDS lock timeout')

                logger.debug('v2v_start(%d): utl_v2v_status timout(%d). try later.' % (os.getpid(), timeout_cnt))

                time.sleep(3)
                timeout_cnt += 1

            time.sleep(3)

        if len(files_needed_remove):
            set_status(pid, "Deleting temp files")
            for i in files_needed_remove:
                sp.utl_remove(i)

        set_status(pid, "Done")
        logger.info("v2v_start(%d): done" % pid)
        _remove_pid_file()

        os._exit(0)
    # FIXME KeyboardInterrupt and SystemExit don't inherit from Exception
    # anymore in version 2.5
    except (KeyboardInterrupt, SystemExit):
        logger.info("v2v_start(%d): cancelling" % os.getpid())
        set_status(os.getpid(), "Error:cancelling")
        if utl_session_id:
            sp.utl_v2v_cancel(utl_session_id)

        if len(files_needed_remove):
            for i in files_needed_remove:
                sp.utl_remove(i)
        sp.utl_remove(os.path.join(real_dest_dir_path, 'vm.cfg'))
        set_status(os.getpid(), "Error:canceled")
        _remove_pid_file()
        logger.info("v2v_start(%d): canceled" % os.getpid())
        os._exit(0)
    except Exception,e:
        if utl_session_id:
            sp.utl_v2v_cancel(utl_session_id)

        errmsg = fmt_err_msg(e)
        if debug:
            print e
        logger.error("v2v_start(%d): failed! => %s" % (os.getpid(), errmsg))
        if len(files_needed_remove):
            for i in files_needed_remove:
                sp.utl_remove(i)
        sp.utl_remove(os.path.join(real_dest_dir_path, 'vm.cfg'))
        set_status(os.getpid(), "Error:%s" % errmsg)
        _remove_pid_file()
        os._exit(0)

@exposed
def v2v_status(session_id):
    """Get the status of the v2v sessions

       If the session_id is null, return the a list containing all the sessions

       Return:
           success: "success:%d,%s;..."%(session_id, status)...
                    "success:%d,[Nonexisted:]Error:%s" % error_msg
                    "success:%d,[Nonexisted:]Done"
                    "success:%d,[Nonexisted:]Running:%s" % last_status
           failed:  "failed:%s"%error_message
    """
    if session_id and type(session_id) == type(''):
        session_id = int(session_id)

    _setprocname('v2v_status %d' % session_id)

    try:
        st = get_status(session_id)
        if st.startswith("Converting"):
            try:
                utl_srv_url = get_utl_srv(session_id)
                if debug:
                    print "v2v_status srv_name='%s'" % srv_name
                sp = ServerProxy(utl_srv_url)
                if sp:
                    utl_pid = st.split()[1]
                    utl_st = sp.utl_v2v_status(utl_pid)
                    if utl_st != 'CDS_lock_timeout':
                        st = utl_st
            except:
                pass
    except Timeout, e:
        st = "Query timeout. Try again."
    except Exception, e:
        errmsg = fmt_err_msg(e)
        return "failed:%s" % errmsg

    ret = "success:%d,Running:%s" % (session_id, st)

    try:
        os.getpgid(session_id)
    except OSError, oserr:
        ret = "success:%d,Nonexisted:%s" % (session_id, st)

    return ret

@exposed
def v2v_cancel(session_id):
    """ Cancel a v2v session

        Return:
            success: "success:%d" % session_id
            failed:  "failed:%s" % error_message
    """
    if session_id and type(session_id) == type(''):
        session_id = int(session_id)

    logger = logging.getLogger("ovs.operation")
    logger.info("v2v_cancel %d." % session_id)

    try:
        os.getpgid(session_id)
    except OSError, oserr:
        return "success:%d,Nonexisted" % session_id

    _setprocname('v2v_cancel %d' % session_id)

    try:
        st = get_status(session_id)
        if st.startswith("Converting"):
            utl_srv_url = get_utl_srv(session_id)
            if debug:
                print "v2v_status srv_name='%s'" % srv_name
            sp = ServerProxy(utl_srv_url)
            if sp:
                utl_pid = st.split()[1]
                st = sp.utl_v2v_cancel(utl_pid)
        os.kill(session_id, signal.SIGINT)
    except OSError, oserr:
        if oserr.errno == errno.ESRCH:
            return "success:%d,Nonexisted" % session_id
    except Exception, e:
        errmsg = fmt_err_msg(e)
        return "failed:%s" % errmsg

    return "success:%d" % session_id

@exposed
def verify_image_v2v(src_dir_path):
    """ Get the src_img_type of the VM

        src_dir_path: path of source directory

        Return:
            success: "success:%s;%s"%(source_type, filename)
                source_type:
                    OracleVM: ovm
                    VMware  : vmx
            failed:  "failed:%s"%error_message
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("verify_image_v2v: %s" % src_dir_path)
        src_dir_path = find_real_path(src_dir_path, True)[0]
        mode = os.stat(src_dir_path)[ST_MODE]
        if not S_ISDIR(mode):
            ret = "verify_image_v2v: failed: %s is not a directory." % src_dir_path
            logger.error(ret)
            return "failed:%s is not a directory." % src_dir_path

        vmxs = []
        for f in os.listdir(src_dir_path):
            path = os.path.join(src_dir_path, f)
            mode = os.stat(path)[ST_MODE]
            if S_ISREG(mode):
                if f.lower() == "vm.cfg":
                    ret = "success:%s;%s" % ('ovm', f)
                    logger.info("verify_image_v2v: %s" % ret)
                    return ret
                elif f.lower().endswith(".vmx"):
                    vmxs.append(f)
        if len(vmxs):
            ret = "success:%s;%s" % ('vmx', vmxs[0])
            logger.info("verify_image_v2v: %s" % ret)
            return ret

        ret = "verify_image_v2v: failed! => Can't find source vm in the path %s" % src_dir_path
        logger.error(ret)
        return "failed:Can't find source vm in the path %s" % src_dir_path
    except Exception,e:
        errmsg = fmt_err_msg(e)
        if debug:
            print e
        ret = "verify_image_v2v: failed! => %s" % errmsg
        logger.error(ret)
        return "failed:%s" % errmsg

if __name__ == "__main__":
    debug = 1
    logging.basicConfig()
    try:
        raise Timeout("timeout test")
    except Timeout,e:
        print e
    #FIXME since the [utl_]v2v_start will run as daemon, the following
    # testing codes don't work anymore
    #ret = utl_v2v_start("/root/win98vm/win98.vmx", "/tmp")
    #print "ret=", ret

    #import time
    #while 1:
    #    print v2v_status(None)
    #    time.sleep(3)
    #v2v_start("Ubuntu.vmx", "/tmp")
