#!/usr/bin/env python
# Copyright (C) 2008, Oracle
# vmdk_parser.py
# OracleVM 2.1.2
# vmdk index file parser

import logging
import struct

from OVSCommons import exposed, fmt_err_msg

debug = 0

#typedef uint64 SectorType;
#typedef uint8 Bool;
#
#typedef struct SparseExtentHeader {
#        uint32 magicNumber;
#        uint32 version;
#        uint32 flags;
#        SectorType capacity;
#        SectorType grainSize;
#        SectorType descriptorOffset;
#        SectorType descriptorSize;
#        uint32 numGTEsPerGT;
#        SectorType rgdOffset;
#        SectorType gdOffset;
#        SectorType overHead;
#        Bool uncleanShutdown;
#        char singleEndLineChar;
#        char nonEndLineChar;
#        char doubleEndLineChar1;
#        char doubleEndLineChar2;
#        uint16 compressAlgorithm;
#        uint8 pad[433];
#} SparseExtentHeader;
#This structure needs to be packed. If you use gcc to compile your application, you must use the
#keyword __attribute__((__packed__)).

class vmdk_parser:
    """vmdk index file parser
       vmdk index file header: # Disk DescriptorFile
    """
    def __init__(self, filename):
        self._filename = filename
        #self._vmdk_list = []
        #self._is_indexer = 0
        self.parse()

    def is_indexer(self):
        return self._is_indexer

    def parse(self):
        index_header = "# Disk DescriptorFile"
        self._vmdk_list = []
        self._is_indexer = 0
        f = open(self._filename, "rb")
        ext_start = "# Extent description"
        ddb_start = "# The Disk Data Base"
        try:
            header = f.read(44)
            if debug:
                print "header=", header
            if header[:len(index_header)] == index_header:
                self._is_indexer = 1
                in_extent_desc = 0
                for line in f:
                    line = line.strip()
                    if debug > 1:
                        print "line:", line

                    if not in_extent_desc:
                        # start of Extent description
                        if line == ext_start:
                            if debug:
                                print "start of Extent description"
                            in_extent_desc = 1
                            continue

                    if in_extent_desc:
                        # end of Extent description
                        if len(line) == 0:
                            in_extent_desc = 0
                            if debug:
                                print "end of Extent description"
                            continue

                        # extent vmdk
                        if debug:
                            print "extent:", line
                        disk = line.split('"')[1]
                        if debug:
                            print "disk:", disk
                        self._vmdk_list.append(disk)

                    if line.startswith("ddb.virtualHWVersion"):
                        self.hwver = int(line.split('"')[1])
                        if debug:
                            print "hwver:", self.hwver
                    if line.startswith("ddb.geometry.cylinders"):
                        self.cylinders = int(line.split('"')[1])
                        if debug:
                            print "cylinders:", self.cylinders
                    if line.startswith("ddb.geometry.heads"):
                        self.heads = int(line.split('"')[1])
                        if debug:
                            print "heads:", self.heads
                    if line.startswith("ddb.geometry.sectors"):
                        self.sectors = int(line.split('"')[1])
                        if debug:
                            print "sectors:", self.sectors
                    if line.startswith("ddb.adapterType"):
                        self.type = line.split('"')[1]
                        if debug:
                            print "type:", self.type
            elif header[:4] == 'KDMV':  # SPARSE_MAGICNUMBER
                self._is_indexer = 0
                magic, ver, flags, capacity, grain_size, descriptor_offset, descriptor_size = struct.unpack('IIIQQQQ', header)
                #print magic, ver, flags, capacity, grain_size, descriptor_offset, descriptor_size
                self._capacity = capacity

            f.close()
        except Exception,e:
            if debug:
                print e
            raise e

    def get_extent_list(self):
        return self._vmdk_list

    def get_disk_size(self):
        """ disk size in M
            
            Return:
                failed: -1
        """
        try:
            if self._is_indexer:
                size = self.cylinders * self.heads * self.sectors
                size = size / 2 / 1024
            else:
                size = self._capacity / 2 / 1024
        except Exception,e:
            if debug:
                print e
            return -1

        return size

@exposed
def get_vmdk_list(filename):
    """Get Extent vmdk list from vmdk index file
       
       filename: path of vmdk file

       Return:
            success: "success:%s;%s..."
            failed:  "failed:%s"
    """
    logger = logging.getLogger("ovs.operation")
    logger.info("get_vmdk_list: %s..." % filename)
    try:
        p = vmdk_parser(filename)
        list = p.get_extent_list()
        ret = "success:"
        for i in list:
            ret = ret + i + ';'
        logger.info("get_vmdk_list: success. =>%s" % ret)
        return ret
    except Exception, e:
        errmsg = fmt_err_msg(e)
        if debug:
            print e
        logger.error("get_vmdk_list: failed. =>%s" % errmsg)
        return "failed:%s" % errmsg

def get_disk_size(filename):
    """ disk size in M

        Return:
            failed: -1
    """
    logger = logging.getLogger("ovs.operation")
    logger.info("get_dist_size: %s" % filename)
    p = vmdk_parser(filename)
    return p.get_disk_size()

if __name__ == "__main__":
    debug = 0
    import sys
    import getopt
    logging.basicConfig()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "d:", [])
    except getopt.error, msg:
        print msg
    for o, a in opts:
        if o in ("-d",):
            debug = int(a)

    if len(args):
        print get_vmdk_list(args[0])
        print "get_disk_size:", get_disk_size(args[0])
    else:
        print get_vmdk_list("/root/ttylinux_vmdk/ttylinux_vm.vmdk")
        print "get_disk_size:", get_disk_size("/root/ttylinux_vmdk/ttylinux_vm.vmdk")
        #print get_vmdk_list("/root/ttylinux_vmdk/ttylinux_vm-s001.vmdk")
        print get_disk_size("/root/ttylinux_vmdk/ttylinux_vm-s001.vmdk")
        #print get_vmdk_list("/root/nonexisted.vmdk")
        #print get_disk_size("/root/nonexisted.vmdk")
