#!/usr/bin/env python
# Copyright (C) 2008, Oracle
# vmx_parser.py
# OracleVM 2.1.2
# vmx config file parser

import re
import logging

from OVSCommons import exposed, fmt_err_msg

debug = 0

# Only 3 parameters are required to make a vmx file:
#   virtualHW.version = "4"
#   config.version = "4"
#   guestOS = "nameofguestos"

class vmx_parser:
    """vmx config file parser
    """
    def __init__(self, filename):
        self._filename = filename
        self.vmxcfg = {}
        self.parse()

    def is_validate_vmx(self):
        i = self.vmxcfg.has_key('virtualHW'.lower())
        i = i and self.vmxcfg['virtualHW'.lower()].has_key('version')
        i = i and self.vmxcfg.has_key('config')
        i = i and self.vmxcfg['config'].has_key('version')
        i = i and self.vmxcfg.has_key('__global__')
        i = i and self.vmxcfg['__global__'].has_key('guestOS')
        return i

    def get_memsize(self):
        try:
            return self.vmxcfg['__global__']['memsize']
        except:
            return None

    def get_config_version(self):
        try:
            return self.vmxcfg['config']['version']
        except:
            return None

    def get_virtualHW_version(self):
        try:
            return self.vmxcfg['virtualHW']['version']
        except:
            return None

    def get_guestOS(self):
        try:
            return self.vmxcfg['__global__']['guestOS']
        except:
            return None

    def get_displayName(self):
        try:
            return self.vmxcfg['__global__']['displayName']
        except:
            return None

    def _get_disks(self, type):
        ret = {}
        if self.vmxcfg.has_key(type):
            for k,v in self.vmxcfg[type].iteritems():
                if (v.has_key('present') and v['present'] == 'FALSE'):  # device present
                    continue
                if not v.has_key('partition'):
                    continue

                for k2,v in v['partition'].iteritems():
                    if v.has_key('present') and v['present'] == 'FALSE': # partition present
                        continue
                    if (v.has_key('deviceType') and v['deviceType'].startswith('cdrom')): # skip cdrom
                        continue
                    if (v.has_key('deviceType') and v['deviceType'].startswith('rawDisk')): # skip rawdisk
                        continue

                    if v.has_key('fileName'):
                        partition_index = int(v['partition_index'])
                        ret[partition_index] = v['fileName']
        return ret

    def get_ethnum(self):
        ret = 0
        type = 'ethernet'
        if self.vmxcfg.has_key(type):
            for k,v in self.vmxcfg[type].iteritems():
                if not (v.has_key('present') and v['present'] == 'FLASE'): # device present
                    ret += 1

        return ret

    def get_vcpusnum(self):
        ret = 1
        type = 'numvcpus'
        if self.vmxcfg.has_key(type):
            ret = self.vmxcfg[type]

        return ret

    def get_scsis(self):
        return self._get_disks('scsi')

    def get_ides(self):
        return self._get_disks('ide')

    def get_vmdk_list(self):
        disklist = self.get_scsis()
        disklist.update(self.get_ides())
        if debug:
            print disklist
        keys = disklist.keys()
        keys.sort()
        return [disklist[key] for key in keys if disklist[key].endswith(".vmdk")]

    def _check_key(self, dict, key):
        """ check if the dict has a key, if not, add it and initialize it as a dict
            dict and key are string
        """
        if not dict.has_key(key):
            dict[key] = {}
            return

    def parse(self):
        f = open(self._filename, "r")
        partition_index = 0
        try:
            for line in f:
                if debug:
                    print ""
                    print "line:", line
                line = line.strip()

                if len(line) < 1:
                    continue

                if line[0] == "#": # the line which starts with # is a comment line
                    continue

                tmp = line.split('=')
                if len(tmp) < 2:    # ignore the lines which are formatted
                    continue        # as target = value

                #target, value = line.split('=')
                target = tmp[0]
                value  = tmp[1]
                target = target.strip()
                value = value.strip(' "')

                sec_var = target.split('.')
                sec_var_len = len(sec_var)

                if sec_var_len == 1:    # e.g. memsize = "512"
                    sec = '__global__'  # '__global__' represents global section
                    var = sec_var[0]
                else:
                    if sec_var_len == 2: # e.g. config.version = "8"
                        sec = sec_var[0].lower()    # section is case insensitive
                        var = sec_var[1]

                sub_sec1 = None     # for the case scsi0.present and
                sub_sec2 = None     # scsi0:0.present
                                    # vmxcfg['scsi']['0']['present']
                                    # vmxcfg['scsi']['0']['partition']['0']['present']
                m = re.search('^([a-z]+)(\d+):(\d+)$', sec)
                if m:
                    if debug:
                        print "group:", m.group(1), m.group(2), m.group(3)
                    sec = m.group(1)
                    sub_sec1 = m.group(2)
                    sub_sec2 = m.group(3)
                else:
                    m = re.search('^([a-z]+)(\d+)$', sec)
                    if m:
                        if debug:
                            print "group:", m.group(1), m.group(2)
                        sec = m.group(1)
                        sub_sec1 = m.group(2)

                # init the dictionary
                self._check_key(self.vmxcfg, sec)
                if sub_sec1:
                    self._check_key(self.vmxcfg[sec], sub_sec1)
                    if sub_sec2:
                        self._check_key(self.vmxcfg[sec][sub_sec1], "partition")
                        self._check_key(self.vmxcfg[sec][sub_sec1]['partition'], sub_sec2)

                # set the value
                if value.endswith("\\"):
                    value = value + "\\"
                exec_cmd2 = ""
                if sub_sec1 and sub_sec2:
                    exec_cmd = "self.vmxcfg['%s']['%s']['partition']['%s']['%s'] = '%s'" % (sec, sub_sec1, sub_sec2, var, value)
                    # save the partition index info
                    if var == 'fileName':
                        exec_cmd2 = "self.vmxcfg['%s']['%s']['partition']['%s']['%s'] = '%d'" % (sec, sub_sec1, sub_sec2, 'partition_index', partition_index)
                        partition_index += 1
                else:
                    if sub_sec1:
                        exec_cmd = "self.vmxcfg['%s']['%s']['%s'] = '%s'" % (sec, sub_sec1, var, value)
                    else:
                        exec_cmd = "self.vmxcfg['%s']['%s'] = '%s'" % (sec, var, value)
                if debug:
                    print "set value:", exec_cmd
                exec(exec_cmd)
                if exec_cmd2:
                    exec(exec_cmd2)
        finally:
            f.close()

@exposed
def get_vmx_configs(filename):
    """ Get configs from vmx file

        filename: vmx config file path

        Return:
            success: "success:memory=%s,diskindex=%s;%s...,vif=%d,vcpus=%d"
            failed:  "failed:%s"
    """
    logger = logging.getLogger("ovs.operation")
    logger.info("get_vmx_configs: %s..." % filename)
    try:
        p = vmx_parser(filename)

        memory = p.get_memsize()

        disklist = p.get_vmdk_list()
        diskindex = ""
        for i in disklist:
            diskindex = diskindex + i + ";"

        vif = p.get_ethnum()

        vcpus = p.get_vcpusnum()

        ret = "success:memory=%s,diskindex=%s,vif=%d,vcpus=%d" % (memory, diskindex, vif, vcpus)
        logger.info("get_vmx_configs: success. =>%s" % ret)
        return ret
    except Exception,e:
        errmsg = fmt_err_msg(e)
        if debug:
            print e
        logger.error("get_vmx_configs: failed. =>%s" % errmsg)
        return "failed:%s" % errmsg

if __name__ == "__main__":
    debug = 0
    import sys
    import getopt
    logging.basicConfig()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "d:", [])
    except getopt.error, msg:
        print msg
    for o, a in opts:
        if o in ("-d",):
            debug = a
    if len(args):
        print get_vmx_configs(args[0])
        p = vmx_parser(args[0])
        print p._filename, p.is_validate_vmx()
        print "scsi desks:", p.get_scsis()
        print "ide disks:", p.get_ides()
        print "disks:", p.get_vmdk_list()
    else:
        p = vmx_parser("/var/www/html/ovm/root/win98vm/win98.vmx")
        print p._filename, p.is_validate_vmx()
        print "scsi disks:"
        print p.get_scsis()
        print "ide disks:"
        print p.get_ides()
        p = vmx_parser("/root/work/OracleVM/Agent/test/Ubuntu.vmx")
        print p._filename, p.is_validate_vmx()
        print "scsi disks:"
        print p.get_scsis()
        print "ide disks:"
        print p.get_ides()
        print p.get_displayName()
        print get_vmx_configs("/root/work/OracleVM/Agent/test/Ubuntu.vmx")
        print get_vmx_configs("/root/work/nonexisted.vmx")

#    for k,v in p.vmxcfg.iteritems():
#        if type(v) == type({}):
#            print k
#            for k,v in v.iteritems():
#                if type(v) == type({}):
#                    print "\t%s" % k
#                    print "\t\t%s" % v
#                else:
#                    print "\t%s" % k, v
#        else:
#            print k, v
